/*------------------------------------------------------------------------
 * terminal characteristics routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_set_term_db() - set term database type
 */
int tcap_set_term_db (const char *term_db_type)
{
	int type;

	if (! is_cmd_pres(term_db_type))
	{
		return (-1);
	}

	type = tcap_db_name_to_val(term_db_type);
	if (type < 0)
	{
		return (-1);
	}

	return tcap_db_set_type(type);
}

/*------------------------------------------------------------------------
 * tcap_get_term_db() - get term database type
 */
const char * tcap_get_term_db (void)
{
	int	type = tcap_db_get_type();

	return tcap_db_val_to_name(type);
}

/*------------------------------------------------------------------------
 * tcap_get_term_db_name() - get term database name
 */
const char * tcap_get_term_db_name (int num)
{
	return tcap_db_num_to_name(num);
}

/*------------------------------------------------------------------------
 * tcap_get_term_db_desc() - get term database desc
 */
const char * tcap_get_term_db_desc (int num)
{
	return tcap_db_num_to_desc(num);
}

/*------------------------------------------------------------------------
 * tcap_dump_term() - dump a term database
 */
int tcap_dump_term (const char *term, FILE *fp)
{
	return tcap_db_dump_term(tcap_db_get_type(), term, fp);
}

/*------------------------------------------------------------------------
 * tcap_set_icon() - set icon for display
 *
 * icon data is as follows:
 *
 *		ICBS esc-sequence
 *		width
 *		height
 *		((width * height) / 8) data bytes ("height" lines of "width" bits)
 *		ICBE esc-seq
 *
 *		width	one byte  encoded as (width  + ' ')
 *		height	one byte  encoded as (height + ' ')
 *		data	two bytes encoded as ((b >> 4) + ' ', (b & 0x0f) + ' ')
 */
int tcap_set_icon (TERMINAL *sp, int width, int height,
	const unsigned char *bits)
{
	if (sp == 0 ||
		width  <= 0 || width  > 64 ||
		height <= 0 || height > 64 ||
		bits == 0)
	{
		return (-1);
	}

	if (! is_cmd_pres(sp->tcap->strs.icbs) ||
	    ! is_cmd_pres(sp->tcap->strs.icbe))
	{
		return (0);
	}

	TERMINAL_LOCK(sp);
	{
		int i;
		int bytes = (width * height) / 8;

		tcap_outcmd(sp, S_ICBS, sp->tcap->strs.icbs);

		tcap_outch(sp, width  + ' ', FALSE);
		tcap_outch(sp, height + ' ', FALSE);

		for (i = 0; i < bytes; i++)
		{
			int	l = bits[i] >> 4;
			int r = bits[i]  & 0x0f;

			tcap_outch(sp, l + ' ', FALSE);
			tcap_outch(sp, r + ' ', FALSE);
		}

		tcap_outcmd(sp, S_ICBE, sp->tcap->strs.icbe);
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_evt_rtn() - set an event routine to be called
 */
int tcap_set_evt_rtn (TERMINAL *sp,
	int (*rtn)(void *data, int ms),
	void *data)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_screen_dev_evt(sp->term, rtn, data);
	}
	TERMINAL_UNLOCK(sp);

	if (rc)
		return (-1);

#ifdef SIGTSTP
	/*--------------------------------------------------------------------
	 * If we are using an event routine, we are not under control of
	 * a shell, so lose our job-control stuff (since there's no shell
	 * to restore it).
	 */
	if (data != 0)
		signal(SIGTSTP, SIG_DFL);
#endif

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_inp_rtn() - set an input routine to be called
 */
int tcap_set_inp_rtn (TERMINAL *sp,
	int (*rtn)(int mode, void *data, int ms),
	void *data)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_screen_dev_inp(sp->term, rtn, data);
	}
	TERMINAL_UNLOCK(sp);

	if (rc)
		return (-1);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_out_rtn() - set an output routine to be called
 */
int tcap_set_out_rtn (TERMINAL *sp,
	int (*rtn)(int mode, void *data, const unsigned char *buf, int n),
	void *data)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_screen_dev_out(sp->term, rtn, data);
	}
	TERMINAL_UNLOCK(sp);

	if (rc)
		return (-1);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_get_window_env() - get whether in a window or not
 */
int tcap_get_window_env (void)
{
	return term_window_check();
}

/*------------------------------------------------------------------------
 * tcap_get_disp_in_win() - get whether in a window or not
 */
int tcap_get_disp_in_win (const TERMINAL *sp)
{
	if (sp == 0)
		return (FALSE);

	return (sp->scrn->disp_in_window);
}

/*------------------------------------------------------------------------
 * tcap_get_orig_attr() - get original screen attribute if possible
 */
attr_t tcap_get_orig_attr (const TERMINAL *sp)
{
	if (sp == 0)
		return (A_UNSET);

#if V_WINDOWS
	if (sp->scrn->orig_attr == A_UNSET &&
	    sp->tcap->db_type   == TCAP_DB_WINDOWS)
	{
		int fg;
		int bg;
		int	rc;

		rc = term_win_get_color(sp->term, &fg, &bg);
		if (rc == 0)
		{
			sp->scrn->orig_attr = A_CLR(fg, bg);
		}
	}
#endif

	return (sp->scrn->orig_attr);
}

/*------------------------------------------------------------------------
 * tcap_get_termname() - get value of TERM env variable
 */
const char *tcap_get_termname (const TERMINAL *sp)
{
	return getenv("TERM");
}

/*------------------------------------------------------------------------
 * tcap_get_termtype() - get name of TERM used to load a TERMINAL struct
 */
const char *tcap_get_termtype (const TERMINAL *sp)
{
	if (sp == 0)
		return ("?");

	return (sp->tcap->tty_type);
}

/*------------------------------------------------------------------------
 * tcap_get_longname() - get description of terminal type
 */
const char *tcap_get_longname (const TERMINAL *sp)
{
	if (sp == 0)
		return ("?");

	if (is_cmd_pres(sp->tcap->strs.desc))
		return (sp->tcap->strs.desc);
	else
		return (sp->tcap->tty_type);
}

/*------------------------------------------------------------------------
 * tcap_get_ttyname() - get name of device for a terminal
 */
const char *tcap_get_ttyname (const TERMINAL *sp)
{
	if (sp == 0)
		return ("?");

	return term_get_tty_name(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_max_rows() - get max rows on screen
 */
int tcap_get_max_rows (const TERMINAL *sp)
{
	if (sp == 0)
		return (0);

	return (sp->tcap->ints.maxrows);
}

/*------------------------------------------------------------------------
 * tcap_get_max_cols() - get max cols on screen
 */
int tcap_get_max_cols (const TERMINAL *sp)
{
	if (sp == 0)
		return (0);

	return (sp->tcap->ints.maxcols);
}

/*------------------------------------------------------------------------
 * tcap_get_tab_size() - get tab width for a screen
 */
int tcap_get_tab_size (const TERMINAL *sp)
{
	int ts = 8;

	if (sp != 0)
	{
		if (sp->tcap->ints.tabsize != 0)
			ts = sp->tcap->ints.tabsize;
	}

	return (ts);
}

/*------------------------------------------------------------------------
 * tcap_get_has_ic() - check whether ic/dc capability is present
 */
int tcap_get_has_ic (const TERMINAL *sp)
{
	if (sp == 0)
		return (FALSE);

	if (is_cmd_pres(sp->tcap->strs.ic) && is_cmd_pres(sp->tcap->strs.dc))
		return (TRUE);
	else
		return (FALSE);
}

/*------------------------------------------------------------------------
 * tcap_get_has_il() - check whether il/dl capability is present
 */
int tcap_get_has_il (const TERMINAL *sp)
{
	if (sp == 0)
		return (FALSE);

	if (is_cmd_pres(sp->tcap->strs.al) && is_cmd_pres(sp->tcap->strs.dl))
		return (TRUE);
	else
		return (FALSE);
}

/*------------------------------------------------------------------------
 * tcap_get_termattrs() - get mask of attributes available
 */
attr_t tcap_get_termattrs (const TERMINAL *sp)
{
	attr_t ch = 0;

	if (sp == 0)
		return (ch);

	if ( is_cmd_pres(sp->tcap->strs.sts) )	ch |= A_STANDOUT;
	if ( is_cmd_pres(sp->tcap->strs.uls) )	ch |= A_UNDERLINE;
	if ( is_cmd_pres(sp->tcap->strs.bds) )	ch |= A_BOLD;
	if ( is_cmd_pres(sp->tcap->strs.bls) )	ch |= A_BLINK;
	if ( is_cmd_pres(sp->tcap->strs.dms) )	ch |= A_DIM;
	if ( is_cmd_pres(sp->tcap->strs.ivs) )	ch |= A_INVIS;
	if ( is_cmd_pres(sp->tcap->strs.prs) )	ch |= A_PROTECT;

	return (ch);
}

/*------------------------------------------------------------------------
 * tcap_get_cmdstr() - get cmd string associated with a TC_DATA entry
 */
static int tcap_get_cmdstr (const TERMINAL *sp, const TC_DATA *tc, char *buf)
{
	const TCAP_DATA *	tp	= sp->tcap;
	const char **		cp;
	const int *			ip;
	const int *			bp;

	switch (tc->trmc_flags & TC_TYPE)
	{
	case TC_NUM:
		ip = TCAP_TC_NUM_GP(tp, tc);
		sprintf(buf, "%d\n", *ip);
		break;

	case TC_BLN:
		bp = TCAP_TC_BLN_GP(tp, tc);
		sprintf(buf, "%d\n", *bp);
		break;

	case TC_STR:
		cp = TCAP_TC_STR_GP(tp, tc);
		if (! is_cmd_pres(*cp))
			*buf = 0;
		else
			strcpy(buf, *cp);
		break;

	default:
		*buf = 0;
		break;
	}

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_get_cmdstr_by_enum() - get cmd string associated with an enum
 */
int tcap_get_cmdstr_by_enum (const TERMINAL *sp, SCRN_CMD cmd, char *buf)
{
	const TC_DATA *		tc;
	const TCAP_DATA *	tp;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0 || buf == 0)
		return (-1);

	tp = sp->tcap;

	*buf = 0;

	/*--------------------------------------------------------------------
	 * special case of database type
	 */
	if (cmd == S_DB)
	{
		sprintf(buf, "%s\n", tcap_db_num_to_name(tp->db_type));
		return (0);
	}

	/*--------------------------------------------------------------------
	 * find entry in tcap table
	 */
	for (tc=tcap_termcap_data; tc->trmc_label; tc++)
	{
		if (tc->trmc_cmd == cmd)
			break;
	}

	if (tc->trmc_label == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * return entry
	 */
	return tcap_get_cmdstr(sp, tc, buf);
}

/*------------------------------------------------------------------------
 * tcap_get_cmdstr_by_name() - get cmd string associated with a name
 */
int tcap_get_cmdstr_by_name (const TERMINAL *sp, const char *name,
	SCRN_CMD *pcmd, char *buf)
{
	const TC_DATA *		tc;
	const KF_TABLE *	kf;
	const TCAP_DATA *	tp;
	int					key;
	int					clr;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0 || buf == 0 || pcmd == 0 || name == 0 || *name == 0)
		return (-1);

	tp = sp->tcap;

	*pcmd	= 0;
	*buf	= 0;

	/*--------------------------------------------------------------------
	 * special case of database type
	 */
	if (tcap_trmcmp(name, tcap_keyword_dbtype) == 0)
	{
		*pcmd = S_DB;
		return tcap_get_cmdstr_by_enum(sp, S_DB, buf);
	}

	/*--------------------------------------------------------------------
	 * find entry in tcap table
	 */
	for (tc=tcap_termcap_data; tc->trmc_label; tc++)
	{
		if (tc->trmc_label  && tcap_trmcmp(name, tc->trmc_label)  == 0)
			break;

		if (tc->trmc_tcname && tcap_trmcmp(name, tc->trmc_tcname) == 0)
			break;

		if (tc->trmc_tiname && tcap_trmcmp(name, tc->trmc_tiname) == 0)
			break;
	}

	/*--------------------------------------------------------------------
	 * if tcap entry found, return it
	 */
	if (tc->trmc_label != 0)
	{
		*pcmd = tc->trmc_cmd;
		return tcap_get_cmdstr(sp, tc, buf);
	}

	/*--------------------------------------------------------------------
	 * if entry not found, check if it is a function key name
	 */
	key = -1;

	for (kf=tcap_tc_keys; kf->kf_tcname; kf++)
	{
		if (tcap_trmcmp(kf->kf_tcname, name) == 0 ||
		    tcap_trmcmp(kf->kf_tiname, name) == 0)
		{
			key = kf->kf_code;
			break;
		}
	}

	if (key == -1)
		key = tcap_get_key_value(name);

	/*--------------------------------------------------------------------
	 * if function key name, look up its value
	 */
	if (key != -1)
	{
		KFUNC *	k;

		for (k=tp->kfuncs; k; k=k->kf_next)
		{
			if (k->kf_code == key)
			{
				strcpy(buf, k->kf_str);
				break;
			}
		}

		return (0);
	}

	/*--------------------------------------------------------------------
	 * if not a function key, check if color name
	 */
	clr = tcap_get_fg_num_by_name(name);
	if (clr != -1)
	{
		if (is_cmd_pres(tp->strs.clr_fg[clr]))
			strcpy(buf, tp->strs.clr_fg[clr]);

		return (0);
	}

	clr = tcap_get_bg_num_by_name(name);
	if (clr != -1)
	{
		if (is_cmd_pres(tp->strs.clr_bg[clr]))
			strcpy(buf, tp->strs.clr_bg[clr]);

		return (0);
	}

	return (-1);
}

/*------------------------------------------------------------------------
 * tcap_get_cmdstr_by_str() - get cmd string associated with a string
 *
 * returns:
 *		-1	error
 *		 0	got parameter string in buf
 *		 1	got parameter values in p1 & p2
 */
int tcap_get_cmdstr_by_str (const TERMINAL *sp, const char *str,
	SCRN_CMD *pcmd, char *buf, int *pp1, int *pp2)
{
	const TC_DATA *	tc;
	SCRN_CMD		cmd;
	char			name[128];
	char			data[128];
	char *			n;
	const char **	vp;
	int				p1;
	int				p2;
	int				rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (sp == 0 || str == 0 || pcmd == 0 || buf == 0 || pp1 == 0 || pp2 == 0)
	{
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * clear args
	 */
	*pcmd	= 0;
	*buf	= 0;
	*pp1	= 0;
	*pp2	= 0;

	p1		= 0;
	p2		= 0;

	/*--------------------------------------------------------------------
	 * collect name part
	 */
	strcpy(name, str);
	n = strchr(name, ':');
	if (n != 0)
		*n++ = 0;
	else
		n = "";

	/*--------------------------------------------------------------------
	 * lookup name
	 */
	if (tcap_trmcmp(name, "fg") == 0)
	{
		cmd = S_FG;
	}
	else if (tcap_trmcmp(name, "bg") == 0)
	{
		cmd = S_BG;
	}
	else
	{
		tc = tcap_tc_find_entry(name);
		if (tc == 0)
			return (-1);
		cmd	= tc->trmc_cmd;
	}

	/*--------------------------------------------------------------------
	 * decode entry
	 */
	switch (cmd)
	{
	case S_CM:
		p1 = atoi(n);
		n = strchr(n, ',');
		if (n != 0)
			p2 = atoi(n+1);
		rc = 1;
		break;

	case S_CS:
		strcpy(data, n);
		n = strchr(data, '/');
		if (n == 0)
			n = "";
		else
			*n++ = 0;
		p2 = tcap_get_fg_num_by_name(data);
		p1 = tcap_get_bg_num_by_name(n);
		rc = 1;
		break;

	case S_FG:
		p1 = tcap_get_fg_num_by_name(n);
		cmd = S_FG;
		rc = 1;
		break;

	case S_BG:
		p1 = tcap_get_bg_num_by_name(n);
		cmd = S_BG;
		rc = 1;
		break;

	default:
		vp = TCAP_TC_STR_GP(sp->tcap, tc);
		if (is_cmd_pres(*vp))
			strcpy(buf, *vp);
		rc = 0;
		break;
	}

	/*--------------------------------------------------------------------
	 * store results
	 */
	*pcmd	= cmd;
	*pp1	= p1;
	*pp2	= p2;

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_get_default_attrs() - get default attributes
 */
int tcap_get_default_attrs (const TERMINAL *sp, attr_t *pr, attr_t *ps,
	int use_color, int use_win)
{
	attr_t orig;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0 || pr == 0 || ps == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * figure out attributes
	 */
	orig = tcap_get_orig_attr(sp);

	if (use_color && tcap_has_colors(sp))
	{

		if (orig != A_UNSET)
		{
			int f = A_FG_CLRNUM(orig);
			int b = A_BG_CLRNUM(orig);

			*pr	= orig;
			*ps	= A_CLR(b, f);
		}
		else
		{
			/*----------------------------------------------------
			 * if in a window, set to black-on-ltwhite
			 * if not,         set to white-on-black
			 */
			if (use_win && tcap_get_disp_in_win(sp))
			{
				*pr	= A_CLR(COLOR_BLACK,   COLOR_LTWHITE);
				*ps	= A_CLR(COLOR_LTWHITE, COLOR_BLACK);
			}
			else
			{
				*pr	= A_CLR(COLOR_LTWHITE, COLOR_BLACK);
				*ps	= A_CLR(COLOR_BLACK,   COLOR_LTWHITE);
			}
		}
	}
	else
	{
		*pr	= A_NORMAL;
		*ps	= A_REVERSE;
	}

	return (0);
}
