/*------------------------------------------------------------------------
 * screen-saver routine
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * static constants used in routine
 */
static const int tcap_save_intvl	= 5000;		/* # of msecs to wait */
static const int tcap_max_rand		= 0x7fff;	/* random number mask */

/*------------------------------------------------------------------------
 * random function used
 */
#define random()		(rand() & tcap_max_rand)

/*------------------------------------------------------------------------
 * tcap_blank_screen() - blank the screen with a bouncing logo
 */
int tcap_blank_screen (TERMINAL *sp, const char **logo)
{
	int		y, x;
	int		old_y,  old_x;
	int		orig_y, orig_x;
	int		c;
	int		logo_width;
	int		logo_height;
	int		i;
	int		l;
	int		clock_was_running;
	int		cursor;
	int		do_color;

	if (sp == 0)
		return (-1);

#if 0
	/*--------------------------------------------------------------------
	 * If the strings exist to iconify the window,
	 * then reduce to an icon instead of blanking the window.
	 *
	 * The problem is that we do not know when we are
	 * uniconified, i.e. redisplayed.  At that time,
	 * we should clear any elapsed time counters.
	 */
	if (is_cmd_pres(sp->tcap->strs.icc))
	{
		tcap_outcmd(sp, S_ICC, sp->tcap->strs.icc);
		return (0);
	}
#endif

	/*--------------------------------------------------------------------
	 * check for user-supplied refresh routine
	 *
	 * If we don't have one, just bail since we cannot restore the screen.
	 */
	if (! tcap_async_check(sp->scrn, KEY_REFRESH))
		return (-1);

	/*--------------------------------------------------------------------
	 * cache current cursor position
	 */
	orig_x = sp->scrn->curr_col;
	orig_y = sp->scrn->curr_row;

	/*--------------------------------------------------------------------
	 * get number of lines & length of longest line in logo
	 */
	logo_height	= 0;
	logo_width	= 0;
	for (i=0; logo[i]; i++)
	{
		logo_height++;

		l = strlen(logo[i]);
		if (l > logo_width)
			logo_width = l;
	}

	/*--------------------------------------------------------------------
	 * turn off clock if running
	 */
	clock_was_running = term_clock_check();
	if (clock_was_running)
		term_clock_off();

	/*--------------------------------------------------------------------
	 * turn off cursor if possible
	 */
	cursor = tcap_curs_set(sp, VISIBILITY_OFF);

	/*--------------------------------------------------------------------
	 * check if display in color
	 */
	do_color = ((sp->scrn->curr_attr & A_COLOR) != 0);

	/*--------------------------------------------------------------------
	 * now clear the screen, display a bouncing logo, & wait for a keypress
	 */
	old_y = old_x = -1;

	while (TRUE)
	{
		/*----------------------------------------------------------------
		 * if in color, choose a random color
		 */
		if (do_color)
		{
			/*------------------------------------------------------------
			 * random color between (1-(NUM_FG_FND-1))
			 */
			int f = ((random() *
				(sp->tcap->ints.max_fg - 1)) / tcap_max_rand) + 1;
			int b = sp->scrn->disp_in_window ? COLOR_LTWHITE : COLOR_BLACK;

			tcap_outattr(sp, A_CLR(f, b), ' ');
		}

		/*----------------------------------------------------------------
		 * clear either the whole screen or just our logo
		 */
		if (old_x == -1)
		{
			/*------------------------------------------------------------
			 * first time: clear the whole screen
			 */
			tcap_clear_screen(sp);
		}
		else
		{
			/*------------------------------------------------------------
			 * If we have a CL cmd, it is quicker to use
			 */
			if (is_cmd_pres(sp->tcap->strs.cl))
			{
				tcap_outcmd(sp, S_CL, sp->tcap->strs.cl);
			}
			else
			{
				for (i=0; i<logo_height; i++)
				{
					l = strlen(logo[i]);
					tcap_outpos(sp, old_y+i, old_x+(logo_width-l)/2);
					for (; l; l--)
						tcap_outc(sp, ' ');
				}
			}
		}

		/*----------------------------------------------------------------
		 * get random (y, x) to display logo at
		 */
		y = (random() *
			(sp->tcap->ints.maxrows - (logo_height-1))) / tcap_max_rand;
		x = (random() *
			(sp->tcap->ints.maxcols - (logo_width -1))) / tcap_max_rand;

		/*----------------------------------------------------------------
		 * now display our logo at that position
		 */
		for (i=0; i<logo_height; i++)
		{
			l = strlen(logo[i]);
			tcap_outpos(sp, y+i, x+(logo_width-l)/2);
			tcap_outs(sp, logo[i]);
		}

		tcap_outflush(sp);

		/*----------------------------------------------------------------
		 * cache this position for next time around
		 */
		old_y = y;
		old_x = x;

		/*----------------------------------------------------------------
		 * now wait for a key-press & break if we got one
		 */
		c = tcap_kbd_get_key(sp, tcap_save_intvl, TRUE, TRUE, TRUE);
		if (c != -1)
			break;
	}

	/*--------------------------------------------------------------------
	 * call user-specified refresh routine
	 */
	tcap_async_process(sp->scrn, KEY_REFRESH);

	/*--------------------------------------------------------------------
	 * restore cursor state
	 */
	tcap_curs_set(sp, cursor);

	/*--------------------------------------------------------------------
	 * turn clock back on if it was running
	 */
	if (clock_was_running)
		term_clock_on();

	/*--------------------------------------------------------------------
	 * reset original cursor position
	 */
	if (orig_x != -1 && orig_y != -1)
	{
		tcap_outpos(sp, orig_y, orig_x);
		tcap_outflush(sp);
	}

	return (0);
}
