/*------------------------------------------------------------------------
 * terminal init/deinit routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_cinit() - init terminal state stuff
 */
int tcap_cinit (TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * cache old & new terminal states
		 */
		tcap_get_orig_mode(sp);
		tcap_get_prog_mode(sp);

		/*----------------------------------------------------------------
		 * now switch terminal to new state
		 */
		tcap_set_prog_mode(sp);
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_cexit() - uninit terminal state stuff
 */
int tcap_cexit (TERMINAL *sp, int exit_flag)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * switch terminal to orig state
		 */
		tcap_set_orig_mode(sp);

		/*----------------------------------------------------------------
		 * if exiting, clear out everything
		 */
		if (exit_flag)
		{
			/*------------------------------------------------------------
			 * close out debug logs
			 */
			tcap_kbd_debug(sp, 0, FALSE, FALSE);
			tcap_out_debug(sp, 0, FALSE, FALSE);

			/*------------------------------------------------------------
			 * clear signal table
			 */
			tcap_signal_set_rtn(sp, -1, 0, 0);

			/*------------------------------------------------------------
			 * close devices
			 */
			term_screen_dev_close(sp->term);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_exec_shell() - shell out to a command shell
 */
int tcap_exec_shell (TERMINAL *sp)
{
	int status;
	int y;
	int x;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * ignore request if no way to refresh the screen
	 */
	if (! tcap_async_check(sp->scrn, KEY_REFRESH))
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*------------------------------------------------------------
		 * save original screen position & clear the screen
		 */
		y = sp->scrn->curr_row;
		x = sp->scrn->curr_col;

		tcap_clear_screen(sp);
		tcap_outpos(sp, 0, 0);
		tcap_outflush(sp);

		/*------------------------------------------------------------
		 * shell out
		 */
		tcap_set_orig_mode(sp);
		status = tcap_exec_cmd(sp, NULL, NULL, TRUE, NULL);
		tcap_set_prog_mode(sp);

		/*------------------------------------------------------------
		 * refresh the screen
		 */
		tcap_async_process(sp->scrn, KEY_REFRESH);

		/*------------------------------------------------------------
		 * re-position cursor to where it was
		 */
		tcap_outpos(sp, y, x);
		tcap_outflush(sp);
	}
	TERMINAL_UNLOCK(sp);

	return (status);
}

/*------------------------------------------------------------------------
 * tcap_get_prog_mode() - get program terminal mode
 */
int tcap_get_prog_mode (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_get_prog_mode(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_get_orig_mode() - get original terminal mode
 */
int tcap_get_orig_mode (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_get_orig_mode(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_prog_mode() - put screen in program mode
 */
int tcap_set_prog_mode (TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * set terminal to program state
		 */
		term_set_prog_mode(sp->term);

		/*----------------------------------------------------------------
		 * indicate current position unknown
		 */
		sp->scrn->curr_row = -1;
		sp->scrn->curr_col = -1;

		/*----------------------------------------------------------------
		 * reset alarm time if any
		 */
		if (sp->scrn->alarm_time_left)
		{
			term_alarm_start(sp->scrn->alarm_time_left);
			sp->scrn->alarm_time_left = 0;
		}

		/*----------------------------------------------------------------
		 * send term-initialization strings
		 */
		tcap_outcmd(sp, S_TI,  sp->tcap->strs.ti);
		tcap_outcmd(sp, S_ACE, sp->tcap->strs.ace);

		/*----------------------------------------------------------------
		 * inform parent to turn on any keyboard mapping
		 */
		tcap_outcmd(sp, S_KBO, sp->tcap->strs.kbo);

		/*----------------------------------------------------------------
		 * check if mouse was in effect
		 */
		if (sp->scrn->mouse_restore)
		{
			sp->scrn->mouse_restore = FALSE;
			tcap_mouse_init(sp);
		}

		/*----------------------------------------------------------------
		 * restore window title if there was one
		 */
		if (*sp->scrn->win_title != 0)
		{
			tcap_set_window_title(sp, 0);
		}

		/*----------------------------------------------------------------
		 * flush any pending output
		 */
		tcap_outflush(sp);

		/*----------------------------------------------------------------
		 * initialize signals
		 */
		tcap_sig_init(sp);

		/*----------------------------------------------------------------
		 * set mode on
		 */
		sp->scrn->mode = TRUE;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_orig_mode() - restore screen to original mode
 */
int tcap_set_orig_mode (TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * check if already in orig mode
	 */
	if (! sp->scrn->mode)
		return (0);

	/*--------------------------------------------------------------------
	 * now do it
	 */
	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * end signal processing
		 */
		tcap_sig_end(sp);

		/*----------------------------------------------------------------
		 * save old alarm time
		 */
		sp->scrn->alarm_time_left = term_alarm_stop();

		/*----------------------------------------------------------------
		 * check if mouse in effect
		 */
		if (tcap_mouse_get_is_on(sp))
		{
			sp->scrn->mouse_restore = TRUE;
			tcap_mouse_end(sp);
		}

		/*----------------------------------------------------------------
		 * set fonts to normal
		 */
		if (sp->scrn->curr_reg_font)
		{
			tcap_outcmd(sp, tcap_reg_font_cmds[sp->scrn->curr_reg_font].off,
				sp->tcap->strs.rne[sp->scrn->curr_reg_font-1]);
			sp->scrn->curr_reg_font = 0;
		}

		if (sp->scrn->curr_alt_font)
		{
			tcap_outcmd(sp, tcap_alt_font_cmds[sp->scrn->curr_alt_font].off,
				sp->tcap->strs.ane[sp->scrn->curr_alt_font-1]);
			sp->scrn->curr_alt_font = 0;
		}

		/*----------------------------------------------------------------
		 * reset screen attrs if original known
		 */
		if (sp->scrn->orig_attr != A_UNSET)
		{
			tcap_outattr(sp, sp->scrn->orig_attr, ' ');
			tcap_clear_screen(sp);
			tcap_outpos(sp, 0, 0);
		}

		/*----------------------------------------------------------------
		 * restore cursor if different from normal
		 */
		tcap_curs_set(sp, VISIBILITY_ON);

		/*----------------------------------------------------------------
		 * inform parent to turn off any keyboard mapping
		 */
		tcap_outcmd(sp, S_KBF, sp->tcap->strs.kbf);

		/*----------------------------------------------------------------
		 * send term-uninit strings
		 */
		tcap_outcmd(sp, S_ACD, sp->tcap->strs.acd);
		tcap_outcmd(sp, S_TE,  sp->tcap->strs.te);

		/*----------------------------------------------------------------
		 * flush any pending output
		 */
		tcap_outflush(sp);

		/*----------------------------------------------------------------
		 * Before flushing any input, give the terminal emulator time
		 * to finish up what he is told to do.
		 */
		term_nap(100);

		/*----------------------------------------------------------------
		 * This won't flush any input (that was already done, but
		 * it will give any event handlers one last chance.
		 */
		tcap_outflush(sp);

		/*----------------------------------------------------------------
		 * Flush any pending input.
		 * This is needed to clear out any "mapped" keyboard input.
		 */
		tcap_kbd_clr_ta(sp);

		/*----------------------------------------------------------------
		 * set terminal to original state
		 */
		term_set_orig_mode(sp->term);

		/*----------------------------------------------------------------
		 * indicate in orig mode
		 */
		sp->scrn->mode = FALSE;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_get_save_mode() - save current terminal mode
 */
int tcap_get_save_mode (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_get_save_mode(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_save_mode() - set terminal to saved terminal mode
 */
int tcap_set_save_mode (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_save_mode(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}
