/*------------------------------------------------------------------------
 * device output routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_outflush() - flush any buffered output
 */
int tcap_outflush (TERMINAL *sp)
{
	int rc;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * now do the flush
	 */
	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * write any buffered chars
		 */
		rc = term_screen_dev_write(sp->term, sp->scrn->out_buf,
			sp->scrn->out_buf_count);

		/*----------------------------------------------------------------
		 * set buffered char count to 0
		 */
		sp->scrn->out_buf_count = 0;

		/*----------------------------------------------------------------
		 * flush debug file
		 */
		if (sp->scrn->debug_scrn_fp != 0)
			fflush(sp->scrn->debug_scrn_fp);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_outch() - output a character (either data or cmd part)
 */
int tcap_outch (TERMINAL *sp, int c, int data)
{
	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * output the char
	 */
	TERMINAL_LOCK(sp);
	{
#if V_WINDOWS
		/*----------------------------------------------------------------
		 * Win98 has a bug in that if you automatically move to the last
		 * column, the system will "wrap" it to the first col of that row.
		 * So, to avoid that problem, if we are about to buffer a char in
		 * the last col, we "move" to that position first.
		 */
		if (sp->tcap->db_type == TCAP_DB_WINDOWS)
		{
			if (sp->scrn->curr_col == (sp->tcap->ints.maxcols - 1))
			{
				tcap_outparm(sp, S_CM, sp->scrn->curr_row, sp->scrn->curr_col);
			}
		}
#endif

		/*----------------------------------------------------------------
		 * buffer this char
		 */
		sp->scrn->out_buf[sp->scrn->out_buf_count++] = (unsigned char)c;

		/*----------------------------------------------------------------
		 * if buffer is full, flush it
		 */
		if (sp->scrn->out_buf_count == sp->scrn->out_buf_size)
			tcap_outflush(sp);

		/*----------------------------------------------------------------
		 * debug print
		 */
		if (sp->scrn->debug_scrn_fp != 0)
		{
			if (data || ! sp->scrn->debug_scrn_text)
				putc(c, sp->scrn->debug_scrn_fp);
			sp->scrn->debug_scrn_mode = ! data;

			/*------------------------------------------------------------
			 * If doing text debugging & we have AM & we are outputting
			 * data & we are at EOL, output an nl to avoid real long lines.
			 */
			if (data && sp->scrn->debug_scrn_text && sp->tcap->bools.am)
			{
				if (sp->scrn->curr_col == (sp->tcap->ints.maxcols - 1))
				{
					putc('\n', sp->scrn->debug_scrn_fp);
					sp->scrn->debug_scrn_mode = TRUE;
				}
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_outchar() - output a char from a parameterized cmd string
 */
int tcap_outchar (TERMINAL *sp, int c)
{
	return tcap_outch(sp, c, FALSE);
}

/*------------------------------------------------------------------------
 * tcap_outcmd() - output a cmd
 */
int tcap_outcmd (TERMINAL *sp, SCRN_CMD cmd, const char *s)
{
	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * bail if string not present
	 */
	if (! is_cmd_pres(s))
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * debug print
		 */
		if (sp->scrn->debug_scrn_fp != 0 && sp->scrn->debug_scrn_text)
		{
			const char *p = tcap_tc_get_cmd(sp->tcap, cmd);

			if (p != 0)
			{
				switch (cmd)
				{
				case S_FG:
				case S_BG:
				case S_CM:
				case S_CS:
					break;

				default:
					if (! sp->scrn->debug_scrn_mode)
						fprintf(sp->scrn->debug_scrn_fp, "\n");
					sp->scrn->debug_scrn_mode = TRUE;
					fprintf(sp->scrn->debug_scrn_fp, "<%s>\n", p);
					break;
				}
				fflush(sp->scrn->debug_scrn_fp);
			}
		}

		/*----------------------------------------------------------------
		 * now output the cmd
		 */
#if V_WINDOWS
		if (sp->tcap->db_type == TCAP_DB_WINDOWS)
		{
			switch (cmd)
			{
			case S_BL:			/* bell */
				break;

			case S_VE:			/* cursor normal */
			case S_VS:			/* cursor very visible */
				term_win_set_cursor(sp->term, TRUE);
				break;

			case S_VI:			/* cursor invisible */
				term_win_set_cursor(sp->term, FALSE);
				break;

			case S_STS:			/* standout start */
				tcap_outparm(sp, S_CS,
					A_FG_CLRNUM(sp->scrn->orig_attr),
					A_BG_CLRNUM(sp->scrn->orig_attr));
				break;

			case S_STE:			/* standout stop */
				tcap_outparm(sp, S_CS,
					A_BG_CLRNUM(sp->scrn->orig_attr),
					A_FG_CLRNUM(sp->scrn->orig_attr));
				break;

			case S_MBEG:		/* mouse start */
				term_win_set_mouse(sp->term, TRUE);
				break;

			case S_MEND:		/* mouse stop */
				term_win_set_mouse(sp->term, FALSE);
				break;

			case S_TTS:			/* start window title */
				term_win_set_title(sp->term, s);
				break;

			case S_TTE:			/* end window title */
				break;
			}
		}
		else
#endif
		{
			if (*s)
			{
				tcap_outflush(sp);
				tcap_db_outs(sp, s, tcap_outchar);
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_outparm() - process a cmd which is parameterized
 */
int tcap_outparm (TERMINAL *sp, SCRN_CMD cmd, int p1, int p2)
{
	int	rc = -1;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * now do it
	 */
	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * debug print
		 */
		if (sp->scrn->debug_scrn_fp != 0 && sp->scrn->debug_scrn_text)
		{
			const char *p = tcap_tc_get_cmd(sp->tcap, cmd);

			if (! sp->scrn->debug_scrn_mode)
				fprintf(sp->scrn->debug_scrn_fp, "\n");
			sp->scrn->debug_scrn_mode = TRUE;

			switch (cmd)
			{
			case S_CM:
				fprintf(sp->scrn->debug_scrn_fp, "<%s:%d,%d>\n",
					p, p1, p2);
				fflush(sp->scrn->debug_scrn_fp);
				break;

			case S_CS:
				fprintf(sp->scrn->debug_scrn_fp, "<%s:%s/%s>\n", p,
					tcap_get_fg_name_by_num(p2), tcap_get_bg_name_by_num(p1));
				fflush(sp->scrn->debug_scrn_fp);
				break;

			case S_FG:
				fprintf(sp->scrn->debug_scrn_fp, "<%s:%s>\n",
					p, tcap_get_fg_name_by_num(p1));
				fflush(sp->scrn->debug_scrn_fp);
				break;

			case S_BG:
				fprintf(sp->scrn->debug_scrn_fp, "<%s:%s>\n",
					p, tcap_get_bg_name_by_num(p1));
				fflush(sp->scrn->debug_scrn_fp);
				break;
			}
		}

		/*----------------------------------------------------------------
		 * parameterize the data & output it
		 */
#if V_WINDOWS
		if (sp->tcap->db_type == TCAP_DB_WINDOWS)
		{
			tcap_outflush(sp);

			switch (cmd)
			{
			case S_CM:			/* cursor move */
				rc = term_win_set_position(sp->term, p1, p2);
				break;

			case S_CS:			/* color set */
				rc = term_win_set_color(sp->term, p2, p1);
				break;
			}
		}
		else
#endif
		{
			char buf[128];

			switch (cmd)
			{
			case S_CM:
			case S_CS:
				tcap_eval_parm(sp, buf, cmd, p1, p2);
				rc = tcap_outcmd(sp, cmd, buf);
				break;

			case S_FG:
				rc = tcap_outcmd(sp, S_FG, sp->tcap->strs.clr_fg[p1]);
				break;

			case S_BG:
				rc = tcap_outcmd(sp, S_BG, sp->tcap->strs.clr_bg[p1]);
				break;
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}
