/*------------------------------------------------------------------------
 * mouse routines
 *
 * Note: all these routines assume a valid TERMINAL pointer
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * defines
 */
#define MOUSE_STAT(sp)		( term_screen_dev_mouse(sp->term) )

/*------------------------------------------------------------------------
 * tcap_mouse_init() - initialize mouse processing
 */
int tcap_mouse_init (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	if (! tcap_mouse_get_display(sp))
	{
		return (-1);
	}

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);

		if (! mp->is_init)
		{
			int i;

			if (sp->tcap->ints.nb > NUM_BUTTONS)
				sp->tcap->ints.nb = NUM_BUTTONS;

			for (i=0; i<NUM_BUTTONS; i++)
			{
				mp->button[i]			= BUTTON_UP;
				mp->btime[i].tv_sec		= 0;
				mp->btime[i].tv_usec	= 0;
			}

			mp->is_on				= FALSE;
			mp->dbl_clk_interval	= sp->tcap->ints.dblclk_intvl;
			mp->y					= -1;
			mp->x					= -1;
			mp->old_y				= -1;
			mp->old_x				= -1;
			mp->is_init				= TRUE;

			rc = tcap_mouse_show(sp);
		}
		else
		{
			rc = 0;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_end() - end mouse processing
 */
int tcap_mouse_end (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		if (mp->is_on)
			tcap_mouse_hide(sp);
		mp->is_init = FALSE;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_show() - turn the mouse on
 */
int tcap_mouse_show (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	if (! tcap_mouse_get_display(sp))
	{
		return (-1);
	}

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		if (! mp->is_on)
		{
			mp->is_on = TRUE;
			tcap_outcmd(sp, S_MBEG, sp->tcap->strs.mbeg);
			tcap_outflush(sp);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_hide() - turn the mouse off
 */
int tcap_mouse_hide (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		if (mp->is_on)
		{
			mp->is_on = FALSE;
			tcap_outcmd(sp, S_MEND, sp->tcap->strs.mend);
			tcap_outflush(sp);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_is_on() - check whether mouse is on or off
 */
int tcap_mouse_get_is_on (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (FALSE);

	mp = MOUSE_STAT(sp);

	return (mp->is_on);
}

/*------------------------------------------------------------------------
 * tcap_mouse_invalidate() - invalidate mouse position
 */
int tcap_mouse_invalidate (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		mp->y		= -1;
		mp->x		= -1;
		mp->old_y	= -1;
		mp->old_x	= -1;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_display() - check whether mouse should be displayed
 */
int tcap_mouse_get_display (const TERMINAL *sp)
{
	if (sp == 0)
		return (FALSE);

	return (sp->tcap->bools.md &&
		is_cmd_pres(sp->tcap->strs.mbeg) &&
	    is_cmd_pres(sp->tcap->strs.mend) &&
	    is_cmd_pres(sp->tcap->strs.mstr) );
}

/*------------------------------------------------------------------------
 * tcap_mouse_set_dbl_clk_intvl() - set double click interval
 */
int tcap_mouse_set_dbl_clk_intvl (TERMINAL *sp, int ms)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	if (ms <= 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		mp->dbl_clk_interval = ms;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_event() - get mouse event
 */
int tcap_mouse_get_event (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	return (mp->event);
}

/*------------------------------------------------------------------------
 * tcap_mouse_set_event() - set event types we are interested in
 */
int tcap_mouse_set_event (TERMINAL *sp, int event)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);
		mp->event = event;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_x() - get current mouse x position
 */
int tcap_mouse_get_x (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	return (mp->x);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_y() - get current mouse y position
 */
int tcap_mouse_get_y (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	return (mp->y);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_buttons() - get current mouse button mask
 */
int tcap_mouse_get_buttons (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;
	int i;
	int c;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	c = 0;
	for (i=0; i<NUM_BUTTONS; i++)
	{
		if (mp->button[i])
			c |= (1 << i);
	}

	return (c);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_still_down() - check if any btn is still down (drag detect)
 */
int tcap_mouse_get_still_down (const TERMINAL *sp)
{
	MOUSE_STATUS *mp;
	int i;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	for (i=0; i<NUM_BUTTONS; i++)
	{
		if (mp->button[i])
			return (TRUE);
	}

	return (FALSE);
}

/*------------------------------------------------------------------------
 * tcap_mouse_check_btn() - check for double click for a button
 */
int tcap_mouse_check_btn (TERMINAL *sp, int i)
{
	MOUSE_STATUS *mp;
	TIME_VAL now;
	long secs;
	long usecs;
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);

		term_timeval(&now);

		secs  = (mp->dbl_clk_interval / 1000);
		usecs = (mp->dbl_clk_interval % 1000) * 1000;

		rc = FALSE;
		if ((now.tv_sec - mp->btime[i].tv_sec) < secs)
		{
			rc = TRUE;
		}
		else if ((now.tv_sec  - mp->btime[i].tv_sec ) ==  secs &&
			     (now.tv_usec - mp->btime[i].tv_usec) <= usecs)
		{
			rc = TRUE;
		}

		if (rc)
		{
			if ((mp->x != mp->old_x) ||
				(mp->y != mp->old_y))
			{
				rc = FALSE;
			}
		}

		mp->btime[i].tv_sec		= now.tv_sec;
		mp->btime[i].tv_usec	= now.tv_usec;
		mp->old_x				= mp->x;
		mp->old_y				= mp->y;
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_get_input() - process mouse input string
 */
int tcap_mouse_get_input (TERMINAL *sp, const unsigned char *input)
{
	MOUSE_STATUS *			mp;
	const unsigned char *	str;
	const char *			template;
	int						args[3];
	int						i;
	int						base_offset = 0;
	int						reverse	= 0;
	int						button_change;
	int						button_is_down;
	int						bi;
	int						x;
	int						y;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	/*--------------------------------------------------------------------
	 * check if we have a non-empty mouse-input string defined
	 */
	if (! is_cmd_pres(sp->tcap->strs.mstr))
		return (0);

	if (*sp->tcap->strs.mstr == 0)
		return (0);

	/*--------------------------------------------------------------------
	 *	convert string to args[]
	 *	template will be something like "\e[M%+ %+ %+ "
	 *	args[0] = button info
	 *	args[1] = x (or y)
	 *	args[2] = y (or x)
	 */
	template = sp->tcap->strs.mstr;

	str = input;
	args[0] = args[1] = args[2] = 0;

	for (i=0; *str && *template; template++)
	{
		if (*template == '%')
		{
			template++;
			switch (*template)
			{
			case '+':
				template++;
				args[i] = *str++ - *template;
				i++;
				break;

			case 'd':
				if (isdigit(*str))
				{
					while (isdigit(*str))
					{
						args[i] = args[i] * 10 + *str - '0';
						str++;
					}
					i++;
					break;
				}
				else
				{
					return (0);
				}

			case 'i':
				base_offset = 1;
				break;

			case 'r':
				reverse = 1;
				break;

			case 'c':
				args[i] = *str++;
				i++;
				break;
			}
		}
		else
		{
			if (*str != *template)
				return (0);
			str++;
		}
	}

	/*--------------------------------------------------------------------
	 * check if template used up
	 */
	if (*template)
		return (-1);

	/*--------------------------------------------------------------------
	 *	If (btn-info & 0x80) == 0, then we have input in "xterm" mode.
	 *	If (btn-info & 0x80) != 0, then we have input in "xvt"   mode.
	 *
	 *	1.	xterm mode:
	 *		Button info is encoded as:
	 *
	 *			bits 0-1	(0x03)	button info:
	 *
	 *							0	button 1 pressed
	 *							1	button 2 pressed
	 *							2	button 3 pressed
	 *							3	some button was released (no info on which)
	 *
	 *			bit  2		(0x04)	shift is down
	 *			bit  3		(0x08)	meta (alt) is down
	 *			bit  4		(0x10)	control is down
	 *			bit  5		(0x20)	reserved (for + ' ' encoding)
	 *			bit  6		(0x40)	always 0
	 *			bit  7		(0x80)	always 0
	 *	
	 *	2.	xvt mode:
	 *		Button info is encoded as:
	 *
	 *			bit  0		(0x01)	button 1 state (0 = dn, 1 = up)
	 *			bit  1		(0x02)	button 2 state (0 = dn, 1 = up)
	 *			bit  2		(0x04)	button 3 state (0 = dn, 1 = up)
	 *			bit  3		(0x08)	1 = scroll-wheel up one notch
	 *			bit  4		(0x10)	1 = scroll-wheel dn one notch
	 *			bit  5		(0x20)	reserved (for + ' ' encoding)
	 *			bit  6		(0x40)	1 = no button change, motion only
	 *			bit  7		(0x80)	always 1
	 *
	 *	Note that in either case, only one button's worth of change
	 *	is sent at a time, although the x-y position may change.
	 */
	button_is_down	= tcap_mouse_get_still_down(sp);
	button_change	= -1;
	bi				= args[0];

	TERMINAL_LOCK(sp);
	{
		if (bi & 0x80)
		{
			/*------------------------------------------------------------
			 * "xvt" mode
			 */
			if ((bi & 0x40) == 0)
			{
				if      (bi & (1 << BUTTON_SCRLUP))
				{
					button_change = MOUSE_SCROLLUP;
				}
				else if (bi & (1 << BUTTON_SCRLDN))
				{
					button_change = MOUSE_SCROLLDN;
				}
				else
				{
					for (i=0; i<NUM_BUTTONS; i++)
					{
						if (bi & (1 << i))
						{
							if (! mp->button[i])
							{
								mp->button[i] = BUTTON_DOWN;
								if (tcap_mouse_check_btn(sp, i))
									button_change = term_button_dbl[i];
								else
									button_change = term_button_down[i];
								break;
							}
						}
						else
						{
							if (mp->button[i])
							{
								mp->button[i] = BUTTON_UP;
								button_change = term_button_up[i];
								break;
							}
						}
					}
				}
			}
		}
		else
		{
			/*------------------------------------------------------------
			 * "xterm" mode
			 */
			button_change = -1;
			i = args[0];
			if (i < NUM_BUTTONS)
			{
				if (sp->tcap->ints.nb == 2 && i == BUTTON_MIDDLE)
					i = BUTTON_RIGHT;
				mp->button[i] = BUTTON_DOWN;
				if (tcap_mouse_check_btn(sp, i))
					button_change = term_button_dbl[i];
				else
					button_change = term_button_down[i];
				mp->saved_buttons = i;
			}
			else
			{
				if (mp->saved_buttons < NUM_BUTTONS)
				{
					mp->button[mp->saved_buttons] = BUTTON_UP;
					button_change = term_button_up[mp->saved_buttons];
				}
				else
				{
					for (i=0; i<NUM_BUTTONS; i++)
					{
						if (mp->button[i])
						{
							mp->button[i] = BUTTON_UP;
							button_change = term_button_up[i];
							break;
						}
					}
				}
				mp->saved_buttons = args[0];
			}
		}

		/*----------------------------------------------------------------
		 * adjust x & y positions and offsets
		 */
		x = args[1+reverse] - base_offset;
		y = args[2-reverse] - base_offset;

		/*----------------------------------------------------------------
		 *	At this point, either x-y has changed, a button changed, or both.
		 *	If both, queue the button changed info & return x-y change.
		 *	Else return what happened.
		 */
		if (x != mp->x || y != mp->y)
		{
			if (button_change != -1)
			{
				mp->queued_event = button_change;
				tcap_kbd_enq_ta(sp, KEY_MOUSE, TRUE);
			}

			mp->x = x;
			mp->y = y;
			if (button_is_down)
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
		}
		else
		{
			mp->event = button_change;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (KEY_MOUSE);
}

/*------------------------------------------------------------------------
 * tcap_mouse_deq_event() - dequeue last mouse event
 */
int tcap_mouse_deq_event (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);

		if (mp->queued_event)
		{
			mp->event = mp->queued_event;
			mp->queued_event = 0;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_check_position() - position mouse if position unknown
 */
int tcap_mouse_check_position (TERMINAL *sp)
{
	MOUSE_STATUS *mp;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		mp = MOUSE_STAT(sp);

		if (mp->y == -1 || mp->x == -1)
		{
			mp->y = sp->tcap->ints.maxrows / 2;
			mp->x = sp->tcap->ints.maxcols / 2;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_up() - move mouse up
 */
int tcap_mouse_move_up (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y > 0)
		{
			mp->y--;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_down() - move mouse down
 */
int tcap_mouse_move_down (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y < (sp->tcap->ints.maxrows-1))
		{
			mp->y++;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_left() - move mouse left
 */
int tcap_mouse_move_left (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->x > 0)
		{
			mp->x--;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_right() - move mouse right
 */
int tcap_mouse_move_right (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->x < (sp->tcap->ints.maxcols-1))
		{
			mp->x++;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_ul() - move mouse up & left
 */
int tcap_mouse_move_ul (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y > 0 && mp->x > 0)
		{
			mp->y--;
			mp->x--;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_ur() - move mouse up & right
 */
int tcap_mouse_move_ur (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y > 0 && mp->x < (sp->tcap->ints.maxcols-1))
		{
			mp->y--;
			mp->x++;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_dl() - move mouse down & left
 */
int tcap_mouse_move_dl (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y < (sp->tcap->ints.maxrows-1) && mp->x > 0)
		{
			mp->y++;
			mp->x--;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_move_dr() - move mouse down & right
 */
int tcap_mouse_move_dr (TERMINAL *sp)
{
	MOUSE_STATUS *	mp;
	int				rc;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	TERMINAL_LOCK(sp);
	{
		if (mp->y < (sp->tcap->ints.maxrows-1) &&
	    	mp->x < (sp->tcap->ints.maxcols-1))
		{
			mp->y--;
			mp->x--;
			if (tcap_mouse_get_still_down(sp))
				mp->event = MOUSE_DRAG;
			else
				mp->event = MOUSE_MOVE;
			mp->queued_event = 0;

			rc = 0;
		}
		else
		{
			rc = -1;
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_mouse_set_button() - set button up or down
 */
int tcap_mouse_set_button (TERMINAL *sp, int but, int pos)
{
	MOUSE_STATUS *	mp;

	if (sp == 0)
		return (-1);

	mp = MOUSE_STAT(sp);

	tcap_mouse_check_position(sp);

	if (but < 0 || but > NUM_BUTTONS)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		if (pos == BUTTON_DOWN)
		{
			mp->button[but]  = BUTTON_DOWN;
			mp->queued_event = 0;
			if (tcap_mouse_check_btn(sp, but))
				mp->event = term_button_dbl[but];
			else
				mp->event = term_button_down[but];
		}
		else
		{
			mp->button[but]		= BUTTON_UP;
			mp->queued_event	= 0;
			mp->event			= term_button_up[but];
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}
