/*------------------------------------------------------------------------
 * screen output routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_outc() - output a character
 */
int tcap_outc (TERMINAL *sp, int c)
{
	int ch;
	int ic = c;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * Check if negative (from calling with a high-order char as a
	 * signed char) & force modulo 256.
	 */
	ic %= 256;
	if (ic < 0)
		ic += 256;

	/*--------------------------------------------------------------------
	 * output the char
	 */
	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * translate char according to appropriate table
		 */
		if (sp->scrn->alt_chars_on)
		{
			/*------------------------------------------------------------
			 * translate using the alt-chars table
			 */
			ch = sp->tcap->alt_tbl[ic];

			/*------------------------------------------------------------
			 * If the alt-font number for this char is different from the
			 * current alt-font number, re-output the current attributes,
			 * but with them suitable for the new char, not the old.
			 */
			if (sp->tcap->alt_font_tbl[ic] != sp->scrn->curr_alt_font)
				tcap_outattr(sp, sp->scrn->curr_attr, ic);
		}
		else
		{
			/*------------------------------------------------------------
			 * translate using the reg-chars table
			 */
			ch = sp->tcap->reg_tbl[ic];

			/*------------------------------------------------------------
			 * If the reg-font number for this char is different from the
			 * current reg-font number, re-output the current attributes,
			 * but with them suitable for the new char, not the old.
			 */
			if (sp->tcap->reg_font_tbl[ic] != sp->scrn->curr_reg_font)
				tcap_outattr(sp, sp->scrn->curr_attr, ic);
		}

		/*----------------------------------------------------------------
		 * output the char
		 */
		tcap_outch(sp, ch, TRUE);

		/*----------------------------------------------------------------
		 * check current position
		 */
		if (! sp->scrn->alt_chars_on && c < ' ')
		{
			/*------------------------------------------------------------
			 * process control char
			 */
			switch (ic)
			{
			case '\b':
				if (sp->scrn->curr_col)
					sp->scrn->curr_col--;
				break;

			case '\r':
				sp->scrn->curr_col = 0;
				break;

			case '\n':
				sp->scrn->curr_col = 0;
				if (++sp->scrn->curr_row >= sp->tcap->ints.maxrows)
					sp->scrn->curr_row = sp->tcap->ints.maxrows-1;
				break;
			}
		}
		else
		{
			/*------------------------------------------------------------
			 * process past EOL condition
			 */
			if (++sp->scrn->curr_col >= sp->tcap->ints.maxcols)
			{
				if (sp->tcap->bools.am)
				{
					/* wrap to next line */

					sp->scrn->curr_col = 0;
					if (++sp->scrn->curr_row >= sp->tcap->ints.maxrows)
						sp->scrn->curr_row = sp->tcap->ints.maxrows-1;
				}
				else
				{
					/* just stay where we were */

					sp->scrn->curr_col = sp->tcap->ints.maxcols-1;
				}
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_outs() - output a string
 */
int tcap_outs (TERMINAL *sp, const char *s)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		while (*s)
			tcap_outc(sp, *s++);
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_mono() - set mono attributes
 */
static void tcap_set_mono (TERMINAL *sp, int on, attr_t a, attr_t t, int force)
{
	if (on)
	{
		if ((a & A_REVERSE)   && ((t & A_REVERSE)   || force))
			tcap_outcmd(sp, S_STS, sp->tcap->strs.sts);

		if ((a & A_UNDERLINE) && ((t & A_UNDERLINE) || force))
			tcap_outcmd(sp, S_ULS, sp->tcap->strs.uls);

		if ((a & A_BOLD)      && ((t & A_BOLD)      || force))
			tcap_outcmd(sp, S_BDS, sp->tcap->strs.bds);

		if ((a & A_BLINK)     && ((t & A_BLINK)     || force))
			tcap_outcmd(sp, S_BLS, sp->tcap->strs.bls);

		if ((a & A_DIM)       && ((t & A_DIM)       || force))
			tcap_outcmd(sp, S_DMS, sp->tcap->strs.dms);

		if ((a & A_INVIS)     && ((t & A_INVIS)     || force))
			tcap_outcmd(sp, S_IVS, sp->tcap->strs.ivs);

		if ((a & A_PROTECT)   && ((t & A_PROTECT)   || force))
			tcap_outcmd(sp, S_PRS, sp->tcap->strs.prs);
	}
	else
	{
		if ((a & A_REVERSE)   && (t & A_REVERSE))
			tcap_outcmd(sp, S_STE, sp->tcap->strs.ste);

		if ((a & A_UNDERLINE) && (t & A_UNDERLINE))
			tcap_outcmd(sp, S_ULE, sp->tcap->strs.ule);

		if ((a & A_BOLD)      && (t & A_BOLD))
			tcap_outcmd(sp, S_BDE, sp->tcap->strs.bde);

		if ((a & A_BLINK)     && (t & A_BLINK))
			tcap_outcmd(sp, S_BLE, sp->tcap->strs.ble);

		if ((a & A_DIM)       && (t & A_DIM))
			tcap_outcmd(sp, S_DME, sp->tcap->strs.dme);

		if ((a & A_INVIS)     && (t & A_INVIS))
			tcap_outcmd(sp, S_IVE, sp->tcap->strs.ive);

		if ((a & A_PROTECT)   && (t & A_PROTECT))
			tcap_outcmd(sp, S_PRE, sp->tcap->strs.pre);
	}
}

/*------------------------------------------------------------------------
 * tcap_set_color() - set color attributes
 */
static void tcap_set_color (TERMINAL *sp, attr_t a)
{
	TCAP_DATA *tp = sp->tcap;

	if (sp->scrn->color_flag)
	{
		int	fg;
		int	bg;

		/*----------------------------------------------------------------
		 * get fg & bg values
		 */
		fg = A_FG_CLRNUM(a) % tp->ints.max_fg;
		bg = A_BG_CLRNUM(a) % tp->ints.max_bg;

		/*----------------------------------------------------------------
		 * output appropriate cmds
		 */
		if (is_cmd_pres(tp->parms.csr)    || is_cmd_pres(tp->parms.csm)    ||
			is_cmd_pres(tp->parms.csr_ti) || is_cmd_pres(tp->parms.csm_ti) ||
			is_cmd_pres(tp->parms.csr_tc) || is_cmd_pres(tp->parms.csm_tc) )
		{
			tcap_outparm(sp, S_CS, bg, fg);
		}
		else
		{
			tcap_outparm(sp, S_FG, fg, 0);
			tcap_outparm(sp, S_BG, bg, 0);
		}
	}
}

/*------------------------------------------------------------------------
 * tcap_outattr() - set screen attributes
 */
int tcap_outattr (TERMINAL *sp, attr_t a, int c)
{
	attr_t	o;
	attr_t	t;
	int		f;
	int		char_set_changed = FALSE;
	int		old_is_color;
	int		new_is_color;
	int		need_init = FALSE;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * if attribute is UNSET, set to "default"
	 */
	if (a == A_UNSET)
	{
		attr_t r;
		attr_t s;

		tcap_get_default_attrs(sp, &r, &s, TRUE, TRUE);

		a = r;
	}

	/*--------------------------------------------------------------------
	 * if both color & mono attrs are specified, bail
	 */
	if ((a & A_COLOR) != 0 && (a & A_MONO) != 0)
	{
		/*----------------------------------------------------------------
		 * before we bail, convert BOLD/BLINK to FG-BRIGHT/BG-BRIGHT
		 */
		int am = a & (A_BOLD | A_BLINK);
		int at = a & (A_MONO);

		if (am != at)
			return (-1);

		if (am & A_BOLD)
		{
			a |= COLOR_FG_BRIGHT;
			a &= ~A_BOLD;
		}

		if (am & A_BLINK)
		{
			a |= COLOR_BG_BRIGHT;
			a &= ~A_BLINK;
		}
	}

	/*--------------------------------------------------------------------
	 * if current attribute is UNSET, treat as 0
	 */
	o = sp->scrn->curr_attr;
	if (c == 0)
	{
		o = A_UNSET;
		c = ' ';
	}

	if (o == A_UNSET)
	{
		need_init = TRUE;
		o = 0;
	}

	/*--------------------------------------------------------------------
	 * determine whether old & new attrs are in color
	 *
	 * Note that this test fails if an attr is color of (black/black),
	 * but this doesn't happen very often.
	 */
	old_is_color = ( (o & A_COLOR) != 0 );
	new_is_color = ( (a & A_COLOR) != 0 );

	/*--------------------------------------------------------------------
	 * now do the attributes
	 */
	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * check if we are switching from color->mono or mono->color
		 */
		if (old_is_color != new_is_color && sp->scrn->color_flag)
		{
			if (old_is_color)
			{
				attr_t r;
				attr_t s;

				/*--------------------------------------------------------
				 * switching color->mono
				 *
				 * Switch color to "std" mono-type color (e.g. white/black).
				 */
				tcap_get_default_attrs(sp, &r, &s, TRUE, TRUE);

				tcap_set_color(sp, r);

				/*--------------------------------------------------------
				 * set reverse on/off
				 * (in hopes this will set the emulator to mono)
				 */
				tcap_outcmd(sp, S_STS, sp->tcap->strs.sts);
				tcap_outcmd(sp, S_STE, sp->tcap->strs.ste);
			}
			else
			{
				/*--------------------------------------------------------
				 * switching mono->color
				 *
				 * Turn off all attributes so we have a "normal" screen.
				 */
				tcap_set_mono(sp, FALSE, o, o, TRUE);
			}
		}

		/*----------------------------------------------------------------
		 * store new attribute as current
		 */
		sp->scrn->curr_attr = a;

		/*----------------------------------------------------------------
		 * get diffs between old & new attrs
		 */
		t = o ^ a;

		/*----------------------------------------------------------------
		 * First do any offs if in mono mode
		 */
		if (! old_is_color)
		{
			tcap_set_mono(sp, FALSE, o, t, char_set_changed);
		}

		/*----------------------------------------------------------------
		 * Do any alt char switching first, since brain-dead
		 * SCO XENIX ansi driver resets color on font change.
		 */
		{
			/*------------------------------------------------------------
			 * first do the offs
			 */
			if (o & A_ALTCHARSET)
			{
				f = (a & A_ALTCHARSET) ? sp->tcap->alt_font_tbl[c] : 0;

				if (sp->scrn->alt_chars_on && sp->scrn->curr_alt_font != f)
				{
					if (sp->scrn->curr_alt_font != 0)
					{
						tcap_outcmd(sp,
							tcap_alt_font_cmds[sp->scrn->curr_alt_font].off,
							sp->tcap->strs.ane[sp->scrn->curr_alt_font-1]);
					}
					sp->scrn->curr_alt_font = 0;
					char_set_changed = TRUE;
					sp->scrn->alt_chars_on = FALSE;
				}
			}
			else
			{
				f = (! (a & A_ALTCHARSET)) ? sp->tcap->reg_font_tbl[c] : 0;

				if (! sp->scrn->alt_chars_on && sp->scrn->curr_reg_font != f)
				{
					if (sp->scrn->curr_reg_font != 0)
					{
						tcap_outcmd(sp,
							tcap_reg_font_cmds[sp->scrn->curr_reg_font].off,
							sp->tcap->strs.rne[sp->scrn->curr_reg_font-1]);
					}
					sp->scrn->curr_reg_font = 0;
					char_set_changed = TRUE;
				}
			}

			/*------------------------------------------------------------
			 * now do the ons
			 */
			if (a & A_ALTCHARSET)
			{
				f = sp->tcap->alt_font_tbl[c];

				if (f && sp->scrn->curr_alt_font != f)
				{
					if (f != 0)
					{
						tcap_outcmd(sp, tcap_alt_font_cmds[f].on,
							sp->tcap->strs.ans[f-1]);
					}
					sp->scrn->curr_alt_font = f;
					char_set_changed = TRUE;
				}
				sp->scrn->alt_chars_on = TRUE;
			}
			else
			{
				f = sp->tcap->reg_font_tbl[c];
	
				if (f && sp->scrn->curr_reg_font != f)
				{
					if (f != 0)
					{
						tcap_outcmd(sp, tcap_reg_font_cmds[f].on,
							sp->tcap->strs.rns[f-1]);
					}
					sp->scrn->curr_reg_font = f;
					char_set_changed = TRUE;
				}
			}
		}

		/*----------------------------------------------------------------
		 * now do any color processing if we have color
		 */
		if (new_is_color)
		{
			/*------------------------------------------------------------
			 * output new color if changed, or
			 * if we are switching to or from alt-char set, or
			 * if fonts are different.
			 */
			if (t & A_COLOR || char_set_changed)
			{
				tcap_set_color(sp, a);
			}
		}

		/*----------------------------------------------------------------
		 * Now do any ons if in mono mode
		 */
		if (! new_is_color)
		{
			tcap_set_mono(sp, TRUE, a, t, char_set_changed);
		}

		/*----------------------------------------------------------------
		 * Now if any attribute changes happened & we are in alt-char mode,
		 * re-issue the alt cmd (since brain-dead eterm (linux) loses it).
		 */
		if (t != 0 && sp->scrn->alt_chars_on == TRUE)
		{
			if (sp->scrn->curr_alt_font != 0)
			{
				tcap_outcmd(sp, tcap_alt_font_cmds[sp->scrn->curr_alt_font].on,
					sp->tcap->strs.ans[sp->scrn->curr_alt_font-1]);
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_outpos() - set screen position
 */
int tcap_outpos (TERMINAL *sp, int row, int col)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		if (row != sp->scrn->curr_row || col != sp->scrn->curr_col)
		{
			if (row < 0 || row >= sp->tcap->ints.maxrows ||
				col < 0 || col >= sp->tcap->ints.maxcols)
			{
				rc = -1;
			}
			else
			{
				tcap_outparm(sp, S_CM, row, col);

				sp->scrn->curr_row = row;
				sp->scrn->curr_col = col;

				rc = 0;
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_eval_parm() - evaluate a parameterized expression
 */
char * tcap_eval_parm (const TERMINAL *sp, char *buf, SCRN_CMD cmd,
	int p1, int p2)
{
	TCAP_DATA *	tp	= sp->tcap;
	char *		s;
	int			t;

	switch (cmd)
	{
	/*--------------------------------------------------------------------
	 * move cursor
	 */
	case S_CM:
		t = tp->pdata.cm.type;
		s = tp->pdata.cm.str;
		break;

	/*--------------------------------------------------------------------
	 * color string
	 */
	case S_CS:
		t = tp->pdata.cs.type;
		s = tp->pdata.cs.str;
		if (tp->pdata.bg.data)
		{
			p1 = tp->ints.color_bg_map[p1];
			p2 = tp->ints.color_fg_map[p2];
		}
		break;

	/*--------------------------------------------------------------------
	 * foreground color
	 */
	case S_FG:
		t = tp->pdata.fg.type;
		s = tp->pdata.fg.str;
		if (tp->pdata.bg.data)
		{
			p1 = tp->ints.color_fg_map[p1];
		}
		break;

	/*--------------------------------------------------------------------
	 * background color
	 */
	case S_BG:
		t = tp->pdata.bg.type;
		s = tp->pdata.bg.str;
		if (tp->pdata.bg.data)
		{
			p1 = tp->ints.color_bg_map[p1];
		}
		break;

	/*--------------------------------------------------------------------
	 * invalid cmd
	 */
	default:
		t = -1;
		s = 0;
		break;
	}

	if (! is_cmd_pres(s) || *s == 0)
	{
		return ("");
	}

	return tcap_db_calc(t, buf, s, p1, p2);
}

/*------------------------------------------------------------------------
 * tcap_get_valid_acs() - check if a char is a valid alt-char
 */
int tcap_get_valid_acs (const TERMINAL *sp, int c)
{
	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (sp == 0)
		return (FALSE);

	/*--------------------------------------------------------------------
	 * get font number for this char
	 */
	c = sp->tcap->alt_font_tbl[c & 0xff];

	/*--------------------------------------------------------------------
	 * if font is 0, then nothing needs to be sent, but it is valid
	 */
	if (c == 0)
		return (TRUE);

	/*--------------------------------------------------------------------
	 * if user set no-alt-chars, then it is not valid
	 */
	if (! sp->tcap->bools.ala)
		return (FALSE);

	/*--------------------------------------------------------------------
	 * if both on & off strings exist for this font, then it is valid
	 */
	if (is_cmd_pres(sp->tcap->strs.ans[c-1]) &&
	    is_cmd_pres(sp->tcap->strs.ane[c-1]))
	{
		return (TRUE);
	}

	return (FALSE);
}

/*------------------------------------------------------------------------
 * tcap_set_valid_acs() - set if a char is a valid alt-char
 */
int tcap_set_valid_acs (TERMINAL *sp, int bf)
{
	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * store new flag
	 */
	TERMINAL_LOCK(sp);
	{
		sp->tcap->bools.ala = bf;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_curs_set() - set cursor mode
 */
int tcap_curs_set (TERMINAL *sp, int vis)
{
	int old_vis;
	int rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * if new-visibility == old-visibility, then we're done
	 */
	old_vis = sp->scrn->visibility;

	if (vis == old_vis)
	{
		return (old_vis);
	}

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * set new visibility mode
		 */
		rc = 0;
		switch (vis)
		{
		case VISIBILITY_OFF:
			if (! is_cmd_pres(sp->tcap->strs.vi))
				rc = -1;
			tcap_outcmd(sp, S_VI, sp->tcap->strs.vi);
			break;

		case VISIBILITY_ON:
			if (! is_cmd_pres(sp->tcap->strs.ve))
				rc = -1;
			tcap_outcmd(sp, S_VE, sp->tcap->strs.ve);
			break;

		case VISIBILITY_VERY_ON:
			if (! is_cmd_pres(sp->tcap->strs.vs))
				rc = -1;
			tcap_outcmd(sp, S_VS, sp->tcap->strs.vs);
			break;

		default:
			rc = -1;
			break;
		}

		/*----------------------------------------------------------------
		 * cache current mode if successful in changing it
		 */
		if (rc == 0)
			sp->scrn->visibility = vis;
	}
	TERMINAL_UNLOCK(sp);

	return (old_vis);
}

/*------------------------------------------------------------------------
 * tcap_delay_output() - output delay chars for a calculated delay
 */
int tcap_delay_output (TERMINAL *sp, int ms, int (*outrtn)(TERMINAL *sp, int c))
{
	int (*rtn)(TERMINAL *sp, int c) = (outrtn == 0 ? tcap_outchar : outrtn);
	int n;
	int i;
	int cps;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (sp == 0)
		return (-1);

	if (ms < 0)
		return (-1);

	if (ms == 0)
		return (0);

	/*--------------------------------------------------------------------
	 * get chars-per-sec
	 *
	 * Note that the baud-rate is bits-per-sec, which includes a start
	 * and stop bit, so 10 bits = 1 char.
	 */
	cps = tcap_get_baud_rate(sp) / 10;

	/*--------------------------------------------------------------------
	 * now calculate number of pad-chars to output
	 */
	n = (ms * cps) / 1000;
	if (n == 0)
		return (0);

	/*--------------------------------------------------------------------
	 * now output them
	 */
	TERMINAL_LOCK(sp);
	{
		for (i=0; i<n; i++)
		{
			(*rtn)(sp, 0);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_out_debug() - start/stop screen debugging
 */
int tcap_out_debug (TERMINAL *sp, const char *path, int flag, int text)
{
	int rc;

	if (sp == 0)
		return (FALSE);

	TERMINAL_LOCK(sp);
	{
#if V_WINDOWS
		if (sp->tcap->db_type == TCAP_DB_WINDOWS)
			text = TRUE;
#endif
		sp->scrn->debug_scrn_text	= text;

		if (path == 0)
		{
			if (*sp->scrn->debug_scrn_path == 0)
				path = text ? TCAP_SCRN_LOG_TXT : TCAP_SCRN_LOG_BIN;
			else
				path = sp->scrn->debug_scrn_path;
		}
		strcpy(sp->scrn->debug_scrn_path, path);

		if (flag)
		{
			if (sp->scrn->debug_scrn_fp == 0)
			{
				sp->scrn->debug_scrn_mode	= TRUE;
				sp->scrn->debug_scrn_fp		= fopen(path, "w");

				if (sp->scrn->mode)
					tcap_set_window_title(sp, 0);
			}
		}
		else
		{
			if (sp->scrn->debug_scrn_fp != 0)
			{
				fclose(sp->scrn->debug_scrn_fp);
				sp->scrn->debug_scrn_fp = 0;

				if (sp->scrn->mode)
					tcap_set_window_title(sp, 0);
			}
		}

		rc = (sp->scrn->debug_scrn_fp != 0);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_get_curr_row() - get current row number
 */
int tcap_get_curr_row (const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return (sp->scrn->curr_row);
}

/*------------------------------------------------------------------------
 * tcap_get_curr_col() - get current col number
 */
int tcap_get_curr_col (const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return (sp->scrn->curr_col);
}

/*------------------------------------------------------------------------
 * tcap_get_curr_attr() - get current attributes
 */
attr_t tcap_get_curr_attr (const TERMINAL *sp)
{
	if (sp == 0)
		return (A_UNSET);

	return (sp->scrn->curr_attr);
}

/*------------------------------------------------------------------------
 * tcap_set_window_title() - set window title
 *
 * Note: we always add any debug log info
 */
int tcap_set_window_title (TERMINAL *sp, const char *s)
{
	if (sp == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * We only send the title if we have both a TTS & TTE string.
	 */
	if (! is_cmd_pres(sp->tcap->strs.tts) || ! is_cmd_pres(sp->tcap->strs.tte))
		return (-1);

	/*--------------------------------------------------------------------
	 * OK - let's do it
	 */
	TERMINAL_LOCK(sp);
	{
		char	new_title[128];

		/*----------------------------------------------------------------
		 * A NULL string implies we output the cached string.
		 */
		if (s == 0)
		{
			/*------------------------------------------------------------
			 * use cached string
			 */
			s = sp->scrn->win_title;
		}
		else
		{
			/*------------------------------------------------------------
			 * cache real string
			 */
			strncpy(sp->scrn->win_title, s, sizeof(sp->scrn->win_title));
		}

		/*----------------------------------------------------------------
		 * Check if we are appending anything
		 */
		if (*s != 0)
		{
			if (sp->scrn->debug_keys_fp != 0 ||
				sp->scrn->debug_scrn_fp != 0)
			{
				strcpy(new_title, s);
				s = new_title;

				strcat(new_title, " [logs:");

				if (sp->scrn->debug_keys_fp != 0)
				{
					strcat(new_title, " ");
					strcat(new_title, sp->scrn->debug_keys_text ?
						TCAP_KEYS_LOG_TXT : TCAP_KEYS_LOG_BIN);
				}

				if (sp->scrn->debug_scrn_fp != 0)
				{
					strcat(new_title, " ");
					strcat(new_title, sp->scrn->debug_scrn_text ?
						TCAP_SCRN_LOG_TXT : TCAP_SCRN_LOG_BIN);
				}

				strcat(new_title, "]");
			}
		}

		/*----------------------------------------------------------------
		 * send the title string
		 */
#if V_WINDOWS
		if (sp->tcap->db_type == TCAP_DB_WINDOWS)
		{
			tcap_outcmd(sp, S_TTS, s);
			tcap_outcmd(sp, S_TTE, 0);
		}
		else
#endif
		{
			tcap_outcmd(sp, S_TTS, sp->tcap->strs.tts);
			tcap_outs(sp, s);
			tcap_outcmd(sp, S_TTE, sp->tcap->strs.tte);
			tcap_outflush(sp);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_clear_screen() - clear the screen using current attributes
 */
int tcap_clear_screen (TERMINAL *sp)
{
	int i;
	int j;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		/*----------------------------------------------------------------
		 * use any easy methods if possible
		 */
		if (is_cmd_pres(sp->tcap->strs.cl))
		{
			/*------------------------------------------------------------
			 * clear-screen cmd is present
			 */
			tcap_outcmd(sp, S_CL, sp->tcap->strs.cl);
		}
		else if (is_cmd_pres(sp->tcap->strs.cd))
		{
			/*------------------------------------------------------------
			 * clear-to-end-of-screen cmd is present
			 */
			tcap_outpos(sp, 0, 0);
			tcap_outcmd(sp, S_CD, sp->tcap->strs.cd);
		}
		else if (is_cmd_pres(sp->tcap->strs.ce))
		{
			/*------------------------------------------------------------
			 * clear-to-end-of-line cmd is present
			 */
			for (i=0; i<sp->tcap->ints.maxrows; i++)
			{
				tcap_outpos(sp, i, 0);
				tcap_outcmd(sp, S_CE, sp->tcap->strs.ce);
			}
		}
		else
		{
			/*------------------------------------------------------------
			 * just blank the whole friggin screen
			 */
			for (i=0; i<sp->tcap->ints.maxrows; i++)
			{
				tcap_outpos(sp, i, 0);
				for (j=0; j<sp->tcap->ints.maxcols; j++)
					tcap_outc(sp, ' ');
			}
		}
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_need_ins_final() - check whether final write call needed for
 * last col of last line
 */
int tcap_need_ins_final (const TERMINAL *sp)
{
	if (sp == 0)
		return (FALSE);

	return (sp->tcap->bools.ifc);
}

/*------------------------------------------------------------------------
 * tcap_write_final_char() - output final char via insert logic
 */
int tcap_write_final_char (TERMINAL *sp, int cpc, attr_t cpa,
	int chc, attr_t cha)
{
	if (sp == 0)
		return (-1);

	if (! is_cmd_pres(sp->tcap->strs.ic))
		return (-1);

	TERMINAL_LOCK(sp);
	{
		tcap_outpos(sp, sp->tcap->ints.maxrows-1, sp->tcap->ints.maxcols-2);
		tcap_outattr(sp, cha, chc);
		tcap_outc(sp, chc);
		tcap_outpos(sp, sp->tcap->ints.maxrows-1, sp->tcap->ints.maxcols-2);
		tcap_outcmd(sp, S_IC, sp->tcap->strs.ic);
		tcap_outattr(sp, cpa, cpc);
		tcap_outc(sp, cpc);
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}
