/*------------------------------------------------------------------------
 *	This is a portable method of getting the name of a signal given
 *	the number.  We do a linear search on the table to find the right
 *	signal name and return the requested long or short name.
 *
 *	This is not a particurly efficient routine, but the alternative
 *	is to try to juggle signal number indices, and on a lot of OSs
 *	this is *madness*.  Besides, this is probably only called once in
 *	an error handler anyway.
 */
#include "termcommon.h"

/*------------------------------------------------------------------------
 * signal info struct
 */
struct signm
{
	int				si_signo;		/* signal number	*/
	const char *	si_signame;		/* SIG name			*/
	const char *	si_sigdesc;		/* SIG description	*/
};
typedef struct signm SIGNM;

static const SIGNM term_sigtbl[] =
{
#ifdef SIGHUP
	{ SIGHUP,		"SIGHUP",		"Hangup"					},
#endif

#ifdef SIGINT
	{ SIGINT,		"SIGINT",		"Interrupt"					},
#endif

#ifdef SIGQUIT
	{ SIGQUIT,		"SIGQUIT",		"Quit"						},
#endif

#ifdef SIGILL
	{ SIGILL,		"SIGILL",		"Illegal instruction"		},
#endif

#ifdef SIGTRAP
	{ SIGTRAP,		"SIGTRAP",		"Trace trap"				},
#endif

#ifdef SIGABRT
	{ SIGABRT,		"SIGABRT",		"Abort"						},
#endif

#ifdef SIGIOT
	{ SIGIOT,		"SIGIOT",		"IOT trap"					},
#endif

#ifdef SIGEMT
	{ SIGEMT,		"SIGEMT",		"EMT trap"					},
#endif

#ifdef SIGFPE
	{ SIGFPE,		"SIGFPE",		"Floating point exception"	},
#endif

#ifdef SIGKILL
	{ SIGKILL,		"SIGKILL",		"Kill with prejudice"		},
#endif

#ifdef SIGBUS
	{ SIGBUS,		"SIGBUS",		"Bus error"					},
#endif

#ifdef SIGSEGV
	{ SIGSEGV,		"SIGSEGV",		"Segmentation violation"	},
#endif

#ifdef SIGSYS
	{ SIGSYS,		"SIGSYS",		"Bad system call"			},
#endif

#ifdef SIGPIPE
	{ SIGPIPE,		"SIGPIPE",		"Write to one-ended pipe"	},
#endif

#ifdef SIGALRM
	{ SIGALRM,		"SIGALRM",		"Alarm call"				},
#endif

#ifdef SIGTERM
	{ SIGTERM,		"SIGTERM",		"Software termination"		},
#endif

#ifdef SIGUSR1
	{ SIGUSR1,		"SIGUSR1",		"User signal 1"				},
#endif

#ifdef SIGUSR2
	{ SIGUSR2,		"SIGUSR2",		"User signal 2"				},
#endif

#ifdef SIGCLD
	{ SIGCLD,		"SIGCLD",		"Child death"				},
#endif

#ifdef SIGCHLD
	{ SIGCHLD,		"SIGCHLD",		"Child death"				},
#endif

#ifdef SIGPWR
	{ SIGPWR,		"SIGPWR",		"Power fail"				},
#endif

#ifdef SIGWINCH
	{ SIGWINCH,		"SIGWINCH",		"Window change"				},
#endif

#ifdef SIGURG
	{ SIGURG,		"SIGURG",		"Urgent I/O"				},
#endif

#ifdef SIGIO
	{ SIGIO,		"SIGIO",		"I/O available"				},
#endif

#ifdef SIGPOLL
	{ SIGPOLL,		"SIGPOLL",		"Pollable event occurred"	},
#endif

#ifdef SIGSTOP
	{ SIGSTOP,		"SIGSTOP",		"Software stop"				},
#endif

#ifdef SIGTSTP
	{ SIGTSTP,		"SIGTSTP",		"Terminal stop"				},
#endif

#ifdef SIGCONT
	{ SIGCONT,		"SIGCONT",		"Continue after stop"		},
#endif

#ifdef SIGTTIN
	{ SIGTTIN,		"SIGTTIN",		"bg tty read attempted"		},
#endif

#ifdef SIGTTOU
	{ SIGTTOU,		"SIGTTOU",		"bg tty write attempted"	},
#endif

#ifdef SIGVTALRM
	{ SIGVTALRM,	"SIGVTALRM",	"virtual timer expired"		},
#endif

#ifdef SIGPROF
	{ SIGPROF,		"SIGPROF",		"Profiling timer expired"	},
#endif

#ifdef SIGXCPU
	{ SIGXCPU,		"SIGXCPU",		"CPU limit exceeded"		},
#endif

#ifdef SIGXFSZ
	{ SIGXFSZ,		"SIGXFSZ",		"File size limit exceeded"	},
#endif

#ifdef SIGLOST
	{ SIGLOST,		"SIGLOST",		"Resource lost"				},
#endif

	{ 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * term_sig_info() - get info for a given signal number
 */
void term_sig_info (int sig, char *namebuf, char *descbuf)
{
	const SIGNM *sn;

	for (sn = term_sigtbl; sn->si_signo >= 0; sn++)
	{
		if (sn->si_signo == sig)
		{
			strcpy(namebuf, sn->si_signame);
			strcpy(descbuf, sn->si_sigdesc);
			return;
		}
	}

	sprintf(namebuf, "<signal %d>", sig);
	sprintf(descbuf, "<unknown>");
}

/*------------------------------------------------------------------------
 * term_sig_name() - get name for a given signal number
 */
char * term_sig_name (int sig, char *buf)
{
	const SIGNM *sn;

	for (sn = term_sigtbl; sn->si_signo >= 0; sn++)
	{
		if (sn->si_signo == sig)
		{
			strcpy(buf, sn->si_signame);
			return (buf);
		}
	}

	sprintf(buf, "<signal %d>", sig);
	return (buf);
}

/*------------------------------------------------------------------------
 * term_sig-desc() - get description for a given signal number
 */
char * term_sig_desc (int sig, char *buf)
{
	const SIGNM *sn;

	for (sn = term_sigtbl; sn->si_signo >= 0; sn++)
	{
		if (sn->si_signo == sig)
		{
			strcpy(buf, sn->si_sigdesc);
			return (buf);
		}
	}

	sprintf(buf, "<signal %d>", sig);
	return (buf);
}
