/*------------------------------------------------------------------------
 * libxvt header
 *
 *	The libxvt library provides a terminal-emulator running in
 *	an X11 window, with a pseudo-tty providing input & output to
 *	that window.
 *
 *	This terminal-emulator is designed to emulate the xterm terminal-
 *	emulator, and uses the same escape sequences.  It is, however,
 *	a "slimmed" down version, and doesn't provide all the functionality
 *	of xterm, such as scroll bars, multi-byte support, background
 *	pixmaps, etc.
 *
 *	This window is intended to provide an IBM-PC type of display for
 *	a character application, with support for full-color and mouse.
 *
 *------------------------------------------------------------------------
 *
 *	This library can be used on one of two ways:
 *
 *	1.	Run in "task" mode.
 *
 *		In "task" mode, the library runs a specified cmd as a
 *		sub-task and returns when the sub-task dies (with all display
 *		structs destroyed).
 *
 *	2.	Run in "here" mode.
 *
 *		In "here" mode, all initial setup is done, and then either a
 *		sub-thread is run to process all events, or the caller is given
 *		a pointer to an event-routine which must be called frequently
 *		to "activate" the event-routine.
 *
 *------------------------------------------------------------------------
 *
 *	Notes
 *
 *	1.	In either case of running in "task" mode or in "detached-here" mode
 *		("here" mode with detach set TRUE), the streams stdin/stdout/stderr
 *		are set to point to the pseudo-tty created.
 *
 *		This means that, if running in "detached-here" mode, your program has
 *		NO access to the original stdin/stdout/stderr your program was
 *		started with, unless you copied it somewhere.
 *
 *		If you are running in "task" mode, only the child's streams are
 *		changed.
 *
 *	2.	When running in "detached-here" mode, the calling program is completely
 *		detached from the original controlling terminal.  Thus, there is
 *		no "/dev/tty" available to output to, unless you copied the original
 *		/dev/tty prior to the call.
 *
 *	3.	When running as a thread, do NOT attempt to read from or write to
 *		the pseudo-tty (i.e. to/from stdin/stdout/stderr) when in a
 *		signal routine.  Some Unixes (e.g. Linux) cause all threads to
 *		stop when an interrupt occurs, and thus your program will "hang"
 *		with no way to stop it except via the "kill" command.
 *
 *	4.	If any error occurs, the following values are returned in the
 *		XVT_DATA struct (except in the case of a NULL XVT_DATA pointer):
 *
 *			error_num		xvt error code
 *			error_quit		TRUE if it is not possible to continue
 *			error_buf		displayable error message
 *
 *	5.	Users of the "ecurses" library can automate the calling of the
 *		event-routine when running in "here" mode with the call:
 *
 *			set_event_rtn(xvt_data_get_event_rtn(xd),
 *				xvt_data_get_event_data(xd));
 *
 *		This will cause the event-routine to be automatically called
 *		whenever a "keyboard" query is done, anything is written to
 *		the "screen", and while the caller is "shelled-out" to a sub-task.
 *
 *------------------------------------------------------------------------
 *
 *	Example of using this library in conjunction with the ecurses library:
 */

#if 0	/* sample program */
/*------------------------------------------------------------------------
 *	sample program using ecurses/xvt
 */
#include <stdlib.h>
#include "curses.h"
#include "libxvt.h"

int main (int argc, char **argv)
{
	XVT_DATA *	xd;
	const char *string		= "Hello world!";
	int			use_screen	= FALSE;
	int			rc;
	int			c;

	/*--------------------------------------------------------------------
	 *	Initialize xvt data struct prior to processing options.
	 */
	xd = xvt_data_alloc();					/* allocate data struct			*/
	xvt_data_set_win_label(xd, argv[0]);	/* set pgm name as window title	*/

	/*--------------------------------------------------------------------
	 *	Extract any toolkit options on cmd line.
	 *	This must be done before any calls to getopt().
	 */
	xvt_tk_args(xd, argc, argv);

	argc = xvt_data_get_argc(xd);
	argv = xvt_data_get_argv(xd);

	/*--------------------------------------------------------------------
	 *	Do any option/arg processing.
	 */
	while ((c = getopt(argc, argv, "s")) != EOF)
	{
		switch (c)
		{
		case 's':				/*	use screen						*/
			use_screen = xvt_available();
			break;

		case '?':				/*	error: msg already displayed	*/
			return (EXIT_FAILURE);
		}
	}

	if (optind < argc)
		string = argv[optind++];

	/*--------------------------------------------------------------------
	 *	Initialize xvt processing if requested.
	 */
	if (use_screen)
	{
		rc = xvt_here_run(xd, TRUE);
		if (rc != XVT_OK)
		{
			/*------------------------------------------------------------
			 *	If an error occured, check if possible to continue.
			 */
			if (xvt_data_get_error_quit(xd))
			{
				/*--------------------------------------------------------
				 *	Note that you cannot display an error msg
				 *	since you have no stdio streams to write to,
				 *	which is why you can't continue.
				 */
				return (EXIT_FAILURE);
			}
		}
		else
		{
			/*------------------------------------------------------------
			 *	This is only needed if you are using any advanced
			 *	features of the ecurses library.
			 */
			newscreen_defs(xvt_data_get_term_type(xd),
				xvt_data_get_term_defs(xd),
				xvt_data_get_fd_inp(xd),
				xvt_data_get_fd_out(xd));
		}
	}

	/*--------------------------------------------------------------------
	 *	Initialize curses & setup any event processing needed.
	 *	Note the set_event_rtn() call is a no-op if args are zero.
	 */
	initscr();
	set_event_rtn(xvt_data_get_event_rtn(xd), xvt_data_get_event_data(xd));

	/*--------------------------------------------------------------------
	 *	Initialize the mouse, with the mouse position in REVERSE and
	 *	mouse cmd strings in BOLD.
	 */
	mouse_setup(stdscr, A_REVERSE, A_BOLD);
	mouse_init(stdscr);

	/*--------------------------------------------------------------------
	 *	Do your program stuff.
	 */
	{
		/*----------------------------------------------------------------
		 *	Display our string.
		 */
		move(10, 20);
		caddstr(string, KEY_RETURN);

		move(0,0);
		refresh();

		/*----------------------------------------------------------------
		 *	Wait for a keypress of RETURN.
		 */
		while ((c = mgetch()) != KEY_RETURN)
			;
	}

	/*--------------------------------------------------------------------
	 *	cleanup & termination
	 */
	clear();								/* clear the screen			*/
	refresh();

	endwin();								/* end curses processing	*/
	xvt_here_close(xd);						/* end xvt processing		*/
	xvt_data_free(xd);						/* free xvt data struct		*/

	return (EXIT_SUCCESS);
}
#endif	/* sample program */

#ifndef LIBXVT_H
#define LIBXVT_H

#include <stdio.h>

/*------------------------------------------------------------------------
 * C++ stuff
 */
#ifdef __cplusplus
extern "C" {
#endif

/*------------------------------------------------------------------------
 * misc generic definitions
 */
#ifndef TRUE
#  define TRUE			1
#endif

#ifndef FALSE
#  define FALSE			0
#endif

#ifndef EXIT_SUCCESS
#  define EXIT_SUCCESS	0
#endif

#ifndef EXIT_FAILURE
#  define EXIT_FAILURE	1
#endif

/*------------------------------------------------------------------------
 * library class names
 */
#define XVT_NAME			"pcxterm"		/* default name					*/
#define XVT_APPCLASS		"Pcxterm"		/* app class name for resources	*/

/*------------------------------------------------------------------------
 * defines for system resources
 */
#define XVT_TERM_NAME		"xterm"			/* default TERM name			*/

#define XVT_X_POINTER		0				/* default pointer type			*/
											/* NULL -> use existing pointer	*/
											/* (xterm uses "xterm")			*/

/*------------------------------------------------------------------------
 * extended-capabilities term names
 *
 * Normally these names are used in conjunction with the ecurses package,
 * since they don't exist in either the termcap or terminfo databases,
 * and they define capabilities only used by the ecurses package.
 */
#define XVT_TERM_MONO		XVT_NAME "-m"	/* mono  term type to use		*/
#define XVT_TERM_COLOR		XVT_NAME "-c"	/* color term type to use		*/

/*------------------------------------------------------------------------
 * misc defaults you should know about
 */
#define XVT_MIN_COLS		80				/* min # cols for window		*/
#define XVT_MIN_ROWS		25				/* min # rows for window		*/

#define XVT_NUM_FONTS		8				/* number of fonts supported	*/

#define XVT_SHELL			"sh"			/* default cmd if no $SHELL		*/

#define XVT_MAX_BITS		64				/* max width & height of icon	*/

#define XVT_SB_WIDTH		11				/* scrollbar width				*/
#define XVT_SB_SHADOW		2				/* scrollbar shadow width		*/

/*------------------------------------------------------------------------
 * color definitions
 */
#define XVT_CLR_BLACK		 0
#define XVT_CLR_RED			 1
#define XVT_CLR_GREEN		 2
#define XVT_CLR_YELLOW		 3
#define XVT_CLR_BLUE		 4
#define XVT_CLR_MAGENTA		 5
#define XVT_CLR_CYAN		 6
#define XVT_CLR_WHITE		 7

#define XVT_CLR_GREY		 8
#define XVT_CLR_LTRED		 9
#define XVT_CLR_LTGREEN		10
#define XVT_CLR_LTYELLOW	11
#define XVT_CLR_LTBLUE		12
#define XVT_CLR_LTMAGENTA	13
#define XVT_CLR_LTCYAN		14
#define XVT_CLR_LTWHITE		15

#define XVT_NUM_COLORS		16

/*------------------------------------------------------------------------
 * string option defaults
 */
#define XVT_S_OPT_DISPLAY	"$DISPLAY"		/* X server to contact			*/
#define XVT_S_OPT_TERMNAME	XVT_TERM_NAME	/* value for TERM variable		*/
#define XVT_S_OPT_RESNAME	XVT_APPCLASS	/* alt resource name			*/
#define XVT_S_OPT_GEOMETRY	0				/* window geometry				*/
#define XVT_S_OPT_ICONGEOM	0				/* icon geometry				*/
#define XVT_S_OPT_TERMMODE	0				/* stty modes for window		*/
#define XVT_S_OPT_ICONFILE	0				/* icon file to load			*/
#define XVT_S_OPT_POINTER	XVT_X_POINTER	/* X cursor type				*/
#define XVT_S_OPT_ENVVAR	0				/* env var name for termtype	*/
#define XVT_S_OPT_TITLE		"<cmd>"			/* window title					*/
#define XVT_S_OPT_ITITLE	"<cmd>"			/* icon title					*/
#define XVT_S_OPT_PRINTER	"<spooler>"		/* printer specification		*/
#define XVT_S_OPT_RESSTR	0				/* resource string				*/
#define XVT_S_OPT_RESFILE	0				/* alternate resource file		*/
#define XVT_S_OPT_PTYSPEC	0				/* pty spec						*/

/*------------------------------------------------------------------------
 * font option defaults
 */
#define XVT_F_OPT_REGFONT0	"7x14"			/* default norm font 0			*/
#define XVT_F_OPT_BLDFONT0	"7x14bold"		/* default bold font 0			*/

#define XVT_F_OPT_REGFONT1	"6x13"			/* default norm font 1			*/
#define XVT_F_OPT_BLDFONT1	"6x13bold"		/* default bold font 1			*/

#define XVT_F_OPT_REGFONT2	"7x13"			/* default norm font 2			*/
#define XVT_F_OPT_BLDFONT2	"7x13bold"		/* default bold font 2			*/

#define XVT_F_OPT_REGFONT3	"8x13"			/* default norm font 3			*/
#define XVT_F_OPT_BLDFONT3	"8x13bold"		/* default bold font 3			*/

#define XVT_F_OPT_REGFONT4	"9x15"			/* default norm font 4			*/
#define XVT_F_OPT_BLDFONT4	"9x15bold"		/* default bold font 4			*/

#define XVT_F_OPT_REGFONT5	0				/* default norm font 5			*/
#define XVT_F_OPT_BLDFONT5	0				/* default bold font 5			*/

#define XVT_F_OPT_REGFONT6	0				/* default norm font 6			*/
#define XVT_F_OPT_BLDFONT6	0				/* default bold font 6			*/

#define XVT_F_OPT_REGFONT7	0				/* default norm font 7			*/
#define XVT_F_OPT_BLDFONT7	0				/* default bold font 7			*/

/*------------------------------------------------------------------------
 * color option defaults
 */
#define XVT_C_OPT_BLACK		"Black"			/* default black      color		*/
#define XVT_C_OPT_RED		"Red3"			/* default red        color		*/
#define XVT_C_OPT_GREEN		"Green3"		/* default green      color		*/
#define XVT_C_OPT_YELLOW	"Yellow3"		/* default yellow     color		*/
#define XVT_C_OPT_BLUE		"Blue3"			/* default blue       color		*/
#define XVT_C_OPT_MAGENTA	"Magenta3"		/* default magenta    color		*/
#define XVT_C_OPT_CYAN		"Cyan3"			/* default cyan       color		*/
#define XVT_C_OPT_WHITE		"AntiqueWhite"	/* default white      color		*/

#define XVT_C_OPT_GREY		"Grey25"		/* default grey       color		*/
#define XVT_C_OPT_LTRED		"Red"			/* default lt-red     color		*/
#define XVT_C_OPT_LTGREEN	"Green"			/* default lt-green   color		*/
#define XVT_C_OPT_LTYELLOW	"Yellow"		/* default lt-yellow  color		*/
#define XVT_C_OPT_LTBLUE	"Blue"			/* default lt-blue    color		*/
#define XVT_C_OPT_LTMAGENTA	"Magenta"		/* default lt-magenta color		*/
#define XVT_C_OPT_LTCYAN	"Cyan"			/* default lt-cyan    color		*/
#define XVT_C_OPT_LTWHITE	"White"			/* default lt-white   color		*/

#define XVT_C_OPT_FG		"Black"			/* default foreground color		*/
#define XVT_C_OPT_BG		"White"			/* default background color		*/

#define XVT_C_OPT_POINTER	"Black"			/* default pointer    color		*/
#define XVT_C_OPT_BORDER	"Black"			/* default border     color		*/
#define XVT_C_OPT_CURSOR	"None"			/* default cursor     color		*/

#define XVT_C_OPT_SB		"Black"			/* default scrollbar  color		*/
#define XVT_C_OPT_ST		"White"			/* default SB trough  color		*/

/*------------------------------------------------------------------------
 * numeric option defaults
 */
#define XVT_N_OPT_COLS		XVT_MIN_COLS	/* starting number of columns	*/
#define XVT_N_OPT_ROWS		XVT_MIN_ROWS	/* starting number of rows		*/
#define XVT_N_OPT_TABWIDTH	8				/* # spaces per tab stop		*/
#define XVT_N_OPT_BORDER	1				/* default border width			*/
#define XVT_N_OPT_MARGIN	0				/* default margin width			*/
#define XVT_N_OPT_MBCOLS	10				/* margin bell cols from right	*/
#define XVT_N_OPT_FONTNO	0				/* default font number			*/
#define XVT_N_OPT_SBLINES	300				/* number of lines to save		*/

/*------------------------------------------------------------------------
 * boolean option defaults
 */
#define XVT_B_OPT_ICONIC	FALSE			/* start iconic					*/
#define XVT_B_OPT_MONO		FALSE			/* force display to mono		*/
#define XVT_B_OPT_LOGIN		FALSE			/* login shell					*/
#define XVT_B_OPT_HILIGHT	FALSE			/* always highlight cursor		*/
#define XVT_B_OPT_REVVID	FALSE			/* reverse video				*/
#define XVT_B_OPT_VISBELL	FALSE			/* visual bell					*/
#define XVT_B_OPT_AUTOWRAP	TRUE			/* auto wraparound				*/
#define XVT_B_OPT_REVWRAP	FALSE			/* reverse wraparound			*/
#define XVT_B_OPT_MBELL		FALSE			/* margin bell					*/
#define XVT_B_OPT_QUIET		FALSE			/* inhibit ^G					*/
#define XVT_B_OPT_IGNPTR	FALSE			/* ignore pointer events		*/
#define XVT_B_OPT_CONSOLE	FALSE			/* intercept console msgs		*/
#define XVT_B_OPT_LBLINFO	FALSE			/* debug info in window label	*/
#define XVT_B_OPT_KPFONT	TRUE			/* KP+/KP- changes fonts		*/
#define XVT_B_OPT_MAPOUTPUT	FALSE			/* map if output when iconic	*/
#define XVT_B_OPT_MAPALERT	FALSE			/* map if ^G output when iconic	*/
#define XVT_B_OPT_C132		FALSE			/* allow 80-132 col switching	*/
#define XVT_B_OPT_XERRORS	FALSE			/* show X error messages		*/
#define XVT_B_OPT_PRTKEEP	FALSE			/* keep printer open until term	*/
#define XVT_B_OPT_SBENABLE	FALSE			/* enable scrollbar				*/
#define XVT_B_OPT_SBRIGHT	TRUE			/* scrollbar on right			*/
#define XVT_B_OPT_SBSHADOW	TRUE			/* shadow scrollbar				*/
#define XVT_B_OPT_SBSCROUT	TRUE			/* scroll on tty output			*/
#define XVT_B_OPT_SBSCRKEY	TRUE			/* scroll on keypress			*/

/*------------------------------------------------------------------------
 * debug option types & log filenames
 */
#define XVT_D_OPT_EVENTS	"events"		/* txt: xlated events			*/
#define XVT_D_OPT_ACTION	"action"		/* txt: xlated X11 events		*/
#define XVT_D_OPT_OUTPUT	"output"		/* txt: xlated child output		*/
#define XVT_D_OPT_SCREEN	"screen"		/* bin: raw    child output		*/
#define XVT_D_OPT_INPKBD	"inpkbd"		/* txt: xlated child input		*/
#define XVT_D_OPT_KEYBRD	"keybrd"		/* bin: raw    child input		*/
#define XVT_D_OPT_RESFIL	"resfil"		/* txt: resource file entries	*/
#define XVT_D_OPT_TKOPTS	"tkopts"		/* txt: dump of all options		*/

#define XVT_D_OPT_EVENTS_LOG	XVT_D_OPT_EVENTS ".log"
#define XVT_D_OPT_ACTION_LOG	XVT_D_OPT_ACTION ".log"
#define XVT_D_OPT_OUTPUT_LOG	XVT_D_OPT_OUTPUT ".log"
#define XVT_D_OPT_SCREEN_LOG	XVT_D_OPT_SCREEN ".log"
#define XVT_D_OPT_INPKBD_LOG	XVT_D_OPT_INPKBD ".log"
#define XVT_D_OPT_KEYBRD_LOG	XVT_D_OPT_KEYBRD ".log"
#define XVT_D_OPT_RESFIL_LOG	XVT_D_OPT_RESFIL ".log"
#define XVT_D_OPT_TKOPTS_LOG	XVT_D_OPT_TKOPTS ".log"

/*------------------------------------------------------------------------
 * event routine definition
 *
 * This is the routine the user must call frequently when in
 * non-threaded "here" mode.
 *
 * "data" is the "event_data" pointer returned in the XVT_DATA struct
 * from the call to xvt_run_here().
 *
 * "ms" is the number of milliseconds to wait for an event.
 * Normal values would be:
 *
 *		0 ms	Just want to check if anything there.
 *				Normal non-threaded "here" type of call.
 *
 *				The idea here is to just "give" a cpu thread
 *				to the interface loop as often as possible.
 *				Normally, this is whenever the "here" routine does any
 *				I/O to the "terminal".
 *
 *		>0 ms	Want to wait while checking. (Usually 5-10 ms).
 *				Normal "task" type of call.
 *				Normal threaded "here" type of call.
 *
 *				Also used in non-threaded "here" mode when it is
 *				in "sub-task" mode. (i.e. when in "here" mode
 *				the program has shelled-out & just needs to keep
 *				the interface alive while waiting for the sub-task
 *				to die.)
 *
 * The pointer to this routine is returned in the "event_rtn" entry in
 * the XVT_DATA struct.  If this pointer is NULL, then a sub-thread was
 * started to process events and no call is needed.
 *
 * This routine returns (0) if OK and (-1) otherwise (usually because of
 * EOF from child).  A (-1) return indicates that the caller should
 * terminate processing (including calling xvt_close_here()).
 */
typedef int		XVT_EVENT_RTN	(void *data,	/* ptr to (our) user data	*/
									int ms);	/* from user - ms to wait	*/

/*------------------------------------------------------------------------
 * xvt-data struct
 *
 * This struct provides all options to the xvt routines, and receives
 * any returned values.
 */
struct xvt_data
{
	/*--------------------------------------------------------------------
	 * default application name
	 *
	 * This is not set via any resource processing, it is used when
	 * displaying defaults.
	 */
	const char *	appname;				/* default app name			*/

	/*--------------------------------------------------------------------
	 * string options
	 */
	const char *	display;				/* display name				*/
	const char *	termname;				/* TERM name to set			*/
	const char *	resname;				/* alternate resource name	*/
	const char *	geometry;				/* window geometry (chars)	*/
	const char *	icongeom;				/* icon geometry (pos only)	*/
	const char *	termmode;				/* stty mode settings		*/
	const char *	ico_file;				/* icon file to use			*/
	const char *	pointer;				/* pointer type				*/
	const char *	envvar;					/* name of env var to set	*/
	const char *	win_label;				/* window label				*/
	const char *	ico_label;				/* icon label				*/
	const char *	printer;				/* printer spec				*/
	const char *	resstr;					/* resource string			*/
	const char *	resfile;				/* alternate resource file	*/
	const char *	ptyspec;				/* pty spec					*/

	/*--------------------------------------------------------------------
	 * font options
	 */
	const char *	reg_fname[XVT_NUM_FONTS];	/* reg  font list		*/
	const char *	bld_fname[XVT_NUM_FONTS];	/* bold font list		*/

	/*--------------------------------------------------------------------
	 * color options
	 */
	const char *	scr_color[XVT_NUM_COLORS];	/* screen colors		*/

	const char *	fg_color;				/* foreground color			*/
	const char *	bg_color;				/* background color			*/
	const char *	cr_color;				/* cursor     color			*/
	const char *	pr_color;				/* pointer    color			*/
	const char *	bd_color;				/* border     color			*/
	const char *	sb_color;				/* scrollbar  color			*/
	const char *	st_color;				/* SB trough  color			*/

	/*--------------------------------------------------------------------
	 * numeric options
	 */
	int				scr_cols;				/* initial # cols in screen	*/
	int				scr_rows;				/* initial # rows in screen	*/
	int				scr_tabs;				/* tab width				*/
	int				border;					/* border width in pixels	*/
	int				margin;					/* margin width in pixels	*/
	int				mbcols;					/* margin bell cols from rt	*/
	int				fontno;					/* font number				*/
	int				sblines;				/* number of lines to save	*/

	/*--------------------------------------------------------------------
	 * boolean options
	 */
	int				iconic;					/* TRUE to start iconic		*/
	int				monodisp;				/* TRUE to force mono		*/
	int				login;					/* TRUE for login shell		*/
	int				ahcursor;				/* TRUE to hilite cursor	*/
	int				reverse;				/* TRUE for reverse video	*/
	int				visual;					/* TRUE for visual bell		*/
	int				autowrap;				/* TRUE for auto wraparound	*/
	int				revwrap;				/* TRUE for reverse wrap	*/
	int				mbell;					/* TRUE for margin bell		*/
	int				quiet;					/* TRUE to inhibit bell		*/
	int				ignptr;					/* TRUE to ignore mouse		*/
	int				console;				/* TRUE to get console msgs	*/
	int				lblinfo;				/* TRUE to show label info	*/
	int				kpfont;					/* TRUE for KP+/- ch fonts	*/
	int				mapoutput;				/* TRUE to map on output	*/
	int				mapalert;				/* TRUE to map on ^G output	*/
	int				c132;					/* TRUE for 80-132 switch	*/
	int				xerrors;				/* TRUE to show X errors	*/
	int				prtkeep;				/* TRUE to keep prtr open	*/
	int				sb_enable;				/* TRUE to enable SB		*/
	int				sb_right;				/* TRUE for SB on right		*/
	int				sb_shadow;				/* TRUE to shadow SB		*/
	int				sb_scrout;				/* TRUE to scroll on output	*/
	int				sb_scrkey;				/* TRUE to scroll on key	*/

	/*--------------------------------------------------------------------
	 * debug streams
	 *
	 * These are used by the libxvt routines if they were opened
	 * by the caller.  Note that if you intend to run in "here" mode,
	 * these streams should not point to stderr.
	 *
	 * They also may be opened by libxvt if the appropriate escape
	 * sequence is received.  libxvt does NOT close these streams, even
	 * if it opens them.
	 */
	FILE *			events_fp;				/* txt: xlated events		*/
	FILE *			action_fp;				/* txt: xlated X11 events	*/
	FILE *			output_fp;				/* txt: xlated child output	*/
	FILE *			screen_fp;				/* bin: raw    child output	*/
	FILE *			inpkbd_fp;				/* txt: xlated child input	*/
	FILE *			keybrd_fp;				/* bin: raw    child input	*/
	FILE *			resfil_fp;				/* txt: resource file		*/
	FILE *			tkopts_fp;				/* txt: options dump		*/

	/*--------------------------------------------------------------------
	 * values returned from xvt_tk_args()
	 */
	int				argc;					/* new argc value			*/

	char **			argv;					/* new argv vector			*/
											/* Note: this vector is		*/
											/* freed when the struct is	*/
											/* freed, and is not copied	*/
											/* by xvt_data_clone().		*/

	/*--------------------------------------------------------------------
	 * values returned from xvt_run_here()
	 *
	 * If threads are not available, "event_rtn" will be set to non-zero.
	 * This is the event routine the caller must be prepared to call
	 * frequently to "activate" the event mechanism.
	 *
	 * If threads are available, "event_rtn" is set to zero.
	 *
	 * Note that "event_data" will always be set, even though event_rtn may
	 * be zero.  This enables the caller to query (xd->event_data != 0) to
	 * determine if he is in an "event-rtn" type of environment.
	 * (Also, xvt_here_close() & xvt_here_clone() uses event_data for its
	 * data pointer).
	 */
	char			term_type[64];			/* term type to use			*/
	const char **	term_defs;				/* list of term definitions	*/

	int				fd_inp;					/* input  file descriptor	*/
	int				fd_out;					/* output file descriptor	*/

	XVT_EVENT_RTN *	event_rtn;				/* event routine to call	*/
	void *			event_data;				/* data to call it with		*/

	/*--------------------------------------------------------------------
	 * values returned from xvt_task_run() & xvt_here_run()
	 * and updated asynchronously by the event routine.
	 */
	int				cur_cols;				/* current # of cols in win	*/
	int				cur_rows;				/* current # of rows in win	*/

	/*--------------------------------------------------------------------
	 * values returned from xvt_task_run()
	 */
	int				exit_code;				/* sub-tasks's exit code	*/

	/*--------------------------------------------------------------------
	 * returned error values
	 */
	int				error_num;				/* error code				*/
	int				error_quit;				/* TRUE if pgm should quit	*/
	char			error_buf[128];			/* error msg buffer			*/

	/*--------------------------------------------------------------------
	 * toolkit option buffer (no user access to this)
	 */
	char			tkbuf[1024];			/* string buffer			*/
	int				tkbuf_index;			/* index to current buf pos	*/
};
typedef struct xvt_data XVT_DATA;

/*------------------------------------------------------------------------
 * routines for accessing the XVT_DATA struct.
 *
 * By using these routines, the XVT_DATA struct can be treated as opaque.
 */
extern XVT_DATA *	xvt_data_alloc				(void);
extern void			xvt_data_free				(XVT_DATA *xd);
extern void			xvt_data_init				(XVT_DATA *xd, int set_dflt);
extern XVT_DATA *	xvt_data_clone				(XVT_DATA *xd);

extern void			xvt_data_set_scr_cols		(XVT_DATA *xd, int n);
extern void			xvt_data_set_scr_rows		(XVT_DATA *xd, int n);
extern void			xvt_data_set_scr_tabs		(XVT_DATA *xd, int n);
extern void			xvt_data_set_border			(XVT_DATA *xd, int n);
extern void			xvt_data_set_margin			(XVT_DATA *xd, int n);
extern void			xvt_data_set_mbcols			(XVT_DATA *xd, int n);
extern void			xvt_data_set_fontno			(XVT_DATA *xd, int n);
extern void			xvt_data_set_sblines		(XVT_DATA *xd, int n);

extern void			xvt_data_set_iconic			(XVT_DATA *xd, int b);
extern void			xvt_data_set_monodisp		(XVT_DATA *xd, int b);
extern void			xvt_data_set_login			(XVT_DATA *xd, int b);
extern void			xvt_data_set_ahcursor		(XVT_DATA *xd, int b);
extern void			xvt_data_set_reverse		(XVT_DATA *xd, int b);
extern void			xvt_data_set_visual			(XVT_DATA *xd, int b);
extern void			xvt_data_set_autowrap		(XVT_DATA *xd, int b);
extern void			xvt_data_set_revwrap		(XVT_DATA *xd, int b);
extern void			xvt_data_set_mbell			(XVT_DATA *xd, int b);
extern void			xvt_data_set_quiet			(XVT_DATA *xd, int b);
extern void			xvt_data_set_ignptr			(XVT_DATA *xd, int b);
extern void			xvt_data_set_console		(XVT_DATA *xd, int b);
extern void			xvt_data_set_lblinfo		(XVT_DATA *xd, int b);
extern void			xvt_data_set_kpfont			(XVT_DATA *xd, int b);
extern void			xvt_data_set_mapoutput		(XVT_DATA *xd, int b);
extern void			xvt_data_set_mapalert		(XVT_DATA *xd, int b);
extern void			xvt_data_set_c132			(XVT_DATA *xd, int b);
extern void			xvt_data_set_xerrors		(XVT_DATA *xd, int b);
extern void			xvt_data_set_prtkeep		(XVT_DATA *xd, int b);
extern void			xvt_data_set_sb_enable		(XVT_DATA *xd, int b);
extern void			xvt_data_set_sb_right		(XVT_DATA *xd, int b);
extern void			xvt_data_set_sb_shadow		(XVT_DATA *xd, int b);
extern void			xvt_data_set_sb_scrout		(XVT_DATA *xd, int b);
extern void			xvt_data_set_sb_scrkey		(XVT_DATA *xd, int b);

extern void			xvt_data_set_appname		(XVT_DATA *xd, const char *s);

extern void			xvt_data_set_reg_fname		(XVT_DATA *xd, const char *s,
													int n);
extern void			xvt_data_set_bld_fname		(XVT_DATA *xd, const char *s,
													int n);

extern void			xvt_data_set_display		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_termname		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_resname		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_geometry		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_icongeom		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_termmode		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_ico_file		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_pointer		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_envvar			(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_win_label		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_ico_label		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_printer		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_resstr			(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_resfile		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_ptyspec		(XVT_DATA *xd, const char *s);

extern void			xvt_data_set_scr_color		(XVT_DATA *xd, const char *s,
													int n);

extern void			xvt_data_set_fg_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_bg_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_cr_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_pr_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_bd_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_sb_color		(XVT_DATA *xd, const char *s);
extern void			xvt_data_set_st_color		(XVT_DATA *xd, const char *s);

extern void			xvt_data_set_events_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_action_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_output_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_screen_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_inpkbd_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_keybrd_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_resfil_fp		(XVT_DATA *xd, FILE *fp);
extern void			xvt_data_set_tkopts_fp		(XVT_DATA *xd, FILE *fp);

extern int			xvt_data_get_scr_cols		(XVT_DATA *xd);
extern int			xvt_data_get_scr_rows		(XVT_DATA *xd);
extern int			xvt_data_get_scr_tabs		(XVT_DATA *xd);
extern int			xvt_data_get_border			(XVT_DATA *xd);
extern int			xvt_data_get_margin			(XVT_DATA *xd);
extern int			xvt_data_get_mbcols			(XVT_DATA *xd);
extern int			xvt_data_get_fontno			(XVT_DATA *xd);
extern int			xvt_data_get_sblines		(XVT_DATA *xd);

extern int			xvt_data_get_iconic			(XVT_DATA *xd);
extern int			xvt_data_get_monodisp		(XVT_DATA *xd);
extern int			xvt_data_get_login			(XVT_DATA *xd);
extern int			xvt_data_get_ahcursor		(XVT_DATA *xd);
extern int			xvt_data_get_reverse		(XVT_DATA *xd);
extern int			xvt_data_get_visual			(XVT_DATA *xd);
extern int			xvt_data_get_autowrap		(XVT_DATA *xd);
extern int			xvt_data_get_revwrap		(XVT_DATA *xd);
extern int			xvt_data_get_mbell			(XVT_DATA *xd);
extern int			xvt_data_get_quiet			(XVT_DATA *xd);
extern int			xvt_data_get_ignptr			(XVT_DATA *xd);
extern int			xvt_data_get_console		(XVT_DATA *xd);
extern int			xvt_data_get_lblinfo		(XVT_DATA *xd);
extern int			xvt_data_get_kpfont			(XVT_DATA *xd);
extern int			xvt_data_get_mapoutput		(XVT_DATA *xd);
extern int			xvt_data_get_mapalert		(XVT_DATA *xd);
extern int			xvt_data_get_c132			(XVT_DATA *xd);
extern int			xvt_data_get_xerrors		(XVT_DATA *xd);
extern int			xvt_data_get_prtkeep		(XVT_DATA *xd);
extern int			xvt_data_get_sb_enable		(XVT_DATA *xd);
extern int			xvt_data_get_sb_right		(XVT_DATA *xd);
extern int			xvt_data_get_sb_shadow		(XVT_DATA *xd);
extern int			xvt_data_get_sb_scrout		(XVT_DATA *xd);
extern int			xvt_data_get_sb_scrkey		(XVT_DATA *xd);

extern const char *	xvt_data_get_appname		(XVT_DATA *xd);

extern const char *	xvt_data_get_reg_fname		(XVT_DATA *xd, int n);
extern const char *	xvt_data_get_bld_fname		(XVT_DATA *xd, int n);

extern const char *	xvt_data_get_display		(XVT_DATA *xd);
extern const char *	xvt_data_get_termname		(XVT_DATA *xd);
extern const char *	xvt_data_get_resname		(XVT_DATA *xd);
extern const char *	xvt_data_get_geometry		(XVT_DATA *xd);
extern const char *	xvt_data_get_icongeom		(XVT_DATA *xd);
extern const char *	xvt_data_get_termmode		(XVT_DATA *xd);
extern const char *	xvt_data_get_ico_file		(XVT_DATA *xd);
extern const char *	xvt_data_get_pointer		(XVT_DATA *xd);
extern const char *	xvt_data_get_envvar			(XVT_DATA *xd);
extern const char *	xvt_data_get_win_label		(XVT_DATA *xd);
extern const char *	xvt_data_get_ico_label		(XVT_DATA *xd);
extern const char *	xvt_data_get_printer		(XVT_DATA *xd);
extern const char *	xvt_data_get_resstr			(XVT_DATA *xd);
extern const char *	xvt_data_get_resfile		(XVT_DATA *xd);
extern const char *	xvt_data_get_ptyspec		(XVT_DATA *xd);

extern const char *	xvt_data_get_scr_color		(XVT_DATA *xd, int n);
extern const char *	xvt_data_get_fg_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_bg_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_cr_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_pr_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_bd_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_sb_color		(XVT_DATA *xd);
extern const char *	xvt_data_get_st_color		(XVT_DATA *xd);

extern FILE *		xvt_data_get_events_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_action_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_output_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_screen_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_inpkbd_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_keybrd_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_resfil_fp		(XVT_DATA *xd);
extern FILE *		xvt_data_get_tkopts_fp		(XVT_DATA *xd);

extern int			xvt_data_get_argc			(XVT_DATA *xd);
extern char **		xvt_data_get_argv			(XVT_DATA *xd);

extern const char *	xvt_data_get_term_type		(XVT_DATA *xd);
extern const char **xvt_data_get_term_defs		(XVT_DATA *xd);
extern int			xvt_data_get_fd_inp			(XVT_DATA *xd);
extern int			xvt_data_get_fd_out			(XVT_DATA *xd);
extern XVT_EVENT_RTN *
					xvt_data_get_event_rtn		(XVT_DATA *xd);
extern void *		xvt_data_get_event_data		(XVT_DATA *xd);
extern int			xvt_data_get_cur_cols		(XVT_DATA *xd);
extern int			xvt_data_get_cur_rows		(XVT_DATA *xd);
extern int			xvt_data_get_exit_code		(XVT_DATA *xd);
extern int			xvt_data_get_error_num		(XVT_DATA *xd);
extern int			xvt_data_get_error_quit		(XVT_DATA *xd);
extern const char *	xvt_data_get_error_buf		(XVT_DATA *xd);

/*------------------------------------------------------------------------
 * xvt error codes
 */
#define XVT_OK					0		/* setup OK							*/
#define XVT_ERR_NOTSUPP			1		/* not supported					*/
#define XVT_ERR_NOTINWIN		2		/* not in windows environment		*/
#define XVT_ERR_CANTOPENDISP	3		/* cannot open display				*/
#define XVT_ERR_CANTOPENPTYS	4		/* cannot open pseudo-ttys			*/
#define XVT_ERR_CANTLOADFONTS	5		/* cannot load requested fonts		*/
#define XVT_ERR_CANTMAKEWIN		6		/* cannot create window				*/
#define XVT_ERR_CANTFORK		7		/* cannot fork subtask				*/
#define XVT_ERR_CANTDETACH		8		/* cannot detach process			*/
#define XVT_ERR_CANTEXEC		9		/* cannot execute cmd				*/
#define XVT_ERR_NOMEM			10		/* no memory						*/
#define XVT_ERR_NOTINIT			11		/* not initialized					*/
#define XVT_ERR_NULLDATA		12		/* null data 						*/
#define XVT_ERR_INVOPT			13		/* invalid option					*/
#define XVT_ERR_INVARG			14		/* invalid option argument			*/

/*------------------------------------------------------------------------
 * Query whether XVT is available.
 *
 * Returns: TRUE or FALSE
 */
extern int	xvt_available		(void);

/*------------------------------------------------------------------------
 * run in "task" mode
 *
 * This runs the "cmd" specified by the argv vector as a sub-task.
 * If argv is NULL or empty, then $SHELL is used.  If $SHELL is
 * NULL or empty, then XVT_SHELL is used.
 *
 * It returns when the sub-task dies.  If it failed in running the
 * specified command, it returns an error code and the entries
 * "error_num" and "error_buf" in the XVT_DATA struct will be set.
 * If it successfully ran the specified command, it returns XVT_OK and the
 * exit code from the command will be stored in the "exit_code" field
 * of the XVT_DATA struct.
 */
extern int	xvt_task_run		(XVT_DATA *xd,
									char **argv);

/*------------------------------------------------------------------------
 * run in "here" mode.
 *
 * This sets up the display and the pseudo-tty for the interface,
 * and either starts a thread to process the interface or
 * stores a pointer to an event routine to be called by the user.
 * The user should be prepared to call the specified event routine frequently
 * if it is not zero.
 *
 * If "detach" is TRUE, then the program is made into a pseudo-daemon,
 * with stdin/stdout/stderr pointing to the pseudo-tty created.
 *
 * This routine returns after the initial setup. The return value is
 * XVT_OK if successful, and an error code if not.  The entries "error_num"
 * and "error_buf" in the XVT_DATA struct will also be set if not successful.
 *
 * If this call fails, then the entry "error_quit" will be set to
 * FALSE if it is possible to continue and TRUE if not.
 */
extern int	xvt_here_run		(XVT_DATA *xd,
									int detach);

/*------------------------------------------------------------------------
 * close out processing when in "here" mode
 *
 * This call will shutdown any sub-thread which was started to process
 * events and will free all data structs which were allocated.
 *
 * This routine returns XVT_OK if successful, and XVT_ERR_NOTINIT if
 * xvt_here_run() was never called (or if the call failed).
 */
extern int	xvt_here_close		(XVT_DATA *xd);

/*------------------------------------------------------------------------
 * create a "clone" of the current task with a new window
 *
 * Note that this routine has two returns: the "parent" and the "child"
 * returns.  After returning, the two tasks are equivalent (i.e. they
 * are both running as "daemons").  This means that after a successful
 * return, there is no way to determine which task was the original one
 * unless you cached your PID prior to the call.
 *
 * Note also that this routine is only valid when running in "here" mode
 * (i.e. after a succesful call to xvt_here_run() is made).
 * It returns XVT_OK if a clone was successfully created, and an error
 * code if not.
 */
extern int	xvt_here_clone		(XVT_DATA *xd);

/*------------------------------------------------------------------------
 * terminal-emulator program
 *
 * This is the working main of the terminal-emulator program.
 * The "real" main just calls this & returns.
 *
 * If this routine is not sucessful in running the specified command,
 * it will return EXIT_FAILURE, and an error msg will be displayed to stderr.
 * Otherwise, it returns the exit code from the command which was run.
 */
extern int	xvt_main			(int argc, char **argv);

/*------------------------------------------------------------------------
 * extract toolkit options from an argv vector
 *
 * This routine will extract all toolkit options from an argv vector
 * and will store a new argc & argv with the toolkit options removed.
 *
 * This routine returns XVT_OK if successful.  Otherwise it returns
 * an error code and a message will be present in error_buf.
 */
extern int	xvt_tk_args			(XVT_DATA *xd,
									int argc,
									char **argv);

/*------------------------------------------------------------------------
 * Process a toolkit option
 *
 * If the option is a boolean option, <arg> is ignored.
 * For string and numeric options, <option> contains the option string
 * and <arg> contains the value.  Optionally, <option> may be in the form
 * "<option>=<arg>", and <arg> is then ignored.
 *
 * If <resent> is FALSE, then string contains option strings,
 * otherwise it contains resource-names.
 *
 * This routine returns:
 *		-1			error
 *		 0			not a toolkit option
 *		 1			toolkit option - option used
 *		 2			toolkit option - option & arg used
 *
 *		XVT_TK_OPT_*	option type in type
 */
#define XVT_TK_OPT_ERR		-1		/* error occurred	*/
#define XVT_TK_OPT_INV		0		/* invalid option	*/
#define XVT_TK_OPT_STR		1		/* string  option	*/
#define XVT_TK_OPT_CLR		2		/* color   option	*/
#define XVT_TK_OPT_FNT		3		/* font    option	*/
#define XVT_TK_OPT_NUM		4		/* numeric option	*/
#define XVT_TK_OPT_BLN		5		/* boolean option	*/
#define XVT_TK_OPT_DBG		6		/* debug   option	*/
#define XVT_TK_OPT_CMD		7		/* cmd list follows	*/

extern int	xvt_tk_opt_set		(XVT_DATA *xd,
									const char *option,
									const char *arg,
									int resent,
									int *type);

/*------------------------------------------------------------------------
 * process a toolkit option list
 *
 * List is in form "<opt>[=<arg>], ..." (semicolons work too)
 *
 * If <resent> is FALSE, then string contains option strings,
 * otherwise it contains resource-names.
 *
 * This routine returns XVT_OK if all options are valid, otherwise
 * an error message is put in error buf.
 */
extern int	xvt_tk_optlist		(XVT_DATA *xd,
									const char *list,
									int resent);

/*------------------------------------------------------------------------
 * load all data a resource file
 *
 * This routine returns XVT_OK if all options are valid, otherwise
 * an error message is put in error buf.
 */
extern int	xvt_tk_resfile_load	(XVT_DATA *xd,
									const char *path,
									const char *resname);

/*------------------------------------------------------------------------
 * save all data to a resource file
 *
 * This routine returns XVT_OK if all options are valid, otherwise
 * an error message is put in error buf.
 */
extern int	xvt_tk_resfile_save	(XVT_DATA *xd,
									const char *path,
									const char *resname);

/*------------------------------------------------------------------------
 * dump all data to a stream
 *
 * Normally, only options which were set are dumped, but if <show_dflt>
 * is set, then all values are shown.
 *
 * This routine returns XVT_OK if all options are valid, otherwise
 * an error message is put in error buf.
 */
extern int	xvt_tk_options_dump	(XVT_DATA *xd,
									FILE *fp,
									int show_dflt);

/*------------------------------------------------------------------------
 * process all known resources
 *
 * This routine returns XVT_OK if all options are valid.
 */
extern int	xvt_tk_resource		(XVT_DATA *xd);

/*------------------------------------------------------------------------
 * output a list of valid toolkit options
 *
 * Note that <xd> may be NULL in this call.
 *
 * verbose:
 *			0		short listing
 *			1		long listing
 *			2		detailed listing
 *
 *	here:
 *			TRUE	running in "here" mode
 *			FALSE	not in "here" mode
 */
extern void	xvt_tk_help			(XVT_DATA *xd,
									FILE *fp,
									int verbose,
									int here);

/*------------------------------------------------------------------------
 * C++ stuff
 */
#ifdef __cplusplus
}
#endif

#endif /* LIBXVT_H */
