/*------------------------------------------------------------------------
 * xvt logging routines
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * get log stream
 */
static FILE * xvt_log_stream (TERMDATA *td, int log)
{
	switch (log)
	{
	case XVT_LOG_EVENTS:	return (td->xd->events_fp);
	case XVT_LOG_ACTION:	return (td->xd->action_fp);
	case XVT_LOG_OUTPUT:	return (td->xd->output_fp);
	case XVT_LOG_SCREEN:	return (td->xd->screen_fp);
	case XVT_LOG_INPKBD:	return (td->xd->inpkbd_fp);
	case XVT_LOG_KEYBRD:	return (td->xd->keybrd_fp);
	default:				return (0);
	}
}

/*------------------------------------------------------------------------
 * output log prefix info (used for events log)
 *
 * Here is the place to add any timestamps, pid, etc.
 */
static void xvt_log_prefix (TERMDATA *td, FILE *fp, const char *str)
{
	if (str != 0)
	{
		if (td->log_pid)
			fprintf(fp, "%-5d: ", td->our_pid);

		fprintf(fp, "%-6s : ", str);
	}
}

/*------------------------------------------------------------------------
 * change log state
 */
void xvt_log_set (TERMDATA *td, int log, int on)
{
	FILE **			pfp;
	const char *	file;

	/*--------------------------------------------------------------------
	 * get pointer to stream-pointer
	 */
	switch (log)
	{
	case XVT_LOG_EVENTS:
		pfp = &td->xd->events_fp;	file = XVT_D_OPT_EVENTS_LOG;	break;

	case XVT_LOG_ACTION:
		pfp = &td->xd->action_fp;	file = XVT_D_OPT_ACTION_LOG;	break;

	case XVT_LOG_OUTPUT:
		pfp = &td->xd->output_fp;	file = XVT_D_OPT_OUTPUT_LOG;	break;

	case XVT_LOG_SCREEN:
		pfp = &td->xd->screen_fp;	file = XVT_D_OPT_SCREEN_LOG;	break;

	case XVT_LOG_INPKBD:
		pfp = &td->xd->inpkbd_fp;	file = XVT_D_OPT_INPKBD_LOG;	break;

	case XVT_LOG_KEYBRD:
		pfp = &td->xd->keybrd_fp;	file = XVT_D_OPT_KEYBRD_LOG;	break;

	default:
		return;
	}

	/*--------------------------------------------------------------------
	 * now set it on/off
	 */
	if (on)
	{
		if (*pfp == 0)
		{
			*pfp = fopen(file, "w");
		}
	}
	else
	{
		if (*pfp != 0)
		{
			fclose(*pfp);
			*pfp = 0;
		}
	}
}

/*------------------------------------------------------------------------
 * query whether a particular log is on
 *
 * We separated this from testing the stream pointer, since we may
 * want to add some debug-level info someday.
 */
int xvt_log_is_on (TERMDATA *td, int log)
{
	FILE *	fp = xvt_log_stream(td, log);

	return (fp != 0);
}

/*------------------------------------------------------------------------
 * event logging
 */
void xvt_log_events_fmt (TERMDATA *td, const char *event,
	const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	xvt_log_events_var(td, event, fmt, args);
	va_end(args);
}

void xvt_log_events_var (TERMDATA *td, const char *event,
	const char *fmt, va_list args)
{
	if (xvt_log_is_on(td, XVT_LOG_EVENTS) && event != 0)
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_EVENTS);

		xvt_log_prefix(td, fp, event);
		vfprintf(fp, fmt, args);
		fputc('\n', fp);
		fflush(fp);
	}
}

void xvt_log_events_buf (TERMDATA *td, const char *event,
	const unsigned char *buf, int num)
{
	if (xvt_log_is_on(td, XVT_LOG_EVENTS) && event != 0)
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_EVENTS);
		int		i;

		xvt_log_prefix(td, fp, event);

		for (i=0; i<num; i++)
		{
			if (buf[i] < ' ')
			{
				fputc('^', fp);
				fputc(buf[i] + '@', fp);
			}
			else
			{
				fputc(buf[i], fp);
			}
		}
		fputc('\n', fp);
		fflush(fp);
	}
}

void xvt_log_events_str (TERMDATA *td, const char *event,
	const char *str)
{
	xvt_log_events_buf(td, event, (const unsigned char *)str, strlen(str));
}

/*------------------------------------------------------------------------
 * output logging
 */
void xvt_log_output_seq (TERMDATA *td, const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	xvt_log_events_var(td, "output", fmt, args);

	if (xvt_log_is_on(td, XVT_LOG_OUTPUT))
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_OUTPUT);

		/*----------------------------------------------------------------
		 * if last output was not an esc-seq, then output a NL
		 */
		if (! td->output_debug_seq)
			fputc('\n', fp);
		fputc('<', fp);
		td->output_debug_seq = TRUE;

		/*----------------------------------------------------------------
		 * output message
		 */
		vfprintf(fp, fmt, args);
		fputs(">\n", fp);
		fflush(fp);
	}
	va_end(args);
}

void xvt_log_output_chr (TERMDATA *td, const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	xvt_log_events_var(td, NULL, fmt, args);

	if (xvt_log_is_on(td, XVT_LOG_OUTPUT))
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_OUTPUT);

		/*----------------------------------------------------------------
		 * indicate not in an esc-seq
		 */
		td->output_debug_seq = FALSE;

		/*----------------------------------------------------------------
		 * output message
		 */
		vfprintf(fp, fmt, args);
		fflush(fp);
	}
	va_end(args);
}

/*------------------------------------------------------------------------
 * action logging
 */
void xvt_log_action_var (TERMDATA *td, const char *fmt, va_list args)
{
	xvt_log_events_var(td, "action", fmt, args);

	if (xvt_log_is_on(td, XVT_LOG_ACTION))
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_ACTION);

		vfprintf(fp, fmt, args);
		fputc('\n', fp);
		fflush(fp);
	}
}

/*------------------------------------------------------------------------
 * screen logging
 */
void xvt_log_screen_buf (TERMDATA *td, const unsigned char *buf, int num)
{
	if (num < 0)
		xvt_log_events_fmt(td, "screen", "EOF");
	else
		xvt_log_events_fmt(td, "screen", "%d bytes", num);

	if (xvt_log_is_on(td, XVT_LOG_SCREEN) && num > 0)
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_SCREEN);

		fwrite(buf, num, 1, fp);
		fflush(fp);
	}
}

/*------------------------------------------------------------------------
 * inpkbd logging
 */
void xvt_log_inpkbd_fmt (TERMDATA *td, const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	xvt_log_events_var(td, "inpkbd", fmt, args);

	if (xvt_log_is_on(td, XVT_LOG_INPKBD))
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_INPKBD);

		vfprintf(fp, fmt, args);
		fputc('\n', fp);
		fflush(fp);
	}
	va_end(args);
}

/*------------------------------------------------------------------------
 * keybrd logging
 */
void xvt_log_keybrd_str (TERMDATA *td, const char *str)
{
	xvt_log_events_str(td, "keybrd", str);

	if (xvt_log_is_on(td, XVT_LOG_KEYBRD))
	{
		FILE *	fp = xvt_log_stream(td, XVT_LOG_KEYBRD);

		fputs(str, fp);
		fputc('\n', fp);
		fflush(fp);
	}
}
