/*------------------------------------------------------------------------
 * "main" program for libxvt terminal-emulator
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * This is needed if getopt() is not available.
 */
#include "sysopt.h"

/*------------------------------------------------------------------------
 * usage display
 */
static void xvt_main_usage (const char *pgm, int level)
{
	if (level == 0)
	{
		fprintf(stderr, "Type \"%s -help\" for help.\n", pgm);
	}
	else
	{
		printf("usage: %s [options] [X-toolkit options] [cmd ...]\n", pgm);
		printf("options:\n");
		printf("  -X list   list of toolkit options "
			"(\"<opt>[=<arg>], ...\")\n");

		printf("\n");
		xvt_tk_help(0, stdout, level - 1, FALSE);

		printf("\n");
		printf("If no cmd is given, $SHELL (or \"%s\") will be used.\n",
			XVT_SHELL);

		printf("\n");
		printf("Type \"%s -help toolkit\"  for a list of toolkit options.\n",
			pgm);
		printf("Type \"%s -help pointers\" for a list of pointer types.\n",
			pgm);
		printf("Type \"%s -help colors\"   for a list of colors.\n",
			pgm);
	}
}

/*------------------------------------------------------------------------
 * main program
 */
int xvt_main (int argc, char **argv)
{
	XVT_DATA *		xd;
	const char *	pgm;
	const char *	p;
	int				c;
	int				rc;

	/*--------------------------------------------------------------------
	 * get program name (minus any directory part)
	 */
	pgm = xvt_path_basename(argv[0]);

	/*--------------------------------------------------------------------
	 * setup xvt-data struct
	 */
	xd = xvt_data_alloc();
	if (xd == 0)
	{
			fprintf(stderr, "%s: No memory\n", pgm);
			return (EXIT_FAILURE);
	}

	/*--------------------------------------------------------------------
	 * set initial rows & cols to env vars LINES & COLUMNS if found
	 */
	p = getenv("LINES");
	if (p != 0 && *p != 0)
		xvt_data_set_scr_rows(xd, atoi(p));

	p = getenv("COLUMNS");
	if (p != 0 && *p != 0)
		xvt_data_set_scr_cols(xd, atoi(p));

	p = getenv("TABSIZE");
	if (p != 0 && *p != 0)
		xvt_data_set_scr_tabs(xd, atoi(p));

	/*--------------------------------------------------------------------
	 * special case for -help (which does not fit getopt standard)
	 */
	if (argc > 1)
	{
		if (strcmp(argv[1], "-?")     == 0 ||
		    strcmp(argv[1], "-help")  == 0 ||
		    strcmp(argv[1], "--help") == 0)
		{
			if (argc > 2)
			{
				if (strcmp(argv[2], "toolkit") == 0)
				{
					xvt_tk_help(xd, stdout, 1, FALSE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}

				if (strcmp(argv[2], "pointers") == 0)
				{
					xvt_term_list_pointers(stdout, FALSE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}

				if (strcmp(argv[2], "colors") == 0)
				{
					xvt_rgb_dump(stdout, FALSE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}
			}

			xvt_main_usage(pgm, 1);
			return (EXIT_SUCCESS);
		}

		if (strcmp(argv[1], "-??")       == 0 ||
		    strcmp(argv[1], "-helpall")  == 0 ||
		    strcmp(argv[1], "--helpall") == 0)
		{
			if (argc > 2)
			{
				if (strcmp(argv[2], "toolkit") == 0)
				{
					xvt_tk_help(xd, stdout, 2, FALSE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}

				if (strcmp(argv[2], "pointers") == 0)
				{
					xvt_term_list_pointers(stdout, TRUE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}

				if (strcmp(argv[2], "colors") == 0)
				{
					xvt_rgb_dump(stdout, TRUE);
					xvt_data_free(xd);
					return (EXIT_SUCCESS);
				}
			}

			xvt_main_usage(pgm, 2);
			return (EXIT_SUCCESS);
		}

		if (strcmp(argv[1], "-toolkit")  == 0 ||
		    strcmp(argv[1], "--toolkit") == 0)
		{
			xvt_tk_help(xd, stdout, 2, FALSE);
			xvt_data_free(xd);
			return (EXIT_SUCCESS);
		}
	}

	/*--------------------------------------------------------------------
	 * extract X-toolkit options (which don't fit getopt() standard)
	 */
	rc = xvt_tk_args(xd, argc, argv);
	if (rc != XVT_OK)
	{
		fprintf(stderr, "%s: %s\n",
			pgm, xvt_data_get_error_buf(xd));
		xvt_data_free(xd);
		return (EXIT_FAILURE);
	}

	argc = xvt_data_get_argc(xd);
	argv = xvt_data_get_argv(xd);

	/*--------------------------------------------------------------------
	 * process cmd-line options
	 */
	while ((c = getopt(argc, argv, "X:")) != EOF)
	{
		switch (c)
		{
		case 'X':
			rc = xvt_tk_optlist(xd, optarg, FALSE);
			if (rc != XVT_OK)
			{
				fprintf(stderr, "%s: %s\n",
					pgm, xvt_data_get_error_buf(xd));
				xvt_data_free(xd);
				return (EXIT_FAILURE);
			}
			break;
			
		default:
			xvt_main_usage(pgm, 0);
			xvt_data_free(xd);
			return (EXIT_FAILURE);
		}
	}

	/*--------------------------------------------------------------------
	 * Now process any resource files
	 */
	xvt_tk_resource(xd);

	/*--------------------------------------------------------------------
	 * Finally, run the cmd.
	 */
	rc = xvt_task_run(xd, argv + optind);
	if (rc != XVT_OK)
	{
		fprintf(stderr, "%s: %s\n",
			pgm, xvt_data_get_error_buf(xd));
		xvt_data_free(xd);
		return (EXIT_FAILURE);
	}

	/*--------------------------------------------------------------------
	 * close out
	 */
	rc = xvt_data_get_exit_code(xd);
	xvt_data_free(xd);

	/*--------------------------------------------------------------------
	 * Execution was successful, so return the sub-tasks's exit-code.
	 */
	return (rc);
}
