/*------------------------------------------------------------------------
 * libxvt external procedures
 */
#include "xvtcommon.h"
#include "systhread.h"

/*------------------------------------------------------------------------
 * xvt data struct ptr
 *
 * We need this struct pointer to be globally available,
 * since the asyncronous procedure xvt_procs_child_rtn() needs access to it.
 * We limit knowledge of it, however, to those routines that need it.
 */
static TERMDATA *	xvt_procs_tdp	= 0;

/*------------------------------------------------------------------------
 * asynchronous signal routine for death of a child
 *
 * Note that there may be a race condition where the child finishes
 * before the parent even deals with the child.  Therefore, we wait for
 * *any* pid, since we can't insure that we know the subtask PID.
 *
 * Another problem here is that the PID we catch may be for a printer
 * subtask.  But, at least, we know that pid so we can deal with it.
 */
static void xvt_procs_child_rtn (int sig)
{
	TERMDATA *	td		= xvt_procs_tdp;
	int			pid;
	int			status;
	int			exit_code;

	/*--------------------------------------------------------------------
	 * wait for child
	 */
	while (TRUE)
	{
		status = 0;
		pid = waitpid((pid_t)-1, &status, 0);
		if (pid < 0 && errno == EINTR)
			continue;
		break;
	}

	exit_code = (status >> 8);

	/*--------------------------------------------------------------------
	 * check if this PID is for the printer
	 */
	if (td->prt != 0 && pid == td->prt->prt_pid)
	{
		td->prt->prt_rc = exit_code;
	}
	else
	{
		/*----------------------------------------------------------------
		 * clear PID of child & store exit code
		 */
		td->pid_died	= pid;
		td->pid			= 0;
		td->exit_code	= exit_code;
	}

	/*--------------------------------------------------------------------
	 * re-enable this routine
	 */
	signal(SIGCHLD, xvt_procs_child_rtn);
}

/*------------------------------------------------------------------------
 * close out everything
 */
static void xvt_procs_close (TERMDATA *td)
{
	if (td->init)
	{
		xvt_term_close_all(td);
	}

	xvt_mem_free(td);
}

/*------------------------------------------------------------------------
 * setup display structs
 */
static TERMDATA * xvt_procs_setup (XVT_DATA *xd, char **argv)
{
	TERMDATA *	td;
	int			rc;

	/*--------------------------------------------------------------------
	 * validate termmode if specified
	 */
	if (xd->termmode != 0)
	{
		char msgbuf[128];

		rc = xvt_tio_build(0, xd->termmode, FALSE, msgbuf);
		if (rc)
		{
			sprintf(xd->error_buf, "termmode: %s", msgbuf);
			xd->error_num = XVT_ERR_INVARG;
			return (0);
		}
	}

	/*--------------------------------------------------------------------
	 * validate any colors specified
	 */
	rc = xvt_rgb_check(xd);
	if (rc)
		return (0);

	/*--------------------------------------------------------------------
	 * initialize data struct
	 */
	td = (TERMDATA *)xvt_mem_alloc(sizeof(*td));
	if (td == 0)
	{
		sprintf(td->xd->error_buf, "No memory");
		td->xd->error_num = XVT_ERR_NOMEM;
		return (0);
	}

	memset(td, 0, sizeof(*td));

	td->our_pid = getpid();

	/*--------------------------------------------------------------------
	 * set all "unset" values in the user-data to default values
	 */
	xvt_tk_data_init(xd, TRUE);

	/*--------------------------------------------------------------------
	 * now do some sanity checks
	 */
	if (xd->scr_tabs >= xd->scr_cols)
		xd->scr_tabs = 0;

	if (xd->mbcols >= xd->scr_cols)
		xd->mbcols = (xd->scr_cols - 1);

	if (xd->fontno < 0 || xd->fontno >= XVT_NUM_FONTS)
	{
		xd->fontno = 0;
	}

	/*--------------------------------------------------------------------
	 * cache the program info into prog struct
	 */
	td->xd			= xd;
	td->argv		= argv;
	td->disp_info	= xd->lblinfo;

	/*--------------------------------------------------------------------
	 * clear all return values in user data
	 */
	xd->term_type[0]	= 0;
	xd->term_defs		= 0;

	xd->event_rtn		= 0;
	xd->event_data		= 0;

	xd->fd_inp			= -1;
	xd->fd_out			= -1;

	xd->cur_cols		= xd->scr_cols;
	xd->cur_rows		= xd->scr_rows;

	xd->exit_code		= 0;

	xd->error_num		= XVT_OK;
	xd->error_quit		= FALSE;
	xd->error_buf[0]	= 0;

	/*--------------------------------------------------------------------
	 * cache window label & set to cmd if not specified
	 */
	if (td->xd->win_label != 0)
		strcpy(td->win_label, td->xd->win_label);
	else if (td->xd->appname != 0)
		strcpy(td->win_label, td->xd->appname);

	if (*td->win_label == 0)
	{
		if (td->argv != 0 && td->argv[0] != 0)
		{
			strcpy(td->win_label, xvt_path_basename(td->argv[0]));
		}
		else
		{
			if (td->xd->resname != 0)
				strcpy(td->ico_label, td->xd->resname);
			else
				strcpy(td->win_label, XVT_NAME);
		}
	}

	strcpy(td->org_label, td->win_label);

	/*--------------------------------------------------------------------
	 * now process icon label
	 */
	if (td->xd->ico_label != 0)
		strcpy(td->ico_label, td->xd->ico_label);
	else if (td->xd->appname != 0)
		strcpy(td->ico_label, td->xd->appname);
	else
		strcpy(td->ico_label, td->win_label);

	/*--------------------------------------------------------------------
	 * open the display
	 */
	rc = xvt_term_setup_all(td);
	if (rc)
	{
		return (0);
	}

	/*--------------------------------------------------------------------
	 * set some initial states
	 */
	td->run_interval	= 5;			/* normal timeout in ms	*/
	td->die_interval	= 100;			/* dying  timeout in ms	*/


	/*--------------------------------------------------------------------
	 * set DISPLAY var string if specified
	 */
	if (xd->display != 0 && *xd->display != 0)
	{
		sprintf(td->dsp_var_str, "DISPLAY=%s", xd->display);
	}

	/*--------------------------------------------------------------------
	 * set env var string if specified
	 */
	if (xd->envvar != 0 && *xd->envvar != 0)
	{
		sprintf(td->env_var_str, "%s=%s",
			xd->envvar, xd->term_type);
	}

	/*--------------------------------------------------------------------
	 * Set TERM variable string
	 */
	sprintf(td->trm_var_str, "TERM=%s", xd->termname);

	/*--------------------------------------------------------------------
	 * Set WINDOWID variable string
	 */
	sprintf(td->wid_var_str, "WINDOWID=%u", xvt_w_get_windowid(td->tw));

	/*--------------------------------------------------------------------
	 * Open pseudo-tty master
	 */
	td->pd = xvt_pty_init(xd->ptyspec);
	if (td->pd == 0)
	{
		sprintf(td->xd->error_buf, "Cannot open pseudo-tty");
		xvt_procs_close(td);

		td->xd->error_num = XVT_ERR_CANTOPENPTYS;
		return (0);
	}

	/*--------------------------------------------------------------------
	 * indicate we are initialized
	 */
	td->init = TRUE;

	return (td);
}

/*------------------------------------------------------------------------
 * exec routine
 */
static int xvt_procs_exec (TERMDATA *td, int need_login)
{
	int	pid;

	/*--------------------------------------------------------------------
	 * setup signal catchers:
	 * cache our data pointer & enable catching of child death.
	 */
	xvt_procs_tdp = td;
	signal(SIGCHLD, xvt_procs_child_rtn);

	/*--------------------------------------------------------------------
	 * Execute child process (this returns as the parent)
	 */
	pid = xvt_child_setup(td, need_login);
	if (pid < 0)
	{
		return (td->xd->error_num);
	}
	td->pid	= pid;

	/*--------------------------------------------------------------------
	 * setup signal catchers:
	 * We want to ignore these signals so they are just passed down
	 * to the child.
	 */
	signal(SIGINT,  SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	/*--------------------------------------------------------------------
	 * open the slave pty so we can report window changes
	 */
#if 0	/* DEFERED until resize happens */
	xvt_pty_open(td->pd, FALSE, td->xd->termmode);
#endif

	/*--------------------------------------------------------------------
	 * now enter main loop
	 */
	xvt_term_main_loop(td);

	/*--------------------------------------------------------------------
	 * reset signal handlers
	 */
	signal(SIGCHLD, SIG_DFL);
	signal(SIGINT,  SIG_DFL);
	signal(SIGQUIT, SIG_DFL);

	/*--------------------------------------------------------------------
	 * We get here by any of the following:
	 *
	 *	1.	The child died.
	 *	2.	We got an EOF reading child output.
	 *	3.	The WM told us to quit.
	 *
	 * If child is still alive, terminate him with extreme prejudice.
	 * Note that we already gave up on waiting for him by resetting
	 * our SIGCHLD catcher.
	 */
	if (td->pid_died != 0 &&
	    td->pid != 0 &&
	    td->pid != td->our_pid &&
	    td->pid != (pid_t)-1)
	{
		kill(td->pid, SIGKILL);
		td->pid = 0;
	}

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * detach a task
 *
 * returns:
 *		-1	error
 *		 0	child  return
 *		>0	parent return (rc == pid of child)
 */
static int xvt_procs_detach (TERMDATA *td)
{
	int	pid;

	/*--------------------------------------------------------------------
	 * Check if no-detach option specified. This is for debugging.
	 */
	if (getenv("XVT_NO_DETACH") != 0)
	{
		return (0);
	}

	/*--------------------------------------------------------------------
	 * Detach ourselves by forking and having the parent return.
	 * All the other happy BS needed to become a proper daemon
	 * (controlling terminals, etc.) is dealt with when we open the slave pty.
	 */
	if (getenv("XVT_NO_THREADS") != 0)
		pid = fork();
	else
		pid = systhread_fork();

	switch (pid)
	{
	case -1:		/* error */
		td->xd->error_num = XVT_ERR_CANTFORK;
		sprintf(td->xd->error_buf, "Cannot fork");
		return (-1);

	case 0:			/* in child */
		td->our_pid = getpid();
		break;

	default:		/* in parent */
		return (pid);
	}

	return (0);
}

/*------------------------------------------------------------------------
 * thread routine
 */
static int xvt_procs_thread (TERMDATA *td)
{
	THREAD_ID tid;

	/*--------------------------------------------------------------------
	 * This tells the main loop that there is a valid process,
	 * but it should not be messed with.
	 */
	td->pid = getpid();

	/*--------------------------------------------------------------------
	 * spawn off main-loop in a thread
	 */
	if (getenv("XVT_NO_THREADS") != 0)
	{
		tid = 0;
	}
	else
	{
		tid = systhread_begin(xvt_term_main_loop, td);
	}
	td->our_tid = tid;
	
	return (tid > 0 ? 0 : -1);
}

/*------------------------------------------------------------------------
 * run one xvt event loop
 */
static int xvt_procs_run_once (void *data, int ms)
{
	TERMDATA *	td	= (TERMDATA *)data;
	int			rc;

	/*--------------------------------------------------------------------
	 * validate that we are initialized
	 */
	if (! td->init)
	{
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * run the main loop
	 */
	rc = xvt_term_loop_once(td, ms);

	return (rc);
}

/*------------------------------------------------------------------------
 * run xvt in "here" mode
 */
int xvt_here_run (XVT_DATA *xd, int detach)
{
	TERMDATA *	td;
	int			rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
	{
		return (XVT_ERR_NULLDATA);
	}

	/*--------------------------------------------------------------------
	 * setup all structs
	 */
	td = xvt_procs_setup(xd, 0);
	if (td == 0)
	{
		return (xd->error_num);
	}

	/*--------------------------------------------------------------------
	 * If we are detaching, first check all debug streams and
	 * close any which point to stderr.
	 */
	if (detach)
	{
		if (xd->events_fp == stderr)	xd->events_fp = 0;
		if (xd->action_fp == stderr)	xd->action_fp = 0;
		if (xd->output_fp == stderr)	xd->output_fp = 0;
		if (xd->screen_fp == stderr)	xd->screen_fp = 0;
		if (xd->inpkbd_fp == stderr)	xd->inpkbd_fp = 0;
		if (xd->keybrd_fp == stderr)	xd->keybrd_fp = 0;
		if (xd->resfil_fp == stderr)	xd->resfil_fp = 0;
		if (xd->tkopts_fp == stderr)	xd->tkopts_fp = 0;
	}

	/*--------------------------------------------------------------------
	 * detach from foreground if requested
	 */
	if (detach)
	{
		rc = xvt_procs_detach(td);
		if (rc < 0)
		{
			xvt_procs_close(td);
			return (XVT_ERR_CANTFORK);
		}

		if (rc > 0)
		{
			/*------------------------------------------------------------
			 * parent return - we just exit
			 */
			_exit(EXIT_SUCCESS);
		}

		td->detached = TRUE;
	}

	/*--------------------------------------------------------------------
	 * Now open our slave pty.
	 * This pty will become our new controlling terminal.
	 */
	rc = xvt_pty_open(td->pd, td->detached, td->xd->termmode);
	if (rc)
	{
		sprintf(td->xd->error_buf, "Cannot open pseudo-tty");
		td->xd->error_num = XVT_ERR_CANTOPENPTYS;
		xd->error_quit = TRUE;
		xvt_procs_close(td);
		return (XVT_ERR_CANTOPENPTYS);
	}

	/*--------------------------------------------------------------------
	 * set our env variables
	 *
	 * We do this only after we are sure that initialization was successful,
	 * since otherwise we still have to use the original term type.
	 */
	if (*td->dsp_var_str != 0)
		putenv(td->dsp_var_str);
	putenv(td->trm_var_str);
	putenv(td->wid_var_str);

	/*--------------------------------------------------------------------
	 * unset TERMCAP env variable if present, since it may be incorrect
	 */
	putenv("TERMCAP=");

	/*--------------------------------------------------------------------
	 * set window size
	 */
	xvt_pty_resize(td->pd, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td), 0);

	/*--------------------------------------------------------------------
	 * start thread for X11 process
	 */
	rc = xvt_procs_thread(td);

	/*--------------------------------------------------------------------
	 * set return values in prog struct
	 *
	 * Note that we always set the event data even though there may not
	 * be an event routine, since that way the caller can determine if he
	 * is in an "event-rtn" type of environment, even though there is
	 * no actual event routine to be called.
	 * Also, xvt_close_here() & xvt_clone() get the data pointer from there.
	 */
	xd->event_rtn	= rc ? xvt_procs_run_once : 0;
	xd->event_data	= td;

	if (td->detached)
	{
		xd->fd_inp	= fileno(stdin);
		xd->fd_out	= fileno(stdout);
	}
	else
	{
		xd->fd_inp	= td->pd->slv_fd;
		xd->fd_out	= td->pd->slv_fd;
	}

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * close out from here mode
 */
int xvt_here_close (XVT_DATA *xd)
{
	TERMDATA *	td;

	/*--------------------------------------------------------------------
	 * check if ever initialized
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	td = (TERMDATA *)xd->event_data;
	if (td == 0 || ! td->init)
	{
		sprintf(xd->error_buf, "Not initialized");
		xd->error_num = XVT_ERR_NOTINIT;
		return (XVT_ERR_NOTINIT);
	}

	/*--------------------------------------------------------------------
	 * If we have a thread running, we have to stop it.
	 */
	if (td->our_tid != 0)
	{
		td->bail = XVT_DIE_LOOPS;
		systhread_wait(td->our_tid);
	}

	/*--------------------------------------------------------------------
	 * free all allocated resources
	 */
	xvt_procs_close(td);

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * run xvt in "task" mode
 */
int xvt_task_run (XVT_DATA *xd, char **argv)
{
	TERMDATA *	td;
	char		fullpath[256];
	char *		temp_argv[2];
	int			need_login	= FALSE;
	int			rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	/*--------------------------------------------------------------------
	 * check if any command specified
	 */
	if (argv == 0 || *argv == 0)
	{
		const char *	env;

		/*----------------------------------------------------------------
		 * no command specified, just use shell
		 */
		env = getenv("SHELL");
		if (env == 0)
			env = XVT_SHELL;

		temp_argv[0] = (char *)env;
		temp_argv[1] = 0;

		argv = temp_argv;
		need_login = xd->login;
	}

	/*--------------------------------------------------------------------
	 * validate the cmd to execute
	 */
	rc = xvt_path_find(fullpath, *argv);
	if (rc)
	{
		xd->error_num = XVT_ERR_INVARG;
		sprintf(xd->error_buf, "Filename \"%s\" not in path", *argv);
		return (xd->error_num);
	}

	rc = access(fullpath, X_OK);
	if (rc)
	{
		xd->error_num = XVT_ERR_CANTEXEC;
		sprintf(xd->error_buf, "Filename \"%s\" not executable", *argv);
		return (xd->error_num);
	}

	/*--------------------------------------------------------------------
	 * do all the setup
	 */
	td = xvt_procs_setup(xd, argv);
	if (td == 0)
		return (xd->error_num);

	/*--------------------------------------------------------------------
	 * set window title again if label-info wanted
	 */
	if (td->disp_info)
	{
		xvt_term_set_win_title(td, td->win_label);
	}

	/*--------------------------------------------------------------------
	 * run the main loop
	 */
	rc = xvt_procs_exec(td, need_login);

	/*--------------------------------------------------------------------
	 * store exit code back
	 */
	xd->exit_code = td->exit_code;

	/*--------------------------------------------------------------------
	 * free all allocated resources
	 */
	xvt_procs_close(td);

	return (rc);
}

/*------------------------------------------------------------------------
 * create a clone of the current task
 */
int xvt_here_clone (XVT_DATA *xd)
{
	TERMDATA *	td;
	TTY_STRUCT	tty;
	int			rc;

	/*--------------------------------------------------------------------
	 * can't clone if not initialized
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	td = (TERMDATA *)xd->event_data;
	if (td == 0 || ! td->init)
	{
		sprintf(xd->error_buf, "Not initialized");
		xd->error_num = XVT_ERR_NOTINIT;
		return (XVT_ERR_NOTINIT);
	}

	/*--------------------------------------------------------------------
	 * first thing is to detach ourselves
	 */
	rc = xvt_procs_detach(td);
	if (rc < 0)
	{
		return (XVT_ERR_CANTFORK);
	}

	/*--------------------------------------------------------------------
	 * Turn on PIDs in event log so we can identify which process an entry
	 * is from.
	 */
	td->log_pid = TRUE;

	/*--------------------------------------------------------------------
	 * The parent can just return & keep going.
	 * The child does all the work.
	 */
	if (rc > 0)
	{
		int pid = rc;
		int status;

		/*----------------------------------------------------------------
		 * The parent waits for child to die, since the child is
		 * detaching again (this way the parent has no zombies attached).
		 * Note that we totally ignore any exit status.
		 */
		while (TRUE)
		{
			rc = waitpid(pid, &status, 0);
			if (rc < 0 && errno == EINTR)
				continue;
			break;
		}
		
		return (XVT_OK);
	}

	/*--------------------------------------------------------------------
	 * We are now a child process, so we will either sucessfully restart
	 * or we will exit.
	 *
	 * At this point, we are a proper sub-task of the original program.
	 * So, we detach again & let our parent continue on.
	 *
	 * But first, we do any shutdown needed.
	 */
	{
		/*----------------------------------------------------------------
		 * cache current tty state
		 */
		xvt_pty_getmode(td->pd, &tty);

		/*----------------------------------------------------------------
		 * Now do any shutdown needed
		 */
		xvt_term_shutdown(td);

		/*----------------------------------------------------------------
		 * Now we detach again
		 */
		rc = xvt_procs_detach(td);
		if (rc != 0)
		{
			_exit(EXIT_SUCCESS);
		}
	}

	/*--------------------------------------------------------------------
	 * We *should* be at this point a true daemon. Now open a new window.
	 */
	rc = xvt_term_restart(td);
	if (rc)
	{
		_exit(EXIT_FAILURE);
	}

	/*--------------------------------------------------------------------
	 * Now open our slave pty.
	 *
	 * Note that after here we are no longer a *true* daemon, since
	 * the pseudo-tty we open will become a controlling terminal.
	 */
	rc = xvt_pty_open(td->pd, td->detached, td->xd->termmode);
	if (rc)
	{
		_exit(EXIT_FAILURE);
	}

	/*--------------------------------------------------------------------
	 * reset tty state to that of our parent
	 */
	xvt_pty_setmode(td->pd, &tty);

	/*--------------------------------------------------------------------
	 * close any pending kbd states
	 */
	xvt_term_force_kbd_state(td, 0);

	/*--------------------------------------------------------------------
	 * set window & icon labels
	 */
	xvt_term_set_win_title(td, td->win_label);
	xvt_term_set_ico_title(td, td->ico_label);

	/*--------------------------------------------------------------------
	 * set window size
	 */
	xvt_pty_resize(td->pd, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td), 0);

	/*--------------------------------------------------------------------
	 * Start thread for X11 process.
	 * Note that we ignore the return code, since we assume that if it
	 * succeeded last time, it will succeed this time and vice-versa.
	 */
	xvt_procs_thread(td);

	/*--------------------------------------------------------------------
	 * Note that we don't store new data in the xvt_data struct, because
	 * nothing has changed.
	 */
	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * query whether XVT is available
 */
int xvt_available (void)
{
	return (TRUE);
}
