/*------------------------------------------------------------------------
 * xterm clone logic routines
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * key state masks
 */
static const int xvt_term_key_masks[] =
{
	KEY_MASK_SHFT,
	KEY_MASK_CTRL,
	KEY_MASK_META,
	KEY_MASK_NUML
};

/*------------------------------------------------------------------------
 * free screen arrays
 */
static void xvt_term_close_text (TERMDATA *td)
{
	td->reg_scrn = xvt_scrn_free(td->reg_scrn);
	td->alt_scrn = xvt_scrn_free(td->alt_scrn);
	td->cur_scrn = 0;
}

/*------------------------------------------------------------------------
 * create screen arrays
 */
static void xvt_term_setup_text (TERMDATA *td, int w, int h)
{
	TERMSCRN *	old_reg	= td->reg_scrn;
	TERMSCRN *	old_alt	= td->alt_scrn;
	TERMSCRN *	old_cur	= td->cur_scrn;
	TERMSCRN *	new_reg;
	TERMSCRN *	new_alt;
	int			mode;
	int			fg;
	int			bg;
	int			at;

	/*--------------------------------------------------------------------
	 * adjust current position if outside of new window
	 */
	if (old_cur != 0)
	{
		if (CUR_SCRN_X(td) >= w)
			CUR_SCRN_X(td) = w-1;

		if (CUR_SCRN_Y(td) >= h)
			CUR_SCRN_Y(td) = h-1;
	}

	/*--------------------------------------------------------------------
	 * allocate new reg screen array
	 */
	if (old_reg == 0)
	{
		mode	= td->scrn_mode;
		fg		= XVT_CLR_FG;
		bg		= XVT_CLR_BG;
		at		= NORMAL;
	}
	else
	{
		mode	= SCRN_MODE(old_reg);
		fg		= SCRN_FG(old_reg);
		bg		= SCRN_BG(old_reg);
		at		= SCRN_AT(old_reg);
	}
	new_reg	= xvt_scrn_init(h, w, td->xd->scr_tabs, mode, fg, bg, at);

	/*--------------------------------------------------------------------
	 * allocate new alt screen array
	 */
	if (old_alt == 0)
	{
		mode	= td->scrn_mode;
		fg		= XVT_CLR_FG;
		bg		= XVT_CLR_BG;
		at		= NORMAL;
	}
	else
	{
		mode	= SCRN_MODE(old_alt);
		fg		= SCRN_FG(old_alt);
		bg		= SCRN_BG(old_alt);
		at		= SCRN_AT(old_alt);
	}
	new_alt	= xvt_scrn_init(h, w, td->xd->scr_tabs, mode, fg, bg, at);

	/*--------------------------------------------------------------------
	 * now copy over screen contents
	 */
	xvt_scrn_copy(new_reg, old_reg);
	xvt_scrn_copy(new_alt, old_alt);

	/*--------------------------------------------------------------------
	 * free the old screens
	 */
	xvt_scrn_free(old_reg);
	xvt_scrn_free(old_alt);

	/*--------------------------------------------------------------------
	 * store new screens
	 */
	td->reg_scrn = new_reg;
	td->alt_scrn = new_alt;
	td->cur_scrn = (old_cur == old_reg ? new_reg : new_alt);

	/*--------------------------------------------------------------------
	 * finally, set new size in user data
	 */
	td->xd->cur_cols = w;
	td->xd->cur_rows = h;
}

/*------------------------------------------------------------------------
 * set reg font
 *
 * If we are in color & we have a bold font defined, use it.
 * Otherwise, use the reg font (which is always there).
 */
static void xvt_term_set_reg_font (TERMDATA *td)
{
	int font_type = (td->scrn_mode == SCRN_MODE_COLOR ? FONT_BOLD : FONT_NORM);

	xvt_w_set_font_type(td->tw, font_type);
}

/*------------------------------------------------------------------------
 * write a key sequence to the child
 */
static void xvt_term_write_keys (TERMDATA *td, const char *p)
{
	int l;

	/*--------------------------------------------------------------------
	 * check that string length > 0
	 */
	l = strlen(p);
	if (l == 0)
		return;

	/*--------------------------------------------------------------------
	 * keybrd debugging
	 */
	xvt_log_keybrd_str(td, p);

	/*--------------------------------------------------------------------
	 * now send it
	 */
	write(td->pd->mst_fd, p, l);
}

/*------------------------------------------------------------------------
 * send data to the program (for reporting)
 */
void xvt_term_write_report (TERMDATA *td, const char *fmt, ...)
{
	char	buffer[128];
	va_list	args;

	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);

	xvt_term_write_keys(td, buffer);
}

/*------------------------------------------------------------------------
 * set user's idea of the kbd state to what we think it is
 */
void xvt_term_set_kbd_state (TERMDATA *td, int it_is)
{
	int diff;

	/*--------------------------------------------------------------------
	 * Check if we need to get real kbd state
	 */
	if (it_is == -1)
		it_is = xvt_w_query_kbd(td->tw);

	/*--------------------------------------------------------------------
	 * Only output this if the kbd is "mapped" for these keys
	 */
	if (td->map_kbd > 1)
	{
		/*----------------------------------------------------------------
		 * Now send necessary strings
		 */
		diff  = (td->tw->keys ^ it_is);
		if (diff != 0)
		{
			/*------------------------------------------------------------
			 * check shift key
			 */
			if (diff & KEY_MASK_SHFT)
			{
				if ((it_is & KEY_MASK_SHFT) != 0)
				{
					xvt_term_write_keys(td, td->keys_st_dn[XVT_KEY_SHFT].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_dn[XVT_KEY_SHFT].name);
				}
				else
				{
					xvt_term_write_keys(td, td->keys_st_up[XVT_KEY_SHFT].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_up[XVT_KEY_SHFT].name);
				}
			}

			/*------------------------------------------------------------
			 * check ctrl key
			 */
			if (diff & KEY_MASK_CTRL)
			{
				if ((it_is & KEY_MASK_CTRL) != 0)
				{
					xvt_term_write_keys(td, td->keys_st_dn[XVT_KEY_CTRL].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_dn[XVT_KEY_CTRL].name);
				}
				else
				{
					xvt_term_write_keys(td, td->keys_st_up[XVT_KEY_CTRL].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_up[XVT_KEY_CTRL].name);
				}
			}

			/*------------------------------------------------------------
			 * check meta key
			 */
			if (diff & KEY_MASK_META)
			{
				if ((it_is & KEY_MASK_META) != 0)
				{
					xvt_term_write_keys(td, td->keys_st_dn[XVT_KEY_META].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_dn[XVT_KEY_META].name);
				}
				else
				{
					xvt_term_write_keys(td, td->keys_st_up[XVT_KEY_META].send);
					xvt_log_inpkbd_fmt (td, td->keys_st_up[XVT_KEY_META].name);
				}
			}
		}
	}

	td->tw->keys = it_is;
}

/*------------------------------------------------------------------------
 * set user's idea of the kbd state to what we think it should be
 */
void xvt_term_force_kbd_state (TERMDATA *td, int it_is)
{
	td->tw->keys = ~it_is;
	xvt_term_set_kbd_state(td, it_is);
}

/*------------------------------------------------------------------------
 * switch between color & mono if possible
 */
void xvt_term_set_scrn_mode (TERMDATA *td, int mode)
{
	/*--------------------------------------------------------------------
	 * check if this is needed
	 */
	if (td->scrn_mode == mode)
		return;

	/*--------------------------------------------------------------------
	 * set to requested mode if possible
	 */
	td->scrn_mode = xvt_w_set_scrn_mode(td->tw, mode);

	/*--------------------------------------------------------------------
	 * set our reg font, since we use bold for color if we have it.
	 */
	xvt_term_set_reg_font(td);

	/*--------------------------------------------------------------------
	 * remember that xvt_w_set_scrn_mode() always sets cur_fg/bg
	 * to CLR_FG/BG (but doesn't issue the X call).
	 */
	if (CUR_SCRN(td) != 0)
	{
		CUR_SCRN_AT(td)		= NORMAL;
		CUR_SCRN_MODE(td)	= td->scrn_mode;

		if (td->scrn_mode == SCRN_MODE_COLOR)
		{
			td->tw->cur_fg = CUR_SCRN_FG(td);
			td->tw->cur_bg = CUR_SCRN_BG(td);
		}
		else
		{
			CUR_SCRN_FG(td) = td->tw->cur_fg;
			CUR_SCRN_BG(td) = td->tw->cur_bg;
		}
	}

	/*--------------------------------------------------------------------
	 * now force our fg & bg colors
	 */
	xvt_w_set_fg(td->tw, td->tw->cur_fg, TRUE);
	xvt_w_set_bg(td->tw, td->tw->cur_bg, TRUE);
}

/*------------------------------------------------------------------------
 * low level output a character
 */
void xvt_term_disp_char (TERMDATA *td, int x, int y, int ch,
	int fg, int bg, int at)
{
	int f;
	int b;

	/*--------------------------------------------------------------------
	 * Check if we are in the wrong color mode
	 */
	if (fg < XVT_NUM_COLORS && bg < XVT_NUM_COLORS)
	{
		if (td->scrn_mode == SCRN_MODE_MONO)
			xvt_term_set_scrn_mode(td, SCRN_MODE_COLOR);
	}
	else if (bg != XVT_CLR_CR)
	{
		if (td->scrn_mode == SCRN_MODE_COLOR)
			xvt_term_set_scrn_mode(td, SCRN_MODE_MONO);
	}

	/*--------------------------------------------------------------------
	 * If in color, use attributes to adjust color
	 */
	if (td->scrn_mode == SCRN_MODE_COLOR || td->tw->cr_color != -1)
	{
		if (at & BOLD)
		{
			fg |= 8;
			at &= ~BOLD;
		}
		xvt_w_set_fg(td->tw, fg, FALSE);

		if (at & BLINK)
		{
			bg |= 8;
			at &= ~BLINK;
		}
		xvt_w_set_bg(td->tw, bg, FALSE);

		at &= ~UNDERLINE;
	}

	/*--------------------------------------------------------------------
	 * Draw char according to mono attributes
	 */
	switch (at)
	{
	case NORMAL:
		/*----------------------------------------------------------------
		 * draw a normal char
		 */
		xvt_w_char_disp(td->tw, x, y, ch);
		break;

	case BOLD:
	case BLINK:
		/*----------------------------------------------------------------
		 * draw a bold char
		 */
		xvt_w_char_bold(td->tw, x, y, ch);
		break;

	case UNDERLINE:
		/*----------------------------------------------------------------
		 * draw a underlined char
		 */
		xvt_w_char_disp(td->tw, x, y, ch);
		xvt_w_char_line(td->tw, x, y);
		break;

	case REVERSE:
		/*----------------------------------------------------------------
		 * draw a char in reverse
		 */
		f = td->tw->cur_fg;
		b = td->tw->cur_bg;
		xvt_w_set_fg(td->tw, b, FALSE);
		xvt_w_set_bg(td->tw, f, FALSE);
		xvt_w_char_disp(td->tw, x, y, ch);
		xvt_w_set_fg(td->tw, f, FALSE);
		xvt_w_set_bg(td->tw, b, FALSE);
		break;
	}
}

/*------------------------------------------------------------------------
 * draw box for the cursor
 */
static void xvt_term_cursor_box (TERMDATA *td)
{
	CELL *	cp = CUR_SCRN_CHAR(td, CUR_SCRN_Y(td), CUR_SCRN_X(td));
	int		x = CUR_SCRN_X(td);
	int		y = CUR_SCRN_Y(td);
	int		cb;

	if (td->tw->cr_color != -1)
	{
		cb = XVT_CLR_CR;
	}
	else
	{
		if (td->scrn_mode == SCRN_MODE_COLOR)
			cb = cp->fg;
		else
			cb = (cp->at == REVERSE ? XVT_CLR_BG : XVT_CLR_FG);
	}

	xvt_w_set_fg(td->tw, cb, FALSE);
	xvt_w_char_rect(td->tw, x, y);
}

/*------------------------------------------------------------------------
 * turn cursor display on/offoff
 */
void xvt_term_set_txt_cursor (TERMDATA *td, int state)
{
	CELL *	cp;
	int		old_mode	= CUR_SCRN_MODE(td);
	int		old_fg		= CUR_SCRN_FG(td);
	int		old_bg		= CUR_SCRN_BG(td);
	int		changed		= FALSE;

	if (state)
	{
		/*----------------------------------------------------------------
		 * turn cursor on
		 */
		if (CUR_SCRN_CSV(td) && ! CUR_SCRN_CON(td))
		{
			CUR_SCRN_CON(td) = TRUE;

			if (td->cursor_mode || td->xd->ahcursor)
			{
				int	fg;
				int	bg;
				int	at;

				cp = CUR_SCRN_CHAR(td, CUR_SCRN_Y(td), CUR_SCRN_X(td));

				if (td->tw->cr_color == -1)
				{
					fg = cp->fg;
					bg = cp->bg;
					at = cp->at & REVERSE ?
						cp->at & ~REVERSE :
						cp->at |  REVERSE;
				}
				else
				{
					fg = XVT_CLR_BG;
					bg = XVT_CLR_CR;
					at = NORMAL;
				}

				xvt_term_disp_char(td, CUR_SCRN_X(td), CUR_SCRN_Y(td),
					cp->ch, fg, bg, at);
			}
			else
			{
				xvt_term_cursor_box(td);
			}

			changed = TRUE;
		}
	}
	else
	{
		/*----------------------------------------------------------------
		 * turn cursor off
		 */
		if (CUR_SCRN_CSV(td) && CUR_SCRN_CON(td))
		{
			CUR_SCRN_CON(td) = FALSE;

			cp = CUR_SCRN_CHAR(td, CUR_SCRN_Y(td), CUR_SCRN_X(td));
			xvt_term_disp_char(td, CUR_SCRN_X(td), CUR_SCRN_Y(td), cp->ch,
				cp->fg, cp->bg, cp->at);
			changed = TRUE;
		}
	}

	/*--------------------------------------------------------------------
	 * if any change made, restore color state
	 */
	if (changed)
	{
		CUR_SCRN_FG(td) = old_fg;
		CUR_SCRN_BG(td) = old_bg;
		xvt_term_set_scrn_mode(td, old_mode);
	}
}

/*------------------------------------------------------------------------
 * set X11 cursor type
 */
void xvt_term_set_x11_cursor (TERMDATA *td, int shape)
{
	xvt_w_set_cursor(td->tw, shape);
}

/*------------------------------------------------------------------------
 * fill a rectangle with current attributes
 */
void xvt_term_disp_fill (TERMDATA *td, int dx, int dy, int dw, int dh)
{
	/*--------------------------------------------------------------------
	 * set color if needed
	 */
	xvt_w_set_fg(td->tw, CUR_SCRN_FG(td), FALSE);
	xvt_w_set_bg(td->tw, CUR_SCRN_BG(td), FALSE);

	/*--------------------------------------------------------------------
	 * fill in requested area
	 */
	xvt_w_rect_fill(td->tw, dx, dy, dw, dh);

	/*--------------------------------------------------------------------
	 * check if we drew over the cursor
	 */
	if (CUR_SCRN_X(td) >= dx && CUR_SCRN_X(td) <= (dx + dw) &&
		CUR_SCRN_Y(td) >= dy && CUR_SCRN_Y(td) <= (dy + dh))
	{
		CUR_SCRN_CON(td) = FALSE;
	}
}

/*------------------------------------------------------------------------
 * redraw a specified rectangle
 */
static void xvt_term_redraw_proc (TERMDATA *td, int s_x, int s_y,
	int w_x, int w_y, int invert)
{
	int x, y;
	int c_x, c_y;

	/*--------------------------------------------------------------------
	 * redraw the specified rectangle of chars
	 */
	c_y = s_y;
	for (y=0; y<w_y; y++)
	{
		CELL *	cp	= CUR_SCRN_LINE(td, c_y);

		c_x = s_x;
		for (x=0; x<w_x; x++)
		{
			if (invert)
			{
				xvt_term_disp_char(td, c_x, c_y, cp[c_x].ch,
					cp[c_x].fg, cp[c_x].bg,
					cp[c_x].at & REVERSE ?
						cp[c_x].at & ~REVERSE : cp[c_x].at | REVERSE);
			}
			else
			{
				xvt_term_disp_char(td, c_x, c_y, cp[c_x].ch,
					cp[c_x].fg, cp[c_x].bg, cp[c_x].at);
			}

			c_x++;
		}

		c_y++;
	}    

	/*--------------------------------------------------------------------
	 * check if cursor needs to be redisplayed
	 */
	if (CUR_SCRN_X(td) >= s_x && CUR_SCRN_X(td) < (s_x + w_x) &&
		CUR_SCRN_Y(td) >= s_y && CUR_SCRN_Y(td) < (s_y + w_y))
	{
		if (CUR_SCRN_CSV(td))
		{
			CUR_SCRN_CON(td) = FALSE;
			xvt_term_set_txt_cursor(td, TRUE);
		}
	}
}

/*------------------------------------------------------------------------
 * screen redraw CB routine
 */
static void xvt_term_repaint_proc (TERMDATA *td, int x, int y, int w, int h)
{
	if (x + w > CUR_SCRN_COLS(td))
		w = CUR_SCRN_COLS(td) - x;

	if (y + h > CUR_SCRN_ROWS(td))
		h = CUR_SCRN_ROWS(td) - y;

	if (! td->tw->doing_resize)
		xvt_term_redraw_proc(td, x, y, w, h, FALSE);
}

/*------------------------------------------------------------------------
 * repaint procedure
 */
void xvt_term_repaint (TERMDATA *td)
{
	xvt_term_repaint_proc(td, 0, 0, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td));
}

/*------------------------------------------------------------------------
 * screen interface routines
 */
void xvt_term_raise (TERMDATA *td)
{
	xvt_w_raise(td->tw);
}

void xvt_term_lower (TERMDATA *td)
{
	xvt_w_lower(td->tw);
}

void xvt_term_get_size (TERMDATA *td, int *w, int *h)
{
	xvt_w_get_size(td->tw, w, h);
}

void xvt_term_set_size (TERMDATA *td, int  w, int  h)
{
	xvt_w_set_size(td->tw, w, h);
}

void xvt_term_get_position (TERMDATA *td, int *x, int *y)
{
	xvt_w_get_position(td->tw, x, y);
}

void xvt_term_set_position (TERMDATA *td, int  x, int  y)
{
	xvt_w_set_position(td->tw, x, y);
}

void xvt_term_resize (TERMDATA *td, int cols, int rows)
{
	xvt_w_resize(td->tw, cols, rows);
}

/*------------------------------------------------------------------------
 * screen flash routine
 */
void xvt_term_flash (TERMDATA *td)
{
	xvt_term_redraw_proc(td, 0, 0, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td),
		TRUE);
	xvt_term_redraw_proc(td, 0, 0, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td),
		FALSE);
}

/*------------------------------------------------------------------------
 * low-level area copy routine
 */
void xvt_term_disp_copy (TERMDATA *td, int dx, int dy, int dw, int dh,
	int sx, int sy)
{
	xvt_w_rect_copy(td->tw, dx, dy, dw, dh, sx, sy);
}

/*------------------------------------------------------------------------
 * print the screen
 */
void xvt_term_screen_print (TERMDATA *td)
{
	xvt_term_printer_open(td);
	if (td->prt != 0)
	{
		xvt_scrn_print(CUR_SCRN(td), td->prt);
		xvt_term_printer_close(td);
	}
}

/*------------------------------------------------------------------------
 * copy the screen to the clipboard
 */
void xvt_term_screen_clipboard (TERMDATA *td)
{
	unsigned char *	scr_cb_data;
	int				scr_cb_len;
	unsigned char *	s;
	int				x;
	int				y;

	scr_cb_len  = (CUR_SCRN_ROWS(td) * (CUR_SCRN_COLS(td) + 1)) + 1;
	scr_cb_data = (unsigned char *)xvt_mem_alloc(scr_cb_len);
	if (scr_cb_data == 0)
		return;

	s = scr_cb_data;
	for (y=0; y < CUR_SCRN_ROWS(td); y++)
	{
		CELL *	cp = CUR_SCRN_LINE(td, y);

		for (x=0; x < CUR_SCRN_COLS(td); x++)
		{
			*s++ = (cp++)->ch;
		}
		*s++ = '\n';
	}
	*s = 0;

	xvt_w_clipboard_copy_to(td->tw, scr_cb_data, scr_cb_len);
	xvt_mem_free(scr_cb_data);
}

/*------------------------------------------------------------------------
 * process a screen-resize message
 */
static void xvt_term_resize_proc (TERMDATA *td, int w, int h)
{
	/*--------------------------------------------------------------------
	 * check if window changed size
	 */
	if (td->pid == 0 || (w != CUR_SCRN_COLS(td) || h != CUR_SCRN_ROWS(td)))
	{
		/*----------------------------------------------------------------
		 * create new screen array & copy old contents to it
		 */
		xvt_term_setup_text(td, w, h);

		/*----------------------------------------------------------------
		 * change size of pseudo-tty
		 */
		xvt_pty_resize(td->pd, w, h, td->pid);

		/*----------------------------------------------------------------
		 * debug
		 */
		xvt_log_events_fmt(td, "events", "new screen (%d,%d)", w, h);

		if (td->pid != 0)
		{
			xvt_log_events_fmt(td, "events", "sending SIGWINCH to %d",
				td->pid);
		}
	}

	/*--------------------------------------------------------------------
	 * check cursor position
	 */
	if (CUR_SCRN_X(td) >= w)
		CUR_SCRN_X(td) = w-1;

	if (CUR_SCRN_Y(td) >= h)
		CUR_SCRN_Y(td) = h-1;
}

/*------------------------------------------------------------------------
 * output the current mouse status (position & button state)
 */
static void xvt_term_mouse_output (TERMDATA *td)
{
	int	b	= 0;
	int	x	= td->mouse_x + 1;
	int	y	= td->mouse_y + 1;
	int	s	= FALSE;

	/*--------------------------------------------------------------------
	 * Determine what & when to send mouse data.
	 *
	 *	Mouse data is always sent as:
	 *
	 *			ESC [ M Cb Cx Cy
	 *
	 *	Cb, Cx, & Cy are button info, and x & y position. They are
	 *	each encoded as one byte + ' ' (0x20).  x & y are one-based,
	 *	i.e., the top left is (1,1).
	 *
	 *	1.	xterm mode:
	 *		The following is sent ONLY when a mouse button is
	 *		pressed or released:
	 *
	 *		Button info is encoded as follows:
	 *
	 *			bits 0-1	(0x03)	button info:
	 *
	 *							0	button 1 pressed
	 *							1	button 2 pressed
	 *							2	button 3 pressed
	 *							3	some button was released (no info on which)
	 *
	 *			bit  2		(0x04)	shift is down
	 *			bit  3		(0x08)	meta (alt) is down
	 *			bit  4		(0x10)	control is down
	 *			bit  5		(0x20)	reserved (for + ' ' encoding)
	 *			bit  6		(0x40)	always 0
	 *			bit  7		(0x80)	always 0
	 *	
	 *	2.	xvt mode:
	 *		If the keyboard is not "mapped", then output is in "xterm mode".
	 *		If the keyboard is "mapped", button info is encoded as follows:
	 *
	 *			bit  0		(0x01)	button 1 state (0 = dn, 1 = up)
	 *			bit  1		(0x02)	button 2 state (0 = dn, 1 = up)
	 *			bit  2		(0x04)	button 3 state (0 = dn, 1 = up)
	 *			bit  3		(0x08)	1 = scroll-wheel up one notch
	 *			bit  4		(0x10)	1 = scroll-wheel dn one notch
	 *			bit  5		(0x20)	reserved (for + ' ' encoding)
	 *			bit  6		(0x40)	1 = no button change, motion only
	 *			bit  7		(0x80)	always 1
	 *
	 *		Note that if the button info did not change, but only the
	 *		x,y changed, then this is a motion event. Also, a button
	 *		event can also contain a motion event.
	 *
	 *		Motion events are triggered only by the mouse moving across
	 *		a "character" boundary. Thus, if a mouse only moves a few pixels
	 *		within a character-cell on the screen, no movement is reported.
	 */
	if (! td->map_kbd)
	{
		/*----------------------------------------------------------------
		 * xterm mode - only output data if any mouse button changed
		 */
		if (td->prev_buttons != td->buttons)
		{
			/*------------------------------------------------------------
			 * get button change mask
			 * note that we ignore shift/meta/control
			 */
			if      ((td->buttons & BUT_L) && ! (td->prev_buttons & BUT_L))
				b = 0;
			else if ((td->buttons & BUT_M) && ! (td->prev_buttons & BUT_M))
				b = 1;
			else if ((td->buttons & BUT_R) && ! (td->prev_buttons & BUT_R))
				b = 2;
			else
				b = 3;

			s = TRUE;
		}
	}
	else
	{
		/*----------------------------------------------------------------
		 * xvt mode - always output info
		 */
		b = td->buttons | ((td->prev_buttons == td->buttons) << 6) | 0x80;
		s = TRUE;
	}

	if (s)
	{
		char	mouse_buf[128];

		sprintf(mouse_buf, "\033[M%c%c%c", b + ' ', x + ' ', y + ' ');
		xvt_term_write_keys(td, mouse_buf);

		xvt_log_inpkbd_fmt(td, "mouse(%02x,%3d,%3d)", b, x, y);
	}
}

/*------------------------------------------------------------------------
 * action debugging
 */
static void xvt_term_debug_action (TERMDATA *td, const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	xvt_log_action_var(td, fmt, args);
	va_end(args);
}

/*------------------------------------------------------------------------
 * process key-press event
 *
 * This routine processes key up/down events.  We always process down
 * events, but the only up events we are interested in are for "modifier"
 * keys (shift, control, alt).
 */
static void xvt_term_event_key (TERMDATA *td, XVT_EVENT_KEYPRESS *xe)
{
	char			keybuf[2];
	int				shft;
	int				ctrl;
	int				meta;
	int				rshft;
	int				rctrl;
	int				rmeta;
	int				state;
	int				send;
	int				key;
	int				kbd_state;
	const char *	p = 0;
	const char *	k = 0;

	/*--------------------------------------------------------------------
	 * get state (TRUE if down)
	 */
	state	= xe->key_state;
	send	= state;

	kbd_state	= xe->kbd_state;

	/*--------------------------------------------------------------------
	 * cache modifiers
	 */
	rshft	= shft	= ( (xe->kbd_state & KEY_MASK_SHFT) != 0 );
	rctrl	= ctrl	= ( (xe->kbd_state & KEY_MASK_CTRL) != 0 );
	rmeta	= meta	= ( (xe->kbd_state & KEY_MASK_META) != 0 );

	/*--------------------------------------------------------------------
	 * check key value
	 */
	key = xe->key_value;
	if (key < 0)
		return;

	/*--------------------------------------------------------------------
	 * if lookup was successful, use it
	 */
	if (key <= 0xff)
	{
		/*----------------------------------------------------------------
		 * regular or control key
		 */
		if (key <= 0x1f)
		{
			/*------------------------------------------------------------
			 * use control char table
			 */
			p = td->keys_ct[key].send;
			k = td->keys_ct[key].name;
		}
		else
		{
			/*------------------------------------------------------------
			 * regular key
			 */
			keybuf[0] = key;
			keybuf[1] = 0;

			p = keybuf;
			k = keybuf;
		}

		shft = ctrl = FALSE;
	}
	else
	{
		/*----------------------------------------------------------------
		 * special key
		 */
		int	key_type	= key & 0xff;
		int	key_mask	= key & XVT_KEY_MASK;

		if (key_mask == XVT_KEY_FN)
		{
			/*------------------------------------------------------------
			 * function key
			 */
			p = td->keys_fn[key_type].send;
			k = td->keys_fn[key_type].name;
		}

		else if (key_mask == XVT_KEY_KP)
		{
			/*------------------------------------------------------------
			 * keypad key
			 */
			p = td->keys_kp[key_type].send;
			k = td->keys_kp[key_type].name;

			/*------------------------------------------------------------
			 * check if special key
			 */
			if (send && (td->tw->keys & KEY_MASK_NUML) == 0)
			{
				switch (key)
				{
				case XVT_KEY_KP_ADD:
					if (td->xd->kpfont)
					{
						xvt_w_chg_font_no(td->tw, +1);
						xvt_term_set_win_title(td, td->win_label);
						send = FALSE;
					}
					break;

				case XVT_KEY_KP_SUB:
					if (td->xd->kpfont)
					{
						xvt_w_chg_font_no(td->tw, -1);
						xvt_term_set_win_title(td, td->win_label);
						send = FALSE;
					}
					break;
				}
			}
		}

		else if (key_mask == XVT_KEY_MS)
		{
			/*------------------------------------------------------------
			 * misc key
			 */
			p = td->keys_ms[key_type].send;
			k = td->keys_ms[key_type].name;

			/*------------------------------------------------------------
			 * check if special key
			 */
			if (send)
			{
				switch (key)
				{
				case XVT_KEY_MS_PRTSCR:
					if (meta)
						xvt_term_screen_clipboard(td);
					else
						xvt_term_screen_print(td);
					send = FALSE;
					break;
				}
			}
		}

		else if (key_mask == XVT_KEY_PF)
		{
			/*------------------------------------------------------------
			 * prefix key
			 *
			 * Note that when a prefix key is pressed, the state-bit is
			 * not set (e.g. when the SHIFT key is pressed, the SHIFT bit
			 * in the event struct is not on yet).
			 *
			 * Note that we don't send these keys directly, as they are
			 * sent by the call to xvt_term_set_kbd_state().
			 */
			send = FALSE;
			shft = ctrl = meta = FALSE;

			if (state)
			{
				p = td->keys_st_dn[key_type].send;
				k = td->keys_st_dn[key_type].name;
			}
			else
			{
				p = td->keys_st_up[key_type].send;
				k = td->keys_st_up[key_type].name;
			}

			if (state)
				kbd_state |=  xvt_term_key_masks[key_type];
			else
				kbd_state &= ~xvt_term_key_masks[key_type];
		}
	}

	/*--------------------------------------------------------------------
	 * check if we need to output keyboard state
	 */
	xvt_term_set_kbd_state(td, kbd_state);

	/*--------------------------------------------------------------------
	 * If we got a valid key, output it
	 */
	if (p != 0)
	{
		/*----------------------------------------------------------------
		 * debug output
		 */
		xvt_term_debug_action(td,
			"key %s (%c%c%c %04x) (%s%s%s%s) [%02x]",
			state ? "dn" : "up",
			rmeta ? 'm' : '.',
			rctrl ? 'c' : '.',
			rshft ? 's' : '.',
			xe->key_symbol,
			meta ? td->keys_pf[XVT_KEY_META].name : "",
			ctrl ? td->keys_pf[XVT_KEY_CTRL].name : "",
			shft ? td->keys_pf[XVT_KEY_SHFT].name : "",
			k,
			td->tw->keys);

		/*----------------------------------------------------------------
		 * write key to child if not empty & if we should send this key
		 */
		if (send && *p != 0)
		{
			char buf[128];

			*buf = 0;
			if (meta) strcat(buf, td->keys_pf[XVT_KEY_META].send);
			if (ctrl) strcat(buf, td->keys_pf[XVT_KEY_CTRL].send);
			if (shft) strcat(buf, td->keys_pf[XVT_KEY_SHFT].send);
			strcat(buf, p);
			xvt_term_write_keys(td, buf);

			xvt_log_inpkbd_fmt(td, "%s%s%s%s",
				meta ? td->keys_pf[XVT_KEY_META].name : "",
				ctrl ? td->keys_pf[XVT_KEY_CTRL].name : "",
				shft ? td->keys_pf[XVT_KEY_SHFT].name : "",
				k);

			td->read_pending = TRUE;
		}
	}
	else
	{
		xvt_term_debug_action(td,
			"key %s (%c%c%c %04x)",
			state ? "dn" : "up",
			rmeta ? 'm' : '.',
			rctrl ? 'c' : '.',
			rshft ? 's' : '.',
			xe->key_symbol);
	}
}

/*------------------------------------------------------------------------
 * process keymap-notify event
 *
 * This routine processes keymap notify events.  These events are always
 * sent after we get focus, and it is used to determine the state of
 * "modifier" keys (shift / control / alt) at the time we get focus.
 */
static void xvt_term_event_keymap (TERMDATA *td, XVT_EVENT_KEYMAP *xe)
{
	xvt_term_debug_action(td, "keymap (%c%c%c)",
		xe->kbd_state & KEY_MASK_META ? 'm' : '.',
		xe->kbd_state & KEY_MASK_CTRL ? 'c' : '.',
		xe->kbd_state & KEY_MASK_SHFT ? 's' : '.');

	xvt_term_set_kbd_state(td, xe->kbd_state);
}

/*------------------------------------------------------------------------
 * process button event
 *
 * This routine processes button up/down events, and can also generate
 * a pointer-moved event.
 */
static void xvt_term_event_button (TERMDATA *td, XVT_EVENT_BUTTON *xe)
{
	const char *	but_no;
	const char *	but_pos;

	/*--------------------------------------------------------------------
	 * save old button state
	 */
	td->prev_buttons = td->buttons;

	/*--------------------------------------------------------------------
	 * get button number
	 */
	switch (xe->button)
	{
	case BUT_L:	but_no = "L";	break;
	case BUT_M:	but_no = "M";	break;
	case BUT_R:	but_no = "R";	break;
	case BUT_U:	but_no = "U";	break;
	case BUT_D:	but_no = "D";	break;
	}

	/*--------------------------------------------------------------------
	 * get up/dn state
	 */
	switch (xe->button)
	{
	case BUT_L:
	case BUT_M:
	case BUT_R:
		if (xe->btn_state)
		{
			td->buttons |=  xe->button;
			but_pos = "dn";
		}
		else
		{
			td->buttons &= ~xe->button;
			but_pos = "up";
		}
		break;

	case BUT_U:
	case BUT_D:
		td->buttons |= xe->button;
		but_pos      = "--";
		break;
	}

	/*--------------------------------------------------------------------
	 * get (x,y) in chars
	 */
	td->mouse_x	= xe->x;
	td->mouse_y = xe->y;

	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_term_debug_action(td,
		"ptr btn   (%3d, %3d) (%s %s)",
		td->mouse_x, td->mouse_y, but_no, but_pos);

	/*--------------------------------------------------------------------
	 * send to child if he asked for them
	 */
	if (td->enable_mouse_events)
		xvt_term_mouse_output(td);

	/*--------------------------------------------------------------------
	 * clear out transient (scroll-wheel) events
	 */
	td->buttons &= ~(BUT_U | BUT_D);
}

/*------------------------------------------------------------------------
 * process pointer-moved event
 *
 * This routine processes pointer-movement events.  Note that a pointer-moved
 * event can also be internally generated in conjunction with a button event.
 */
static void xvt_term_event_motion (TERMDATA *td, XVT_EVENT_MOTION *xe)
{
	/*--------------------------------------------------------------------
	 * save old button state
	 */
	td->prev_buttons = td->buttons;

	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_term_debug_action(td,
		"ptr move  (%3d, %3d)",
		xe->x, xe->y);

	/*--------------------------------------------------------------------
	 * set new values if changed
	 */
	if (xe->x != td->mouse_x || xe->y != td->mouse_y)
	{
		td->mouse_x = xe->x;
		td->mouse_y = xe->y;

		/*----------------------------------------------------------------
		 * send to child if he asked for it
		 */
		if (td->enable_mouse_events)
			xvt_term_mouse_output(td);
	}
}

/*------------------------------------------------------------------------
 * process focus event
 *
 * This routine processes focus events.  It is used to show the cursor
 * as "solid" or "box", depending on whether we have focus or not.
 */
static void xvt_term_event_focus (TERMDATA *td, XVT_EVENT_FOCUS *xe)
{
	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_term_debug_action(td,
		"focus %s",
		xe->focus ? "in" : "out");

	/*--------------------------------------------------------------------
	 * set cursor mode according to focus
	 */
	td->cursor_mode = xe->focus;

	/*--------------------------------------------------------------------
	 * If losing focus,  set kbd_state to none.
	 * If gaining focus, set kbd_state to real state.
	 */
	xvt_term_set_kbd_state(td, td->cursor_mode ? -1 : 0);

	/*--------------------------------------------------------------------
	 * now redraw cursor if visible
	 */
	if (CUR_SCRN_CON(td))
	{
		xvt_term_set_txt_cursor(td, FALSE);
		xvt_term_set_txt_cursor(td, TRUE);
	}
}

/*------------------------------------------------------------------------
 * process expose event
 *
 * This routine processes expose events, which are represented as rectangles
 * which must be redrawn.  These events are internal only, i.e. they are
 * not sent to the "child" process.
 */
static void xvt_term_event_expose (TERMDATA *td, XVT_EVENT_EXPOSE *xe)
{
	int x, y;
	int w, h;

	/*--------------------------------------------------------------------
	 * convert info to cells
	 */
	x = xe->x;
	y = xe->y;

	w = xe->w;
	h = xe->h;

	if ((x + w) > CUR_SCRN_COLS(td))
		w = CUR_SCRN_COLS(td) - x;

	if ((y + h) > CUR_SCRN_ROWS(td))
		h = CUR_SCRN_ROWS(td) - y;

	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_term_debug_action(td,
		"expose    (%3d, %3d, %3d, %3d)",
		x, y, w, h);

	/*--------------------------------------------------------------------
	 * now do the repaint
	 */
	xvt_term_repaint_proc(td, x, y, w, h);
}

/*------------------------------------------------------------------------
 * process configure event
 *
 * This routine processes window-resize events. Note that we have set
 * the window to only allow the size to be a multiple of the font-char
 * size, so we can just take the new size & allocate new screen buffers.
 * This will also result in the "child" process being sent a resize signal.
 */
static void xvt_term_event_configure (TERMDATA *td, XVT_EVENT_CONFIG *xe)
{
	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_term_debug_action(td,
		"configure (%3d, %3d)",
		xe->w, xe->h);

	/*--------------------------------------------------------------------
	 * ignore event if it is a resize following a resize
	 */
	if (td->tw->prev_action == XVT_E_CONFIG)
		return;

	/*--------------------------------------------------------------------
	 * do the resize
	 */
	xvt_term_resize_proc(td, xe->w, xe->h);
	td->tw->doing_resize = TRUE;
}

/*------------------------------------------------------------------------
 * process client event
 *
 * This routine processes Window manager events which are sent to us.
 * Right now the only event we look for is WM_DELETE, which is a quit msg.
 */
static void xvt_term_event_client (TERMDATA *td, XVT_EVENT_WINMSG *xe)
{
	/*--------------------------------------------------------------------
	 * process message
	 */
	switch (xe->winmsg)
	{
	case XVT_W_DELETE:
		/*----------------------------------------------------------------
		 * debug output
		 */
		xvt_term_debug_action(td, "wm-delete");

		/*----------------------------------------------------------------
		 * tell child to terminate
		 */
		if (td->pid != 0)
			kill(td->pid, SIGTERM);

		/*----------------------------------------------------------------
		 * start shutdown process
		 */
		td->bail = 1;
		break;
	}
}

/*------------------------------------------------------------------------
 * process map event
 */
static void xvt_term_event_map (TERMDATA *td, XVT_EVENT_MAP *xe)
{
	/*--------------------------------------------------------------------
	 * here is where we could send a MAP/UNMAP key to the program
	 */
	if (xe->mapped)
	{
		xvt_term_debug_action(td, "map window");
	}
	else
	{
		xvt_term_debug_action(td, "unmap window");
	}
}

/*------------------------------------------------------------------------
 * process an event
 *
 * Here is where all X events we are interested in are processed.
 */
static void xvt_term_process_event (TERMDATA *td, XVT_EVENT *xe)
{
	EVENT_DATA *	xk		= &xe->event_data;
	int				action	= xe->event_type;

	td->tw->doing_resize = FALSE;

	/*--------------------------------------------------------------------
	 * perform proper event procedure
	 */
	switch (action)
	{
	case XVT_E_KEYPRESS:
						xvt_term_event_key		(td, &xk->keypress);	break;
	case XVT_E_KEYMAP:	xvt_term_event_keymap	(td, &xk->keymap);		break;
	case XVT_E_BUTTON:	xvt_term_event_button	(td, &xk->button);		break;
	case XVT_E_MOTION:	xvt_term_event_motion	(td, &xk->motion);		break;
	case XVT_E_FOCUS:	xvt_term_event_focus	(td, &xk->focus);		break;
	case XVT_E_EXPOSE:	xvt_term_event_expose	(td, &xk->expose);		break;
	case XVT_E_CONFIG:	xvt_term_event_configure(td, &xk->config);		break;
	case XVT_E_WINMSG:	xvt_term_event_client	(td, &xk->winmsg);		break;
	case XVT_E_MAP:		xvt_term_event_map		(td, &xk->map);			break;

	default:
		xvt_term_debug_action(td, "action %d", action);
		break;
	}

	td->tw->prev_action = action;
}

/*------------------------------------------------------------------------
 * set window title
 */
void xvt_term_set_win_title (TERMDATA *td, char *str)
{
	char	temp_title[256];

	if (str != 0 && *str != 0)
	{
		if (td->disp_info && td->pd != 0)
		{
			strcpy(temp_title, str);
			sprintf(temp_title + strlen(temp_title),
				" [%d: %s: %s]",
				td->our_pid,
				td->pd->slv_name,
				td->xd->reg_fname[td->tw->cur_font_no]);
			str = temp_title;
		}

		xvt_w_set_win_title(td->tw, str);
	}
}

/*------------------------------------------------------------------------
 * set icon title
 */
void xvt_term_set_ico_title (TERMDATA *td, char *str)
{
	if (str != 0 && *str != 0)
		xvt_w_set_ico_title(td->tw, str);
}

/*------------------------------------------------------------------------
 * set the icon for a window
 */
void xvt_term_set_icon (TERMDATA *td, int width, int height,
	unsigned char *bits)
{
	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (width <= 0 || height <= 0 || bits == 0)
		return;

	/*--------------------------------------------------------------------
	 * cache icon data
	 */
	td->ico_bits	= bits;
	td->ico_width	= width;
	td->ico_height	= height;

	/*--------------------------------------------------------------------
	 * now set it
	 */
	xvt_w_set_icon(td->tw, td->ico_width, td->ico_height, td->ico_bits);
}

/*------------------------------------------------------------------------
 * iconify the window
 */
void xvt_term_set_icon_state (TERMDATA *td, int state)
{
	xvt_w_set_icon_state(td->tw, state);
}

/*------------------------------------------------------------------------
 * set font number
 */
void xvt_term_set_font_no (TERMDATA *td, int n)
{
	xvt_w_set_font_no(td->tw, n);
}

/*------------------------------------------------------------------------
 * set reverse video
 */
void xvt_term_set_rev_video (TERMDATA *td, int reverse)
{
	xvt_w_set_rev_video(td->tw, reverse);
	xvt_term_set_scrn_mode(td, SCRN_MODE_MONO);
	xvt_term_repaint(td);
}

/*------------------------------------------------------------------------
 * sound the bell
 */
void xvt_term_bell (TERMDATA *td)
{
	if (! td->xd->quiet)
	{
		/*----------------------------------------------------------------
		 * check if we need to unmap
		 */
		if (td->tw->iconified && td->xd->mapalert)
			xvt_w_set_icon_state(td->tw, FALSE);

		/*----------------------------------------------------------------
		 * check if visual or auditory
		 */
		if (td->xd->visual)
			xvt_term_flash(td);
		else
			xvt_w_bell(td->tw);
	}
}

/*------------------------------------------------------------------------
 * open the printer
 */
void xvt_term_printer_open (TERMDATA *td)
{
	if (td->prt == 0)
	{
		td->prt = xvt_prt_open(P_NONE, td->xd->printer);
	}

	if (td->prt != 0)
			td->prt_mode = PRT_MODE_NORM;
}

/*------------------------------------------------------------------------
 * close the printer
 */
void xvt_term_printer_close (TERMDATA *td)
{
	if (! td->xd->prtkeep)
	{
		if (td->prt != 0)
		{
			xvt_prt_close(td->prt);
			td->prt = 0;
		}
	}
	td->prt_mode = PRT_MODE_NONE;
}

/*------------------------------------------------------------------------
 * fill input buffer from child output
 */
static int xvt_term_fill_buffer (TERMDATA *td)
{
	int	num;

	td->input_ptr	= td->input_buffer;
	td->input_count	= 0;

	num = read(td->pd->mst_fd, td->input_buffer, sizeof(td->input_buffer));
	xvt_log_screen_buf(td, td->input_buffer, num);

	if (num < 0)
		return (-1);

	if (num > 0)
		td->input_count	= num;

	return (0);
}

/*------------------------------------------------------------------------
 * process child output
 */
static int xvt_term_process_child_output (TERMDATA *td)
{
	int count	= 0;
	int rc;

	/*--------------------------------------------------------------------
	 * check if we need to unmap
	 */
	if (td->tw->iconified && td->xd->mapoutput)
		xvt_w_set_icon_state(td->tw, FALSE);

	/*--------------------------------------------------------------------
	 * read in as much child output as we can
	 */
	rc = xvt_term_fill_buffer(td);
	if (rc)
		return (-1);

	/*--------------------------------------------------------------------
	 * now process the input
	 */
	while (td->input_count)
	{
		int ch;

		ch = *td->input_ptr++;
		td->input_count--;

		count++;
		xvt_emul_char_input(td, ch);
	}

	return (count);
}

/*------------------------------------------------------------------------
 * loop once
 *
 *	returns:
 *		-1	time to bail
 *		 0	OK
 */
int xvt_term_loop_once (TERMDATA *td, int ms)
{
	fd_set			readfds;
	struct timeval	tv;
	int				rc;
	int				xfd;
	int				cfd;
	int				max_fd;
	int				usecs;

	/*--------------------------------------------------------------------
	 * check if we are really done
	 */
	if (! td->init || td->bail >= XVT_DIE_LOOPS)
		return (-1);

	/*--------------------------------------------------------------------
	 * get fd of X display & input from child
	 */
	xfd = xvt_w_get_xfd(td->tw);
	cfd = td->pd->mst_fd;

	/*--------------------------------------------------------------------
	 * determine max # fds to look for
	 */
	if (xfd > cfd)
		max_fd = xfd + 1;
	else
		max_fd = cfd + 1;

	/*--------------------------------------------------------------------
	 * determine amount of time to wait
	 */
	if (td->bail > 0)
	{
		/*----------------------------------------------------------------
		 * If the child was told to die, we bump our count &
		 * give him time to die.
		 */
		td->bail++;
		usecs	= td->die_interval * 1000;
	}
	else
	{
		/*----------------------------------------------------------------
		 * use caller's wait interval
		 */
		usecs	= ms * 1000;
	}

	/*--------------------------------------------------------------------
	 * wait for something to happen
	 */
	while (TRUE)
	{
		/*----------------------------------------------------------------
		 * set file-descriptors to wait for
		 *
		 * We have to set this every time through the loop since select
		 * always clears it.
		 */
		FD_ZERO(&readfds);
		FD_SET(cfd, &readfds);
		FD_SET(xfd, &readfds);

		/*----------------------------------------------------------------
		 * set program wait interval
		 *
		 * We have to set this interval every time through the loop
		 * since Linux (unlike other Unixes) clears the timeval struct
		 * when select returns.
		 */
		tv.tv_sec	= 0;
		tv.tv_usec	= usecs;

		/*----------------------------------------------------------------
		 * now do the actual wait
		 */
		rc = select(max_fd, &readfds, 0, 0, &tv);
		if (rc < 0 && errno == EINTR)
			continue;
		break;
	}

	/*--------------------------------------------------------------------
	 * Process any X cmds if any.  We don't believe what select says, but
	 * are hoping that at least any X events would have triggered the
	 * select.
	 */
	while (TRUE)
	{
		XVT_EVENT	xe;

		if (! xvt_w_event(td->tw, &xe))
			break;

		xvt_term_process_event(td, &xe);
	}

	/*--------------------------------------------------------------------
	 * If select returned with error, then bail.  Note we do this after
	 * flushing any lingering X events.
	 */
	if (rc < 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * Now break out if child died.  We do this after we processed any
	 * lingering X cmds the child issued before dying.
	 */
	if (td->pid == 0 || td->pid_died != 0)
		return (-1);

	/*----------------------------------------------------------------
	 * process any program output if any
	 */
	if (FD_ISSET(cfd, &readfds))
	{
		/*------------------------------------------------------------
		 * turn off cursor while processing output
		 */
		xvt_term_set_txt_cursor(td, FALSE);

		/*--------------------------------------------------------
		 * process output
		 */
		rc = xvt_term_process_child_output(td);
		if (rc < 0)
		{
			/*----------------------------------------------------
			 * got EOF - start shutdown process
			 */
			if (td->pid == td->our_pid)
			{
				/*--------------------------------------------------------
				 * Running in "here" mode.  Immediately signal death the
				 * next time around.
				 */
				td->bail = XVT_DIE_LOOPS;
			}
			else
			{
				/*--------------------------------------------------------
				 * Running in "task" mode.  Start slowdown & wait for
				 * child to die.  We essentially give him XVT_DIE_LOOPS
				 * times "td->die_interval" (100) ms.  If he hasn't died
				 * by then, we just bail and terminate him forcibly.
				 */
				td->bail = 1;
			}
		}
		else if (rc == 0)
		{
			/*------------------------------------------------------------
			 * We read nothing, even though select() said we had something.
			 * Ignore what select() says & break out.
			 * But we do turn the cursor back on if needed.
			 */
			xvt_term_set_txt_cursor(td, TRUE);
		}
	}
	else
	{
		/*--------------------------------------------------------
		 * turn cursor back on if needed
		 */
		xvt_term_set_txt_cursor(td, TRUE);
	}

	return (0);
}

/*------------------------------------------------------------------------
 * main loop
 *
 * Note that this routine may be run as a sub-thread.
 */
void * xvt_term_main_loop (void *data)
{
	TERMDATA *	td = (TERMDATA *)data;
	int			rc;

	/*--------------------------------------------------------------------
	 * we loop until told to bail
	 */
	while (TRUE)
	{
		rc = xvt_term_loop_once(td, td->run_interval);
		if (rc < 0)
			break;
	}

	/*--------------------------------------------------------------------
	 * If running as a thread, our returning will terminate the thread,
	 * so clear out our thread-id so we know we aren't running.
	 *
	 * If we are not running as a thread, this is already cleared.
	 */
	td->our_tid = 0;

	return (0);
}

/*------------------------------------------------------------------------
 * setup everything
 */
int xvt_term_setup_all (TERMDATA *td)
{
	/*--------------------------------------------------------------------
	 * do X setup
	 */
	td->tw = xvt_w_setup_all(td->xd);
	if (td->tw == 0)
		return (td->xd->error_num);

	td->scrn_mode	= SCRN_MODE_MONO;

	/*--------------------------------------------------------------------
	 * setup display window color & font
	 */
	xvt_term_set_scrn_mode(td, td->scrn_mode);

	/*--------------------------------------------------------------------
	 * setup current font
	 */
	td->tw->cur_font_type = FONT_NONE;
	xvt_term_set_reg_font(td);

	/*--------------------------------------------------------------------
	 * setup text screen
	 */
	xvt_term_setup_text(td, td->tw->win_cols, td->tw->win_rows);

	/*--------------------------------------------------------------------
	 * set window & icon labels
	 */
	xvt_term_set_win_title(td, td->win_label);
	xvt_term_set_ico_title(td, td->ico_label);

	/*--------------------------------------------------------------------
	 * setup terminal emulator
	 */
	td->cmd_tbl_ptr = &xvt_vt100_cmdtbl;
	(td->cmd_tbl_ptr->init_rtn)(td);

	return (0);
}

/*------------------------------------------------------------------------
 * close out everything
 */
void xvt_term_close_all (TERMDATA *td)
{
	xvt_pty_close(td->pd);
	xvt_term_close_text(td);
	xvt_w_close_all(td->tw);
	if (td->prt != 0)
		xvt_prt_close(td->prt);

	td->tw		= 0;
	td->pd		= 0;
	td->prt		= 0;
	td->init	= FALSE;
}

/*------------------------------------------------------------------------
 * shutdown prior to cloning
 */
void xvt_term_shutdown (TERMDATA *td)
{
	/*--------------------------------------------------------------------
	 * close the event sub-thread if running
	 *
	 * The man page for fork() states that, when using POSIX threads,
	 * fork() will only copy the main thread, not any sub-threads.
	 * But, on Solaris (at least), fork() appears to copy ALL threads.
	 * So, we have to terminate the sub-thread if one is running.
	 *
	 * Note: if the sub-thread is not running at this time, this will
	 * silently do nothing.
	 */
#if 0
	/*--------------------------------------------------------------------
	 * NOTE: this is commented out for now, since Linux apparently
	 * does not copy sub-threads in a fork(), and pthread_join() hangs
	 * when called with a bogus thread id.  For Solaris, we will just
	 * use fork1() for now, which does only copy the main thread.
	 */
	if (td->our_tid != 0)
	{
		td->bail = XVT_DIE_LOOPS;
		xvtthread_wait(td->our_tid);
		td->bail = 0;
	}
#endif

	/*--------------------------------------------------------------------
	 * Now we can close the pty.
	 */
	xvt_pty_close(td->pd);
	td->pd = 0;

	/*----------------------------------------------------------------
	 * Close the display.  All we do is close the connection, since
	 * calling XCloseDisplay() will free parent resources, which will
	 * hose our parent program.
	 *
	 * Note that we don't call any X routines to free anything in the
	 * X info struct (for the same reason), but we do free the struct itself.
	 *
	 * Note that we cache our current font number first.
	 */
	td->xd->fontno = td->tw->cur_font_no;
	xvt_w_close_con(td->tw);

	xvt_w_free(td->tw);
	td->tw = 0;

	/*----------------------------------------------------------------
	 * close any printer that is open, since we can't clone any piped
	 * processes.
	 */
	if (td->prt != 0)
	{
		xvt_prt_close(td->prt);
		td->prt = 0;
		td->prt_mode = PRT_MODE_NONE;
	}
}

/*------------------------------------------------------------------------
 * restart the display
 */
int xvt_term_restart (TERMDATA *td)
{
	/*--------------------------------------------------------------------
	 * Re-open a new display.
	 */
	td->tw = xvt_w_setup_all(td->xd);
	if (td->tw == 0)
	{
		xvt_term_close_all(td);
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * Now open a new pty.
	 */
	td->pd = xvt_pty_init(0);
	if (td->pd == 0)
	{
		xvt_term_close_all(td);
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * Setup display color to what we had before (remember that setup
	 * always sets it to mono).
	 */
	xvt_term_set_scrn_mode(td, td->scrn_mode);

	/*--------------------------------------------------------------------
	 * setup current font
	 */
	td->tw->cur_font_type = FONT_NONE;
	xvt_term_set_reg_font(td);

	/*--------------------------------------------------------------------
	 * set window & icon labels and the icon
	 */
	xvt_term_set_win_title(td, td->win_label);
	xvt_term_set_ico_title(td, td->ico_label);

	if (td->ico_width > 0 && td->ico_height > 0)
		xvt_w_set_icon(td->tw, td->ico_width, td->ico_height, td->ico_bits);

	return (0);
}

/*------------------------------------------------------------------------
 * output list of cursor names
 */
void xvt_term_list_pointers (FILE *fp, int verbose)
{
	int	i;

	for (i=0; ; i++)
	{
		const char *name	= xvt_w_pointer_name(i);
		int			code	= xvt_w_pointer_code(i);

		if (name == 0)
			break;

		if (verbose)
			fprintf(fp, "%3d\t%s\n", code, name);
		else
			fprintf(fp, "%s\n", name);
	}
}
