/*------------------------------------------------------------------------
 * produce help listings
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * type names
 */
static const char *	xvt_tk_help_types[] =
{
	"invalid",
	"string",
	"color",
	"font",
	"numeric",
	"boolean",
	"debug",
	"command"
};

/*------------------------------------------------------------------------
 * list all toolkit options
 *
 * verbose:
 *			0		short listing
 *			1		long listing
 *			2		detailed listing
 *
 *	here:
 *			TRUE	running in "here" mode
 *			FALSE	not in "here" mode
 */
void xvt_tk_help (XVT_DATA *xd, FILE *fp, int verbose, int here)
{
	const TK_OPTIONS *	op;

	/*--------------------------------------------------------------------
	 * sanity check
	 */
	if (fp == 0)
		return;

	/*--------------------------------------------------------------------
	 * check mode
	 */
	fprintf(fp, "X Toolkit options:\n");
	if (verbose == 0)
	{
		/*----------------------------------------------------------------
		 * short mode
		 */
		char	line[76];
		int		maxlen	= sizeof(line) - 1;
		int		len		= 0;
		int		type	= -1;

		for (op=xvt_tk_options; op->opt_type; op++)
		{
			int l;

			/*------------------------------------------------------------
			 * check if this entry should be skipped
			 */
			if (op->opt_flags & (TK_FLAG_NOTINHELP | TK_FLAG_NOHELP_0))
				continue;

			if (op->opt_desc == 0)
				continue;

			if (here && (op->opt_flags & TK_FLAG_NOTHERE))
				continue;

#if 0
			/*------------------------------------------------------------
			 * This will separate each option type.
			 * (Disabled because it makes the help output too long...)
			 */
			if (type != op->opt_type)
			{
				if (len > 0)
				{
					fprintf(fp, "  %s\n", line);
					len = 0;
				}

				fprintf(fp, "  --%s options--\n",
					xvt_tk_help_types[op->opt_type]);
			}
#endif

			/*------------------------------------------------------------
			 * print out buffer if needed
			 */
			l = strlen(op->opt_help) + 3;
			if ((len + l) > maxlen)
			{
				fprintf(fp, "  %s\n", line);
				len = 0;
			}

			/*------------------------------------------------------------
			 * add entry to buffer
			 */
			type = op->opt_type;
			if (len > 0)
				line[len++] = ' ';
			line[len++] = '[';
			len += xvt_strcpy(line + len, op->opt_help);
			line[len++] = ']';
			line[len  ] = 0;
		}

		/*----------------------------------------------------------------
		 * print buffer if anything left in it
		 */
		if (len > 0)
			fprintf(fp, "  %s\n", line);
	}
	else if (verbose == 1)
	{
		/*----------------------------------------------------------------
		 * long listing
		 */
		for (op=xvt_tk_options; op->opt_type; op++)
		{
			/*------------------------------------------------------------
			 * check if this entry should be skipped
			 */
			if (op->opt_flags & TK_FLAG_NOTINHELP)
				continue;

			if (op->opt_desc == 0)
				continue;

			if (here && (op->opt_flags & TK_FLAG_NOTHERE))
				continue;

			/*------------------------------------------------------------
			 * print entry
			 */
			fprintf(fp, "  %-16s  %s\n",
				op->opt_help,
				op->opt_desc);

			/*------------------------------------------------------------
			 * print any extra lines associated with it
			 */
			if (op->opt_lines != 0)
			{
				const char ** l;

				for (l=op->opt_lines; *l; l++)
				{
					fprintf(fp, "  %-16s  %s\n", "", *l);
				}
			}
		}
	}
	else
	{
		/*----------------------------------------------------------------
		 * detailed listing
		 */
		int	type = -1;

		fprintf(fp,
			"  Option            Default       "
			"Resource name      Description\n");
		fprintf(fp,
			"------------------  ------------  "
			"----------------   ------------------------\n");

		for (op=xvt_tk_options; op->opt_type; op++)
		{
			const char *	opt_default;
			int				opt_default_num;
			char			opt_default_buf[24];

			/*------------------------------------------------------------
			 * check if this entry should be skipped
			 */
			if (op->opt_flags & TK_FLAG_NOTINHELP)
				continue;

			if (op->opt_desc == 0)
				continue;

			if (here && (op->opt_flags & TK_FLAG_NOTHERE))
				continue;

			/*------------------------------------------------------------
			 * if option type changed, output header for it
			 */
			if (op->opt_type != type)
			{
				if (type != -1)
					fprintf(fp, "\n");

				type = op->opt_type;
				fprintf(fp, "--%s options--\n", xvt_tk_help_types[type]);
			}

			/*------------------------------------------------------------
			 * get printable version of default
			 */
			switch (op->opt_type)
			{
			case XVT_TK_OPT_STR:
			case XVT_TK_OPT_CLR:
			case XVT_TK_OPT_FNT:
				opt_default = (op->opt_default == 0 ? "" : op->opt_default);
				break;

			case XVT_TK_OPT_NUM:
				opt_default_num = (int)op->opt_default;
				sprintf(opt_default_buf, "%d", opt_default_num);
				opt_default = opt_default_buf;
				break;

			case XVT_TK_OPT_BLN:
				opt_default_num = (int)op->opt_default;
				opt_default = (opt_default_num ? "on" : "off");
				break;

			case XVT_TK_OPT_DBG:
			case XVT_TK_OPT_CMD:
				opt_default = "";
				break;
			}

			if (xd != 0 && xd->appname != 0 &&
			    (op->opt_flags & TK_FLAG_APPOVR))
			{
				opt_default = xd->appname;
			}

			/*------------------------------------------------------------
			 * now print the entry
			 */
			fprintf(fp, "  %-16s  %-12s  %-18s %s\n",
				op->opt_help,
				opt_default,
				op->opt_resname == 0 ? "" : op->opt_resname,
				op->opt_desc);

			/*------------------------------------------------------------
			 * print any extra lines associated with it
			 */
			if (op->opt_lines != 0)
			{
				const char ** l;

				for (l=op->opt_lines; *l; l++)
				{
					fprintf(fp, "  %-16s  %-12s  %-18s %s\n",
						"", "", "", *l);
				}
			}
		}
	}
}
