/*------------------------------------------------------------------------
 * toolkit resource processing
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * list of files to process for resources
 *
 * These are read first.
 */
static const char *	xvt_tk_res_files[] =
{
	"$XENVIRONMENT",
	"~/.Xdefaults-$HOSTNAME",
	"~/.Xresources-$HOSTNAME",
	"~/.Xdefaults",
	"~/.Xresources",

	0
};

/*------------------------------------------------------------------------
 * list of directories to look for resource file in
 *
 * File looked for is "<dir>/<resource-name>"
 *
 * These are read after all files above.
 */
static const char *	xvt_tk_res_dirs[] =
{
	"$XAPPLRESDIR",
	"/usr/lib/X11/app-defaults",
	"/usr/X/lib/app-defaults",
	"/usr/X/lib/locale/$LANG/app-defaults",

	0
};

/*------------------------------------------------------------------------
 * list of paths to look for resource file in
 *
 * File looked for is "<dir>/<resource-name>"
 *
 * These are read after the dirs above.
 */
static const char *	xvt_tk_res_paths[] =
{
	"$XUSERFILESEARCHPATH",

	0
};

/*------------------------------------------------------------------------
 * process a resource file data entry
 *
 * returns: 0 if used, -1 if not
 */
static int xvt_tk_resdata (XVT_DATA *xd, const char *name, const char *data,
	const char *resname)
{
	const TK_OPTIONS *	op;
	const char *		var_name;
	const char *		p;

	/*--------------------------------------------------------------------
	 * <name> *should* be in one of the following forms:
	 *
	 *		<var-name>
	 *		*<var-name>
	 *		.<var-name>
	 *		<resname>.<var-name>
	 *		<resname>*<var-name>
	 *
	 * Any other forms are considered invalid.
	 */
	p = xvt_strpbrk(name, "*.");
	if (p != 0)
	{
		/*----------------------------------------------------------------
		 * second level is invalid
		 */
		if (xvt_strpbrk(p+1, ".*") != 0)
			return (-1);

		/*----------------------------------------------------------------
		 * if not at beginning, check resname
		 */
		if (p != name)
		{
			if (resname == 0)
				return (-1);

			if (xvt_strnccmp(name, resname, p-name) != 0)
				return (-1);
		}

		var_name = (p+1);
	}
	else
	{
		var_name = name;
	}

	/*--------------------------------------------------------------------
	 * we got a valid entry - check if we know about this var-name
	 */
	op = xvt_tk_find_by_resname(var_name);
	if (op == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * ignore this entry if already set
	 */
	switch (op->opt_type)
	{
	case XVT_TK_OPT_STR:
	case XVT_TK_OPT_CLR:
	case XVT_TK_OPT_FNT:
		{
			const char ** pstr = (const char **)((char *)xd + op->opt_offset);

			if (*pstr != 0)
				return (-1);
		}
		break;

	case XVT_TK_OPT_NUM:
	case XVT_TK_OPT_BLN:
		{
			int * pnum = (int *)((char *)xd + op->opt_offset);

			if (*pnum != -1)
				return (-1);
		}
		break;

	case XVT_TK_OPT_DBG:
	case XVT_TK_OPT_CMD:
		/*----------------------------------------------------------------
		 * these entries not allowed in a resource file
		 */
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * OK - set this entry
	 */
	xvt_tk_opt_set(xd, op->opt_string, data, FALSE, 0);

	return (0);
}

/*------------------------------------------------------------------------
 * load a resource file
 */
int xvt_tk_resfile_load (XVT_DATA *xd, const char *path, const char *resname)
{
	FILE *	fp;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	if (path == 0 || *path == 0)
	{
		sprintf(xd->error_buf, "NULL path");
		xd->error_num = XVT_ERR_INVARG;
		return (XVT_ERR_INVARG);
	}

	/*--------------------------------------------------------------------
	 * open the resource file
	 *
	 * Note that a non-existent file is OK.
	 */
	fp = fopen(path, "r");
	if (fp == 0)
		return (XVT_OK);

	/*--------------------------------------------------------------------
	 * process all lines in the file
	 */
	while (TRUE)
	{
		char	line[BUFSIZ];
		char *	lp;
		char *	p;
		int		rc;

		/*----------------------------------------------------------------
		 * read in the next line
		 *
		 * For now, we aren't dealing with continued lines
		 */
		if (fgets(line, sizeof(line), fp) == 0)
			break;

		/*----------------------------------------------------------------
		 * Strip leading/trailing white-space & remove any comments.
		 */
		lp = xvt_strip(line, TRUE);

		p = xvt_strchr(lp, '!');
		if (p != 0)
		{
			*p = 0;
			xvt_strip(lp, FALSE);
		}

		/*----------------------------------------------------------------
		 * Ignore line if only a comment (or blank) line
		 */
		if (*lp == 0)
			continue;

		/*----------------------------------------------------------------
		 * check for include line
		 *
		 * Note: relative pathnames are relative to dir of prev path
		 */
		if (*lp == '#')
		{
			char	inc_name[256];
			char	inc_path[256];

			if (xvt_strccmpn("#include", lp) == 0)
			{
				p = strchr(lp, '"');
				if (p != 0)
				{
					xvt_strcpy(inc_path, p);
					if (*inc_path != '/')
					{
						strcpy(inc_name, inc_path);
						xvt_path_dirname(inc_path, path);
						strcat(inc_path, "/");
						strcat(inc_path, inc_name);
					}
					xvt_tk_resfile_load(xd, inc_path, resname);
				}
			}
			continue;
		}

		/*----------------------------------------------------------------
		 * If not in form of "<name>:<data>", ignore the line.
		 */
		p = xvt_strchr(lp, ':');
		if (p == 0)
			continue;

		*p++ = 0;
		xvt_strip(lp, FALSE);
		p = xvt_strip(p, TRUE);

		/*----------------------------------------------------------------
		 * we got a real line - process it
		 */
		rc = xvt_tk_resdata(xd, lp, p, resname);

		/*----------------------------------------------------------------
		 * If we used this entry, check if resource file debugging is on
		 */
		if (rc == 0 && xd->resfil_fp != 0)
		{
			fprintf(xd->resfil_fp, "%s: %s: %s\n", path, lp, p);
		}
	}

	fclose(fp);
	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * dump all data to a stream
 *
 * If <resname> is NULL, then all entries are written as "<name>: <value>",
 * otherwise entries are written as "<resname>.<name>: <value>".
 *
 * Note, that according to convention, the <resname> written in entries
 * is converted to all lower-case.
 */
int xvt_tk_options_dump (XVT_DATA *xd, FILE *fp, int show_dflt)
{
	const TK_OPTIONS *	op;
	char				data_buf[256];
	char *				p;
	int					maxlen;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	if (fp == 0)
	{
		sprintf(xd->error_buf, "NULL stream");
		xd->error_num = XVT_ERR_INVARG;
		return (XVT_ERR_INVARG);
	}

	/*--------------------------------------------------------------------
	 * get max length of a options string
	 */
	maxlen = 0;
	for (op=xvt_tk_options; op->opt_type; op++)
	{
		if (op->opt_string != 0)
		{
			int l	= strlen(op->opt_string);

			if (l > maxlen)
				maxlen = l;
		}
	}

	/*--------------------------------------------------------------------
	 * process all entries
	 */
	for (op=xvt_tk_options; op->opt_type; op++)
	{
		p = xvt_tk_data_value(xd, op, show_dflt, data_buf);
		if (p != 0)
		{
			fprintf(fp, "%-*s : %s\n", maxlen, op->opt_string, p);
		}
	}

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * save all data to a resource file
 *
 * If <resname> is NULL, then all entries are stored as "<name>: <value>",
 * and we assume that this resource file is specific to this resname,
 * so the entire file is written.
 *
 * If <resname> is not NULL, then it is assumed that this is a generic
 * resource file which has to be updated.  In this case, entries are
 * written as "<resname>.<name>: <value>", and all previous entries
 * in the form "<resname>[.*]..." are deleted.
 *
 * Note, that according to convention, the <resname> written in entries
 * is converted to all lower-case.
 */
int xvt_tk_resfile_save (XVT_DATA *xd, const char *path, const char *resname)
{
	const TK_OPTIONS *	op;
	char				res_path[256];
	char				res_temp[256];
	char				resbuf[128];
	char				data_buf[256];
	FILE *				fp_inp;
	FILE *				fp_out;
	char *				p;
	const char *		s;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	if (path == 0 || *path == 0)
	{
		sprintf(xd->error_buf, "NULL path");
		xd->error_num = XVT_ERR_INVARG;
		return (XVT_ERR_INVARG);
	}

	xvt_path_resolve(res_path, path);

	/*--------------------------------------------------------------------
	 * update or rewrite?
	 */
	if (resname == 0 || *resname == 0)
	{
		/*----------------------------------------------------------------
		 * rewrite entire file
		 */
		{
			/*------------------------------------------------------------
			 * open resource file
			 */
			fp_out = fopen(res_path, "w");
			if (fp_out == 0)
			{
				sprintf(xd->error_buf, "cannot open file %s", res_path);
				xd->error_num = XVT_ERR_INVARG;
				return (XVT_ERR_INVARG);
			}

			/*------------------------------------------------------------
			 * write all set values as "<name>: <value>"
			 */
			for (op=xvt_tk_options; op->opt_type; op++)
			{
				p = xvt_tk_data_value(xd, op, FALSE, data_buf);
				if (p != 0)
				{
					fprintf(fp_out, "%s: %s\n", op->opt_resname, p);
				}
			}
			fclose(fp_out);
		}
	}
	else
	{
		/*----------------------------------------------------------------
		 * update file
		 */
		{
			/*------------------------------------------------------------
			 * get resource name as all lower-case
			 */
			for (p=resbuf, s=resname; *s; s++)
			{
				*p++ = tolower(*s);
			}
			*p = 0;

			/*------------------------------------------------------------
			 * open resource file as input (failure is OK)
			 */
			fp_inp = fopen(res_path, "r");

			/*------------------------------------------------------------
			 * open temp output file
			 */
			strcpy(res_temp, res_path);
			strcat(res_temp, ".tmp");
			fp_out = fopen(res_temp, "w");
			if (fp_out == 0)
			{
				if (fp_inp != 0)
					fclose(fp_inp);

				sprintf(xd->error_buf, "cannot open file %s", res_temp);
				xd->error_num = XVT_ERR_INVARG;
				return (XVT_ERR_INVARG);
			}

			/*------------------------------------------------------------
			 * copy all lines from orig file which don't contain resname
			 */
			while (fp_inp != 0)
			{
				char	line[BUFSIZ];

				if (fgets(line, sizeof(line), fp_inp) == 0)
					break;

				if (xvt_strstr(line, resbuf) == 0)
					fputs(line, fp_out);
			}
			if (fp_inp != 0)
				fclose(fp_inp);

			/*------------------------------------------------------------
			 * write all set entries as "<resname>.<name>: <value>"
			 */
			for (op=xvt_tk_options; op->opt_type; op++)
			{
				p = xvt_tk_data_value(xd, op, FALSE, data_buf);
				if (p != 0)
				{
					fprintf(fp_out, "%s.%s: %s\n", resbuf, op->opt_resname, p);
				}
			}
			fclose(fp_out);

			/*------------------------------------------------------------
			 * rename "<resfile>.tmp" as "<resfile>"
			 */
			if (rename(res_temp, res_path) < 0)
			{
				sprintf(xd->error_buf, "Cannot rename %s as %s",
					res_temp, res_path);
				xd->error_num = XVT_ERR_INVARG;
				return (XVT_ERR_INVARG);
			}
		}
	}

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * process all known resources
 *
 * Note: We assume that all cmd-line options have already been processed,
 * and those are the *most* specific.  We then go from *more* to *less*
 * specific sources, and we don't store an entry if it already has a value.
 *
 * We initialized string pointers to 0 and all bool/num values to -1,
 * so we know if a value has already been stored in it.
 */
int xvt_tk_resource (XVT_DATA *xd)
{
	const char *	resname;
	const char **	p;
	char			res_skel[256];
	char			res_path[256];
	int				rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (xd == 0)
		return (XVT_ERR_NULLDATA);

	/*--------------------------------------------------------------------
	 * check if a resource string was specified
	 */
	if (xd->resstr != 0 && *xd->resstr != 0)
	{
		rc = xvt_tk_optlist(xd, xd->resstr, TRUE);
		if (rc)
			return (rc);
	}

	/*--------------------------------------------------------------------
	 * get resource name to use
	 */
	if (xd->resname != 0 && *xd->resname != 0)
		resname = xd->resname;
	else if (xd->appname != 0 && *xd->appname != 0)
		resname = xd->appname;
	else
		resname = XVT_S_OPT_RESNAME;

	/*--------------------------------------------------------------------
	 * check if a resource file was specified
	 */
	if (xd->resfile != 0)
	{
		xvt_path_resolve(res_path, xd->resfile);
		if (res_path != 0)
			xvt_tk_resfile_load(xd, res_path, resname);
		return (XVT_OK);
	}

	/*--------------------------------------------------------------------
	 * process all files first
	 */
	for (p=xvt_tk_res_files; *p; p++)
	{
		xvt_path_resolve(res_path, *p);
		if (*res_path != 0)
			xvt_tk_resfile_load(xd, res_path, resname);
	}

	/*--------------------------------------------------------------------
	 * now process all dirs
	 */
	for (p=xvt_tk_res_dirs; *p; p++)
	{
		sprintf(res_skel, "%s/%s", *p, resname);
		xvt_path_resolve(res_path, res_skel);
		if (*res_path != 0)
			xvt_tk_resfile_load(xd, res_path, resname);
	}

	/*--------------------------------------------------------------------
	 * now process all paths
	 */
	for (p=xvt_tk_res_paths; *p; p++)
	{
		const char * path = *p;
		const char * ptr;

		if (*path == '$')
		{
			path = getenv(path+1);
			if (path == 0 || *path == 0)
				continue;
		}

		ptr = path;
		while (*ptr != 0)
		{
			char *	t;

			for (t=res_skel; *ptr; ptr++)
			{
				if (*ptr == ':')
					break;
				*t++ = *ptr;
			}
			*t = 0;

			if (*ptr == ':')
				ptr++;

			if (*res_skel == 0)
				strcpy(res_skel, ".");
			strcat(res_skel, "/");
			strcat(res_skel, resname);
			xvt_path_resolve(res_path, res_skel);
			if (*res_path != 0)
				xvt_tk_resfile_load(xd, res_path, resname);
		}
	}

	/*--------------------------------------------------------------------
	 * if options dump wanted, do it here
	 */
	if (xd->tkopts_fp != 0)
	{
		xvt_tk_options_dump(xd, xd->tkopts_fp, TRUE);
	}

	return (XVT_OK);
}
